// Command line interface for the XML<->PDF translation tool.

#include <stdlib.h>
#include <stdio.h>

// Windows crap.
#ifdef WIN32
#include <io.h> // For _setmode()
#include <fcntl.h> // For _O_BINARY
#define fileno(x) _fileno(x)
#endif

#include "XMLReader.hpp"
#include "PDFReader.hpp"
#include "ObjParser.hpp"
#include "PdfObjects.hpp"
#include "pdfc_basicobjects.h"
#include "stringclass.h"
#include "XMLModule_highLevel.hpp"
#include "XMLModule_lowLevel.hpp"

#define CMDLINE_DEBUG false // Debugging output.
#define DEBUG_PROGRESS_STAGES true
#define DEBUG_PROGRESS_STAGES_PAUSE false

int itype; // Input type.
int otype;
char* ifile=NULL; // Input file.
char* ofile=NULL;
bool itype_isbound=false; // Set when arg_identify() sets the input arg type, and checked on subsequent calls to
	// arg_identify(), to prevent user from accidentally specifying input type twice.
bool otype_isbound=false;
bool itype_isboundhard=false; // Set when itype is specified explicitly (this distinguishes between explicit and
	// implicit sets; implicits can be overridden, but explicits can't).
bool otype_isboundhard=false;
bool ifile_isbound=false;
bool ofile_isbound=false;
int special_mode=-1; // Set to zero if command line invokation is with "xml2pdf", to one if "pdf2xml", and
	// remains at -1 if anything else (e.g. the normal "pdfc").

void show_error(int e) {
	printerr(EFATAL|ECOMMANDLINE_GENERIC, "%s%s%s", errmsgs[1], errmsgs[e], errmsgs[2]);
}

void show_help() {
	printf("%s%s%s", helpmsgs[0], helpmsgs[1], helpmsgs[2]);
	exit(1);
}

void arg_parse(char* arg) {
	char buf[1024];
	if((special_mode!=-1)&&(*arg!='-')) { // Argument must be a filename with the -if= or -of= omitted.
		if(ifile_isbound) { // Argument must be the output file.
			if(ofile_isbound) { // Both already bound; this is an error.
				show_error(8);
			}
			strcpy(buf, "-of=");
		} else {
			strcpy(buf, "-if=");
		}
		strcat(buf, arg); // Append arg, making a well-formed standard argument.
		arg=buf;
	}

	String sarg=(arg);
	if((sarg==String("-d"))|(sarg==String("--decompose"))) {
		decompose=true;
		return;
	}
	else if((sarg==String("-c"))|(sarg==String("--compress"))) {
		PdfObject::COMPRESS_DATA = 3;
		return;
	}
	else if((sarg==String("-p"))|(sarg==String("--parse"))) {
		parseObjects=true;
		return;
	}
	else if((sarg==String("-li"))|(sarg==String("--lowLevelInput"))) {
		parseHighLevelXML=false;
		return;
	}
	else if((sarg==String("-lo"))|(sarg==String("--lowLevelOutput"))) {
		renderHighLevelXML=false;
		return;
	}

	if(sarg==String("--help")) show_help();
	if(pdfc_strlen(arg)<5) show_error(14); // Must be of the form -xx=y (and y must have at least one character).
	if(sarg.getsubstring(0, 3)==String("-it=")) {
		if(itype_isboundhard) show_error(3);
		if(sarg.getsubstring(4)==String("xml")) {
			itype_isbound=true;
			itype_isboundhard=true;
			itype=0;
			return;
		}
		if(sarg.getsubstring(4)==String("pdf")) {
			itype_isbound=true;
			itype_isboundhard=true;
			itype=1;
			return;
		}
		show_error(4);
	}
	if(sarg.getsubstring(0, 3)==String("-ot=")) {
		if(otype_isboundhard) show_error(5);
		if(sarg.getsubstring(4)==String("xml")) {
			otype_isbound=true;
			otype_isboundhard=true;
			otype=0;
			return;
		}
		if(sarg.getsubstring(4)==String("pdf")) {
			otype_isbound=true;
			otype_isboundhard=true;
			renderHighLevelXML=false;
			otype=1;
			return;
		}
		show_error(6);
	}
	if(sarg.getsubstring(0, 3)==String("-if=")) {
		if(ifile_isbound) show_error(7);
		ifile_isbound=true;
		if(DEBUG_LOG_NULLTERM) printf("nullterm() #7.\n");
		ifile=sarg.getsubstring(4).nullterm();
		if(!itype_isboundhard) { // Set type based on extension.
			if(sarg.getsubstring(sarg.getlength()-4)==String(".xml")) {
				itype=0;
				itype_isbound=true;
			} else if(sarg.getsubstring(sarg.getlength()-4)==String(".pdf")) {
				itype=1;
				itype_isbound=true;
			}
			// No other conditions here; unknown extension isn't necessarily an error, since the
			// type might be explicitly specified with a later argument.
		}
		return;
	}
	if(sarg.getsubstring(0, 3)==String("-of=")) {
		if(ofile_isbound) show_error(8);
		ofile_isbound=true;
		if(DEBUG_LOG_NULLTERM) printf("nullterm() #8.\n");
		ofile=sarg.getsubstring(4).nullterm();
		if(!otype_isboundhard) {
			if(sarg.getsubstring(sarg.getlength()-4)==String(".xml")) {
				otype=0;
				otype_isbound=true;
			} else if(sarg.getsubstring(sarg.getlength()-4)==String(".pdf")) {
				renderHighLevelXML=false;
				otype=1;
				otype_isbound=true;
			}
		}
		return;
	}
	show_error(9);
}

int main(int argc, char* argv[]) {
try {
#ifdef WIN32 // Needed to prevent Windows from automatically translating \n to \r\n (which screws up my character counts).
_setmode(fileno(stdin), _O_BINARY);
_setmode(fileno(stdout), _O_BINARY);
#endif
	init_pscommands(); // It's really important to do this or stuff will blow up badly.
	FILE* ofilehandle;
	PObj* root = NULL;
	PdfRoot* pdfRoot = NULL;
	PDFReader preader;
	ObjParser objParser;
	String argv0=String(argv[0]);
	if(pdfc_strlen(argv[0])>6) {
		if(argv0.getsubstring(argv0.getlength()-7)==String("xml2pdf")) special_mode=0; // Must not allow any
			// path prefix on the name to make the comparison fail.
		if(argv0.getsubstring(argv0.getlength()-7)==String("pdf2xml")) special_mode=1;
	}
	if((argc==2)&&(strcmp(argv[1], "--help")==0)) show_help();
	if((argc==1)&&(special_mode==-1)) show_help(); // Command with zero parameters is valid only in special mode.

	if(special_mode==0) {
		itype_isbound=otype_isbound=true;
		itype=0;
		otype=1;
	}
	if(special_mode==1) {
		itype_isbound=otype_isbound=true;
		itype=1;
		otype=0;
	}

	for(int c=1; c<argc; c++) arg_parse(argv[c]);

	if(!itype_isbound) show_error(4);
	if(!otype_isbound) show_error(6);
	if(CMDLINE_DEBUG) printf("ifile_isbound=%s\nifile is %s\nofile_isbound=%s\nofile is %s\n",
		ifile_isbound?"true":"false", ifile?ifile:"NULL", ofile_isbound?"true":"false", ofile?ofile:"NULL");
	if(ifile_isbound==(ifile==NULL)) show_error(11);
	if(ofile_isbound==(ofile==NULL)) show_error(11);

	if(DEBUG_PROGRESS_STAGES) printf("Main: beginning input stage...\n");
	if(DEBUG_PROGRESS_STAGES_PAUSE) {
		printf("Press any key to continue\n");
		getchar();
	}

	if(ifile_isbound) {
		if(itype==0) { // XML input
			if(CMDLINE_DEBUG) printf("Read XML from file.\n");

			if(parseHighLevelXML) {
				XMLModule_highLevel module;
				XMLHandler handler(&module);
				XMLReader::parseXmlFile(&handler, ifile);
				pdfRoot = module.getRoot();
			}
			else {
				XMLModule_lowLevel module;
				XMLHandler handler(&module);
				XMLReader::parseXmlFile(&handler, ifile);
				root = module.getRoot();
			}
		} else { // PDF input
			if(CMDLINE_DEBUG) printf("Read PDF from file.\n");
			FileReader data(ifile);
			root= preader.parse(&data);
		}
	} else {
		if(itype==0) {
			if(CMDLINE_DEBUG) printf("Read XML from stream.\n");
			if(parseHighLevelXML) {
				XMLModule_highLevel module;
				XMLHandler handler(&module);
				XMLReader::parseXmlFilePointer(&handler, stdin);
				pdfRoot = module.getRoot();
			}
			else {
				XMLModule_lowLevel module;
				XMLHandler handler(&module);
				XMLReader::parseXmlFilePointer(&handler, stdin);
				root = module.getRoot();
			}
		} else {
			if(CMDLINE_DEBUG) printf("Read PDF from stream.\n");
			FileReader data(ifile);
			root= preader.parse(&data);
		}	
	}

	if(pdfRoot == NULL && (parseObjects || renderHighLevelXML)) {
		if(DEBUG_PROGRESS_STAGES) printf("Main: intermediate input stage and beginning low level to high level conversion ...\n");
		if(DEBUG_PROGRESS_STAGES_PAUSE) {
			printf("Press any key to continue\n");
			getchar();
		}
	
		pdfRoot = objParser.parse(root);

		delete root;
		root = NULL;

		if(pdfRoot == NULL) {
			printerr("ObjParser returned NULL\n");
		}
	}

	if(DEBUG_PROGRESS_STAGES) printf("Main: ending input stage and beginning output stage...\n");
	if(DEBUG_PROGRESS_STAGES_PAUSE) {
		printf("Press any key to continue\n");
		getchar();
	}
	
	if(renderHighLevelXML && otype==0) {
		if(DEBUG_PROGRESS_STAGES) printf("Main: output stage: render high level XML...\n");
		if(DEBUG_PROGRESS_STAGES_PAUSE) {
			printf("Press any key to continue\n");
			getchar();
		}
 
		if(ofile_isbound) {
			ofilehandle=fopen(ofile, "wb");
			if(ofilehandle==NULL) show_error(13);

			if(CMDLINE_DEBUG) printf("Write HighLevel XML to file.\n");
			fprintf(ofilehandle, "%s", pdfRoot->render_xml(0).nullterm());

			fclose(ofilehandle);
		}
		else {
			if(CMDLINE_DEBUG) printf("Write HighLevel XML to stream.\n");
			printf("%s", pdfRoot->render_xml(0).nullterm());
		}
		delete pdfRoot;
		pdfRoot = NULL;
	}
	else {
		if(pdfRoot != NULL) {
			if(DEBUG_PROGRESS_STAGES) printf("Main: output stage: renderBasicObjects...\n");
			if(DEBUG_PROGRESS_STAGES_PAUSE) {
				printf("Press any key to continue\n");
				getchar();
			}
	
			PObj *rootObj = pdfRoot->renderBasicObjects();

			if(DEBUG_PROGRESS_STAGES) printf("Main: output stage: deleting objects...\n");

			if(rootObj != NULL) {
				if(root != NULL)
					delete root;
				root = rootObj;
			}
			else {
				printerr("Error - could not convert high level representation to low level\n");
			}
			delete pdfRoot;
			pdfRoot = NULL;

			if(DEBUG_PROGRESS_STAGES) printf("Main: output stage: done deleting objects...\n");
		}
	
		if(DEBUG_PROGRESS_STAGES) printf("Main: final output stage: rendering to text...\n");
		if(DEBUG_PROGRESS_STAGES_PAUSE) {
			printf("Press any key to continue\n");
			getchar();
		}
	
		if(ofile_isbound) {
			ofilehandle=fopen(ofile, "wb");
			if(ofilehandle==NULL) show_error(13);
			rewind(ofilehandle);
			if(otype==0) { // XML output.
				if(CMDLINE_DEBUG) printf("Write XML to file.\n");
				String s;
				root->render_xml(0, s);
				fprintf(ofilehandle, "%s", s.nullterm());
			} else {
				if(CMDLINE_DEBUG) printf("Write PDF to file.\n");
				String r;
				root->render_pdf(NULL, r);
				uint written=fwrite((void*)(r.getrawstring()), 1, r.getlength(), ofilehandle);
				if(written!=r.getlength()) {
					fprintf(stderr, "Error writing to output file %s\n", ofile);
					exit(1);
				}
			}
			fclose(ofilehandle);
		} else {
			if(otype==0) {
				if(CMDLINE_DEBUG) printf("Write XML to stream.\n");
				String s;
				root->render_xml(0, s);
				printf("%s", s.nullterm());
			} else {
				if(CMDLINE_DEBUG) printf("Write PDF to stream.\n");
				String r;
				root->render_pdf(NULL, r);
				fwrite((void*)(r.getrawstring()), 1, r.getlength(), stdout);
			}
		}
	}
	if(DEBUG_PROGRESS_STAGES) printf("Main: ending output stage...\n");
	if(DEBUG_PROGRESS_STAGES_PAUSE) {
		printf("Press any key to continue\n");
		getchar();
	}
	if(root != NULL)
		delete root;

	if(DEBUG_PROGRESS_STAGES) printf("Main: Done with program...\n");
	if(DEBUG_PROGRESS_STAGES_PAUSE) {
		printf("Press any key to continue\n");
		getchar();
	}
} /* End try block. */ catch(Err* e) {
	fprintf(stderr, "Main - Exception thrown:\n");
	e->print(stderr);
}
	return 0;
} // End main.
