#include "dll_export.h"


#ifndef _PDFOBJECT_HPP
#define _PDFOBJECT_HPP

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include "pdfc_base.h"
#include "pdfc_basicobjects.h"
#include "Hash.hpp"
#include "stringclass.h"

enum DllExport PdfType { p_object, p_root, p_page, p_image, p_colorspace, p_font, p_page_resource, p_annot, p_text_annot, p_link, p_bookmark, p_function, p_ext_graphics_state };

class PdfObject;
class PdfPage;
class PdfResourceRef;
class PdfResource;
class PdfFont;
class PdfColorSpace;
class PdfImage;
class StringList;
class PdfAnnot;
class PdfTextAnnot;
class PdfLink;
class PdfBookmark;
class PdfLinkList;
class PdfFunction;
class PdfList;

class DllExport PdfObject : public HashableObj {
public:
	static const int OUTPUT_LEVEL /* =1 */;
	static int COMPRESS_DATA /* =0 */;

	int id;
	PdfObject *parent;

	PdfObject();
	virtual ~PdfObject();

	virtual PdfObject* copy();
	virtual void copyFields(PdfObject* alpha);
	virtual PdfType objType() { return p_object; }
	virtual char* objTypeStr() { return "object"; }
	virtual String render_xml(int depth);
	virtual PObj* renderBasicObjects();

	int static getNextId();
	int static getNumIds();
	void static resetId();
protected:
	static int nextId;
	static int fontNum;
	static int imageNum;
	static int functionNum;
};

class DllExport PdfRoot : public PdfObject {
public:
	PdfPage *pages;
	PdfResource *resources;
	PdfLinkList *linkList;
	Hash resourceHash;
	Hash resourceIdHash;
	Hash pageHash;

	String author;
	String creationDate;
	String modDate;
	String creator;
	String producer;
	String title;
	String subject;
	String keywords;

	PdfBookmark *bookmarks;

	bool showBookmarks;
	PdfFont *defaultFont;

	PdfList *colorSpaces;
	PdfList *functions;

	PdfRoot();
	virtual ~PdfRoot();

	virtual PdfObject* copy();
	virtual void copyFields(PdfRoot* alpha);
	virtual PdfType objType() { return p_root; }
	virtual char* objTypeStr() { return "root"; }
	virtual String render_xml(int depth);
	virtual void addPage(PdfPage *page);
	virtual PdfPage* getPage(int pageId);
	virtual void addResource(PdfResource *resource);
	virtual void addLink(PdfLink *newLink);
	virtual void addLink(PdfLink *newLink, PObj *newObj);
	virtual void addBookmark(PdfBookmark *newBookmark);
	virtual void addBookmark(String newTitle, PdfPage *newPage);
	virtual void addColorSpace(PdfColorSpace *newColorSpace);
	virtual PdfColorSpace* getColorSpace(int id);
	virtual void addFunction(PdfFunction *newFunction);
	virtual PdfFunction* getFunction(int id);
	virtual PObj* renderBasicObjects();

	virtual String pdfFormatDate(time_t newTime);
private:
};

class DllExport PdfResource : public PdfObject {
public:
	PdfResource *prev, *next;
	String type;
	int contentId;
	int refId;
	int resourceId;

	static int nextResourceId;

	char *data;
	unsigned long dataLength;
	StringList *filters;

	String filename;
	String encoding;

	PdfResource();
	virtual ~PdfResource();
	virtual PdfObject* copy();
	virtual void copyFields(PdfResource* alpha);
	virtual String render_xml(int depth);
	virtual PObj* renderBasicObjects(PDF_root *root);

	virtual void addFilter(String newFilter);
	virtual void decode(String encoding);
	virtual void encode(String encoding);
	virtual void decode();

	virtual void loadData();
	virtual void saveData();

	static int getNextResourceId();
private:
};

class DllExport PdfPage : public PdfResource {
public:
	double mediaBox[4];
	PObj *contentObj;
	PdfAnnot *annotations;
	PdfResourceRef *resourceRefList;
	PdfRoot *root;

	static double DEFAULT_LLX;
	static double DEFAULT_LLY;
	static double DEFAULT_URX;
	static double DEFAULT_URY;

	PdfPage();
	virtual ~PdfPage();

	virtual PdfObject* copy();
	virtual PdfObject* copy(PdfRoot *root, bool copyContent=true);
	virtual void copyFields(PdfPage* alpha, bool copyContent);
	virtual String render_xml(int depth);
	virtual PdfType objType() { return p_page; }
	virtual char* objTypeStr() { return "page"; }
	virtual PObj* renderBasicObjects(PObj *root, PObj *parent, PObj *pagesArray);
	virtual PObj* renderBasicObjectsForContents(PObj *obj);
	virtual void addResourceRef(PdfResourceRef *ref);
	virtual PdfResource* getResource(String name);
	virtual void addAnnotation(PdfAnnot *annot);
	virtual void addCommand(String name);
	virtual void addCommand(PDF_pscommand *command);
	virtual void addContent(PObj *node);

	virtual void enableSharedContent();
	virtual void switchStream();
	virtual void addStream(PDF_collection *newStream);
};

class DllExport PdfFont : public PdfResource {
public:
	String subType;
	String fontName;
	int firstChar;
	int lastChar;
	long *widths;
	int numWidths;

	int fontDescId;
	int ascent;
	int descent;
	int flags;
	double fontBBox[4];

	String dataSubType;

	PdfFont();
	PdfFont(char * newName);
	virtual ~PdfFont();

	virtual PdfObject* copy();
	virtual void copyFields(PdfFont* alpha);
	virtual String render_xml(int depth);
	virtual PdfType objType() { return p_font; }
	virtual char* objTypeStr() { return "font"; }
	virtual PObj* renderBasicObjects(PDF_root *root);
	virtual bool isBase14Font();

	virtual void showWidths();

	void initBase14Font();

	void setFontCourierBold();
	void setFontCourierBoldOblique();
	void setFontCourierOblique();
	void setFontCourier();
	void setFontHelveticaBold();
	void setFontHelveticaBoldOblique();
	void setFontHelveticaOblique();
	void setFontHelvetica();
	void setFontSymbol();
	void setFontTimesBold();
	void setFontTimesBoldItalic();
	void setFontTimesItalic();
	void setFontTimesRoman();
	void setFontZapfDingbats();
private:
	virtual PObj* renderBasicObjectsForFontDescriptor();
};

class DllExport PdfColorSpace : public PdfResource {
public:
	String catagory;
	PdfColorSpace *base;
	int baseId;
	double whitePoint[3]; // [ Xw, Yx, Zw ]
	double blackPoint[3]; // [ Xb, Yb, Zb ]
	double gamma[3]; // [ Xg, Yg, Zg ]

	double matrix[9]; // [ Xr, Yr, Zr, Xg, Yg, Zg, Xb, Yb, Zb  ]

	double range[4]; // aMin, aMax, bMin, bMax;

	int hival;
	String lookup;

	String separationName;
	PdfColorSpace *altSpace;
	PdfFunction *tintTransform;
	int altSpaceId;
	int tintTransformId;

	bool isPageResource;

	PdfColorSpace();
	PdfColorSpace(char *newCatagory);
	virtual ~PdfColorSpace();

	virtual PdfObject* copy();
	virtual void copyFields(PdfColorSpace* alpha);
	virtual String render_xml(int depth);
	virtual PdfType objType() { return p_colorspace; }
	virtual char* objTypeStr() { return "ColorSpace"; }
	virtual PObj* renderBasicObjects(PDF_root *root);
	virtual bool isSimpleCatagory();
};

class DllExport PdfImage : public PdfResource {
public:
	String format;
	int width;
	int height;
	int bitsPerPixel;
	String name;
	PdfColorSpace *colorSpace;
	int colorSpaceId;
	
	PdfImage();
	virtual ~PdfImage();

	virtual PdfObject* copy();
	virtual void copyFields(PdfImage* alpha);
	virtual String render_xml(int depth);
	virtual PdfType objType() { return p_image; }
	virtual char* objTypeStr() { return "image"; }
	virtual PObj* renderBasicObjects(PDF_root *root);
private:
};

class DllExport PdfResourceRef : public PdfObject {
public:
	PdfResourceRef *prev, *next;
	PdfResource *resource;
	String name;
	String type;
	int resourceId;

	PdfResourceRef();

	virtual PdfObject* copy();
	virtual PdfObject* copy(bool copyResourcePointer);
	virtual void copyFields(PdfResourceRef* alpha, bool copyResourcePointer);
	virtual String render_xml(int depth);
	virtual PdfType objType() { return p_page_resource; }
	virtual char* objTypeStr() { return "resourceRef"; }
	virtual PObj* renderBasicObjects();
private:
};

class DllExport PdfAnnot : public PdfObject {
public:
	PdfAnnot *prev, *next;
	String type;
	double rect[4];

	PdfAnnot();

	virtual PdfObject* copy();
	virtual void copyFields(PdfAnnot* alpha);
	virtual String render_xml(int depth);
	virtual PdfType objType() { return p_annot; }
	virtual char* objTypeStr() { return "annot"; }
	virtual PObj* renderBasicObjects();
private:
};

class DllExport PdfTextAnnot : public PdfAnnot {
public:
	String contents;
	bool isOpen;

	PdfTextAnnot();

	virtual PdfObject* copy();
	virtual void copyFields(PdfTextAnnot* alpha);
	virtual PdfType objType() { return p_text_annot; }
	virtual char* objTypeStr() { return "text_annot"; }
	virtual String render_xml(int depth);
	virtual PObj* renderBasicObjects();
private:
};

class DllExport PdfLink : public PdfAnnot {
public:
	String destKey;
	PdfPage *destination;

	PdfLink(PdfRoot* root);

	virtual PdfObject* copy();
	virtual void copyFields(PdfLink* alpha);
	virtual PdfType objType() { return p_link; }
	virtual char* objTypeStr() { return "link"; }
	virtual String render_xml(int depth);
	virtual PObj* renderBasicObjects();
private:
	PdfRoot *rootObj;
};

class DllExport PdfBookmark : public PdfObject {
public:
	String title;
	String destKey;
	PdfPage *destination;
	PdfBookmark *parent;
	PdfBookmark *prev, *next;
	PdfBookmark *first, *last;
	bool isOpen;
	long count;

	PdfBookmark();
	PdfBookmark(String title, PdfPage *page);
	virtual ~PdfBookmark();

	virtual PdfObject* copy();
	virtual void copyFields(PdfBookmark* alpha);
	virtual PdfType objType() { return p_bookmark; }
	virtual char* objTypeStr() { return "bookmark"; }
	virtual String render_xml(int depth);
	virtual PObj* renderBasicObjects();
	virtual PObj* renderBasicObjects(PObj *root);
	virtual void addBookmark(PdfBookmark *newBookmark);
	virtual void addBookmark(String newTitle, PdfPage *newPage);
	virtual void findPages(PdfRoot *root);
};

class DllExport PdfFunction : public PdfResource {
public:
	int functionType;
	int numIn;
	double *domain;
	int numOut;
	double *range;
	int *size;
	int bitsPerSample;
	int order;
	double *encode;
	double *decode;

	PdfFunction();
	virtual ~PdfFunction();
	virtual PdfObject* copy();
	virtual void copyFields(PdfFunction* alpha);
	virtual PdfType objType() { return p_function; }
	virtual char* objTypeStr() { return "function"; }
	virtual String render_xml(int depth);
	virtual PObj* renderBasicObjects();
};

class DllExport PdfExtGraphicsState : public PdfResource {
public:
	int sa;
	int op;
	PdfFunction *bg;
	PdfFunction *ucr;
	// PdfObject *tr;
	// PdfObject *ht;
	int *htp;

	PdfExtGraphicsState();
	virtual ~PdfExtGraphicsState();
	virtual PdfObject* copy();
	virtual void copyFields(PdfExtGraphicsState* alpha);
	virtual PdfType objType() { return p_ext_graphics_state; }
	virtual char* objTypeStr() { return "extended graphics state"; }
	virtual String render_xml(int depth);
	virtual PObj* renderBasicObjects(PDF_root *root);
};

class DllExport PdfLinkList {
public:
	PdfLink *link;
	PObj *obj;
	PdfLinkList *next;

	virtual ~PdfLinkList();

	virtual PdfLinkList* copy();
	virtual void copyFields(PdfLinkList* alpha);

	PdfLinkList(PdfLink *newLink);
	PdfLinkList(PdfLink *newLink, PObj *newObj);
private:
};

class DllExport StringList {
public:
	String str;
	StringList *next;

	StringList();
	StringList(String newString);
	virtual ~StringList();

	virtual StringList* copy();
	virtual void copyFields(StringList* alpha);
private:
};

class DllExport PdfList {
public:
	PdfObject *object;
	PdfList *next;

	PdfList();
	PdfList(PdfObject *newObject);
};

#endif
