#include "dll_export.h"


#ifndef PDFC_BASE_ALREADY_INCLUDED
#define PDFC_BASE_ALREADY_INCLUDED

/*
Base internal structures and functions for the PDF tool.
*/

#include <stdlib.h> // For itoa() function.
#include <stdio.h> // For sprintf() function.
#include <string.h> // For itoa() for unix.

// All of the following stuff just for file I/O.
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <sys/stat.h>
#include <fcntl.h>

#ifdef WIN32
#define stat _stat // M$ compatibility crap again.
#else
#include <unistd.h>
#endif

#include "stringclass.h"
#include "thash.h"
using Thash::HashTable;

unsigned DllExport long pdfc_ntohl(unsigned long arg);
unsigned DllExport short pdfc_ntohs(unsigned short arg);

#define HASHTABLESIZE 42743 // Arbitrary.
#define MAXDEPTH 1000 // Arbitrary.  Sets the maximum depth of the hierarchy.

#define PDFC_BASE_DEBUG 0

enum DllExport Tsorttype {by_id}; // Currently by_id is the only sort type supported.
enum DllExport ObjectType {t_base_object, t_blob, t_boolean, t_integer, t_real, t_string, t_hexstring, t_name,
	t_array, t_dictionary, t_stream, t_document, t_reference, t_root, t_trailer, t_deletion, t_null,
	t_collection, t_numeric, t_pscommand, t_box, t_inline_image, t_shared_collection};

extern char* DllExport PDF_FILE_HEADER;
extern char* DllExport errmsgs[];
extern char* DllExport helpmsgs[];
extern bool DllExport decompose; // Indicates whether blob xml output should decompose streams into separate files.
extern bool DllExport parseObjects; // Indicates if the document structure should pe parsed (2nd pass parsing)
extern bool DllExport parseHighLevelXML;
extern bool DllExport renderHighLevelXML;

class DllExport PObj;
class DllExport PdfPage;

uint DllExport gethashvalue(uint id);

class DllExport PObjList { // Linked list class, for linked lists other than those built into class PObj.
	// Currently used for efficient hierarchical structuring; objects have linked lists of their subobjects.
public:
	PObj* theobject;
	PObjList* next;
	bool deleteObjectOnDestruct;

	PObjList();
	PObjList(bool deleteObjectFlag);
	~PObjList();
};

class DllExport Xrefentry { // Entry in the xref table for PDF output.
public:
	uint objnumber; // Can't just use the index in the array (see definition of Xreftable below) as the object
		// number, since objects might be stored out of order (which is valid in PDF).
	uint gennumber;
	uint offset;
	uint totalsize; // Accumulated during the rendering of an object.  This value plus offset will become
		// the offset of the next Xrefentry in the Xreftable.  Therefore this really does indicate the _total_
		// size of the object, including the "%d %d obj" header and "endobj" footer.
	Xrefentry();
};

class DllExport Xreftable { // Xref table for PDF output.
public:
	static int MAX_TABLE_SIZE;

	Xrefentry* entry; // 65536 is the max number of objects in PDF file.
	uint arraysize; // Indicates the number of elements used in the array (and therefore the index of
		// the first free element).  This is incremented at the end of a rendering function for an indirect
		// object so that the rendering functions for any direct objects contained within that indirect object
		// know that arraysize indicates the array element whose totalsize attribute is supposed to be accumulated.
	Xreftable();
	~Xreftable();
	void accobjsize(int n); // Add n to the totalsize of the element indicated by arraysize.
};

//---------------------------------------------------------

class DllExport PObj {
public:
	uint id; // PObj ID.  Must be globally unique within each hashtable.  Cannot be 0.
	int generation;

	PObj* parentp; // Pointer to this objects pointer.  Set when this object is added as a subobject to some

    uint unique_id; // Used in the case that something other than the encapsulated PObj's ID is needed as the
                      // hash key.
    PObjList* subobjects; // Linked list of subobjects of this object.
    PObjList* lastsub; // Pointer last subobject in list (only a performance optimization).
	PdfPage* parentPage;
	Xreftable* xref; 

	PObj();
	virtual ~PObj();
	virtual PObj* copy();
	virtual void copyFields(PObj *alpha);
	virtual bool isNumeric() { return false; }
	virtual void setValue(char *str);
	virtual void setValue(String str);
	virtual uint getlength() const;
	virtual void render_xml(int depth, String& dest);
	virtual void render_xml_start(int depth, String& dest);
	virtual void render_xml_children(int depth, String& dest);
	virtual void render_xml_end(int depth, String& dest);
	virtual void render_pdf(Xreftable* x, String& dest);
	virtual void render_pdf_start(Xreftable* x, String& dest);
	virtual void render_pdf_children(Xreftable* x, String& dest);
	virtual void render_pdf_end(Xreftable* x, String& dest);
	virtual Xreftable* render_pdf_top_level_start(Xreftable* x);
	virtual void render_pdf_top_level_end(Xreftable* x);
	virtual void rend_pdf_ind_begin(uint id, uint gen, Xreftable* x, String& dest); // Render PDF header for indirect object.
	virtual void rend_pdf_ind_end(Xreftable* x, String& dest); // Render PDF footer for indirect object.
	virtual ObjectType get_type();
	virtual char* get_printable_type();
	virtual void showObject();

    virtual void addsub(PObj* newsub, bool deleteObjectOnDestruct=true); // Add a subobject to the subobjects list.

    void deletelist(); // Delete the linked list chained from this object.
    uint gettotalsize() const;
	void setParentPage(PdfPage *newPage);
    uint countsubs(); // Self explanatory.
    PdfPage* getParentPage();
};

//---------------------------------------------------------

class DllExport PDF_collection : public PObj {
public:
	bool sharedStream;
	bool streamAlreadyRendered;

	PDF_collection();
	virtual PObj* copy();
	virtual ObjectType get_type();
	virtual char* get_printable_type();
	virtual void render_xml(int depth, String& dest);
	virtual void render_xml_start(int depth, String& dest);
	virtual void render_xml_end(int depth, String& dest);
	virtual void render_pdf(Xreftable* x, String& dest);
private:
};

class DllExport Hashtable {
public:
	Hashtable();
	~Hashtable();

	void add(PObj* newobject);
	void add(PObj* newobject, uint id); // add objects with arbitrary ID unrelated to the encapsulated PObj's ID.
	void remove(uint oldid); 
	PObj* findobject(uint id); // Returns a single object.
private:
    HashTable<int, PObj*> h;
};

#endif
