#include "dll_export.h"


#ifndef _PDFC_ERRORS_H
#define _PDFC_ERRORS_H

#include <stdlib.h>

/* Error messages for PDFC.
All error codes are positive (to distinguish from libc errors, which are negative).
Bit 31 is msb (sign bit), 0 is lsb.
Bit 24-27 are the severity code.  0 indicates a minor warning (default severity), 0xF indiates a fatal error,
and other values are reserved.  printerr() (see below) decides whether to continue or exit based on severity;
if the reserved severity codes are used in the future, the switch() in printerr() must be updated.
Bits 16-23 group errors.  Bits 8-15 can be used to subgroup within groups.
Bits 0-7 indicate individual errors within each group.  Bit 7 is one for shared errors (see below), and
zero for group-specific errors.

Note that this mechanism allows multiple versions of each error, one for each possible severity.  This is
a feature, not a bug.
A group code of all zeroes is reserved.  An individual error code of all zeroes is reserved.  A subgroup
code of all zeroes is not reserved; it is the default.

Usage: all errors flagged in the code should be of one of the following forms (pipe character is bitwise OR):
A|B|C where A is an optional severity flag, B is a group code, and C is a shared error code.
A|B where A is an optional severity flag and B is a group-specific error code.
Each group-specific error code already includes the appropriate group code, so it doesn't have to be
included again.

All error-related defines begin with "E".
*/

typedef unsigned int uint;
typedef unsigned char uchar;

extern FILE* errstream; // All error messages should be written to this stream.
void DllExport printerr(int code, const char* first, ... ); // Primary error handler.
void DllExport printerr(const char* first, ... ); // Old handler for compatibility with old code.

class DllExport Err { // Instantiated to be thrown via C++ exception handling mechanism.
public:
	int code; // Mandatory.
	char* message; // Optional.
	Err* next; // Optional; used if a program needs to wrap another error message with its own (e.g. a catch block rethrows an error, but with more information added).
	void print(FILE* destination); // Print this message (and the ones in the linked list, if any) to e.g. stderr.
	Err(uint newcode);
	Err(uint newcode, char* newmessage);
	Err(uint newcode, char* newmessage, Err* nexterror);
	~Err();
};

#define ESTRICTSYNTAX_IS_FATAL false // Temporarily a #define; this will become a user-configurable variable later.
#define EFATAL_IS_FATAL true // Believe it or not, Steve wants this option.

// Severity codes.
#define ESEVERITY			0x0F000000 // Severity code bitmask.  Use e.g. if((errocode&ESEVERITY)==EFATAL) to see whether errcode is FATAL.
#define EMINORWARN			0x00000000 // This is the default.
#define ESTRICTSYNTAX		0x01000000 // Whether this is fatal is configurable.
#define EFATAL				0x0F000000

// Group codes.
#define EGROUP				0x000F0000 // Group bitmask.  Use e.g. if((errocode&EGROUP)==EFOO) to see whether errcode is in group FOO.
#define EOTHER				0x00000000
#define ESYS				0x00010000 // OS, libc, etc. errors.
#define ESTRINGCLASS		0x00020000
#define ECOMMANDLINE		0x00030000 // Error in the command line interface (cmdline.cpp).
#define EPDFIN				0x00040000
#define EPDFOUT				0x00050000
#define EXMLIN				0x00060000
#define EXMLOUT				0x00070000

// Shared errors.  These are errors which are applicable to multiple groups.
#define	ESHARED				0x00008000 // Bitmask for shared errors.
#define EINTERNALERROR		ESHARED|0x00000001 // Caused by a bug in the program.
#define ENOTIMPLEMENTED		ESHARED|0x00000002 // Feature is just a stub, and not yet implemented.
#define ENULLPOINTER		ESHARED|0x00000003 // NULL pointer bug in program.
#define EWRONGARGTYPE		ESHARED|0x00000004 // An argument was of the wrong data type at runtime.
#define EWRONGARGCOUNT		ESHARED|0x00000005 // Wrong number of arguments was given.
#define EARGOUTOFRANGE		ESHARED|0x00000006 // An argument to a function was out of the allowed range.
#define EUNKNOWNCOMMAND		ESHARED|0x00000007 // An unrecognized command was issued somewhere.
#define EINVALIDFILE		ESHARED|0x00000008 // File is corrupt/not of the right format/etc.
#define ENOTFOUND			ESHARED|0x00000009 // A requested object was not found.
#define ETOOMUCHDATA		ESHARED|0x0000000A // More data given than can be logically used in the context of the code that raised the error.
#define EDIVBYZERO			ESHARED|0x0000000B // More data given than can be logically used in the context of the code that raised the error.

// Group-specific errors.

#define ESYS_OUTOFMEM			ESYS|0x00000001
#define ESYS_FILEOPENERROR		ESYS|0x00000002 // Failed to open a specified file.
#define ESYS_FILETOOBIG			ESYS|0x00000003 // File size exceeded available memory.

#define EOTHER_GENERICERROR		EOTHER|0x00000001 // Generic error (placeholder for a better message).  All occurences of this should be replaced with better messages.

#define ESTRINGCLASS_INVALIDNUMBERSTRING	ESTRINGCLASS|0x00000001 // Invalid string passed to atoi() function.
#define ESTRINGCLASS_ATOIOVERFLOW			ESTRINGCLASS|0x00000002 // Too large a number (greater than 2**32 - 1) was passed to atoi() function.

#define ECOMMANDLINE_GENERIC	ECOMMANDLINE|0x00000001

#define EPDFIN_OUTOFSPEC		EPDFIN|0x00000001 // PDF document being read violates the spec.

//#define REDEFINE_NEW_OPERATOR // HACK: this should go in the makefile.
#ifdef REDEFINE_NEW_OPERATOR

void* operator new(size_t size);
void operator delete(void* ptr);

#endif

#endif
