#include "dll_export.h"


#ifndef STRINGCLASS_ALREADY_INCLUDED
#define STRINGCLASS_ALREADY_INCLUDED

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "pdfc_errors.h"

#ifdef WIN32
#include <windows.h> // For global malloc/free stuff.
#endif

#define pdfc_strlen(x) ((x)==NULL?0:strlen(x))

#define DEBUG_LOG_NULLTERM false // Log to stdout all calls to nullterm() and the location of the call.

#ifndef BSTR // This makes this library compatible with, but not dependent on, Microsoft's headers.
typedef unsigned short* BSTR;
#endif
#ifndef NULL
#define NULL 0
#endif

uint DllExport standard_hash(const char* data, const uint size); // Currently just a sum modulo 2**32.

/* About fragmented, reference-counted, copy-on-write strings:
String points to the head and tail of a doubly linked list of SFPointer objects.  Each SFPointer object either
points to a StringFragment object or points to the head of another SFPointer list, each element of which in turn
points to a StringFragment object (and this nesting can go arbitrarily deep).  Each StringFragment object
contains a counted character array.
Each String object also contains a counted character array of fixed size into which the string can grow.
When data is appended to the string, it is copied directly into the end of this array if it will fit.
Otherwise, the fragmentation functionality is used: if the data being appended is a raw character array,
it is copied into a new StringFragment and a SFPointer to that StringFragment is appended to the String's
SFPointer list.  If the data being appended is a String object, then a new SFPointer object is appended to
the old String's SFPointer list, and that new SFPointer points to the SFPointer list of the String being
appended.  This latter operation is what makes append operations be efficient and run in constant time:
no actual character data gets copied (unless that data is small enough to fit in the shortstring buffer
of the destination string).
Of course, when the client attempts to view the string as a flat character array (such as with a call to
getrawstring or nullterm), it is necessary to actually copy all the data for the string to a flat array.
But this string class is designed for efficiently performing very large quantities of variable size
append operations, with low emphasis on raw character array access, so the tradeoff is appropriate.
*/

class SFPointer;

class DllExport StringFragment { // Fragment of a string.  Reference counted so that copy on write can be implemented.
	// A fragment may be a character array or a list of other fragments.
private:
	char* s;
	uint length;
	int references; // Number of SFPointer objects which currently point to this fragment.
public:
	StringFragment(const uint newlength); // Allocate the memory, but don't copy anything (used in String::compact()).
	StringFragment(const char* news); // Takes null terminated string.
	StringFragment(const char* news, const uint newlength); // Takes counted string.
	StringFragment(const StringFragment& frag);
	~StringFragment(); // Destroys just this object.
	void reference(); // Increase references by 1.
	void unreference(); // Reduce references by 1, and delete this object if it reaches zero.
	uint getlength() const;
	const char* getrawstring() const;
};

class DllExport SFPointer { // Pointer to a fragment of a string.
private:
	int references; // Number of SFPointer objects which currently point to this SFPointer.
public:
	StringFragment* p_sf;
	SFPointer* p_sfp; // Either p_sfp or p_sf is used for a particular SFPointer instance, but not both.
	SFPointer* prev;
	SFPointer* next;
	SFPointer(StringFragment* newsf);
	SFPointer(SFPointer* newsfp);
	~SFPointer(); // Destroys the ENTIRE list of which this object is a member.
	void reference(); // Increase references by 1.
	void unreference(); // Reduce references by 1, and delete this object if it reaches zero.
	uint getlength() const;
};

class DllExport String { // Practical string class.  Has enough functionality to be at least moderatly useful.
private:
	char* shortstring; // Short string at the end of this string.  Uses direct copying rather than
		// fragment appends for efficient appends of small strings (i.e. a few characters) to this string.
		// This only gets allocated when it's first actually needed, but when it's allocated, it's of a
		// fixed length (SHORTSTRING_MAXSIZE; see the .cpp file), and remains allocated until the entire String object dies.
	uint shortstringlength;
	SFPointer* mainhead; // Head of list of StringFragment objects which comprise the main body of this string.
	SFPointer* maintail; // Tail of the aforementioned list.  This pointer exists only for efficiency.
	uint length; // Can be computed from other member variables, but is here for efficiency.
	bool iscompacted; // Indicates whether string is already compacted, so compact() can exit quickly.
	void compact(); // Convert a possibly fragmented string to a single character array.
	char* compactrecursively(char* cp, SFPointer* stepper); // Internal utility.
	inline void globshortstring(); // Glob shortstring to the end of the main list so the main list contains the entire string.
	void init(); // Variable initialization function (called from every constructor).
public:
// Assignment.
	String(); // Create empty string.
	String(const char* newstring); // Create object from standard null-terminated (C-style) string.
	String(const char* newstring, const uint newlength); // Create object from counted string.
	String(const BSTR newstring); // Create object from BSTR.
	String(uint num, const uint base); // Convert num to string in specified base.  Opposite of the atoi() function.
	String(const String& newstring); // Copy existing String object.
	~String();
	void operator=(const String& newstring);
	void operator=(const char* newstring);
//	uint deserialize(char* buffer); // Deserialize this object from specified buffer.  Return total number
		// of bytes extracted from buffer.
	void setval(const char* newstring, const uint newlength);

// Comparison.
	bool equals(const char* reference, const uint reflength);
	bool operator==(const String& reference);
	bool operator==(char* reference);
	bool operator!=(String& reference);
	bool operator!=(char* reference);
	bool operator>(String& reference); // Magnitude comparisons intended for alphabetical ordering.
	bool operator<(String& reference);
	bool isBinary();

// Modification.
	void operator+=(const String& reference); // Infix version of append().
	void operator+=(const char* reference);
	void append(const String& newsubstring);
	void append(const char* newsubstring);
	void append(const char* newsubstring, uint newsubstringlength);
	void prepend(const String& newsubstring);
	uint replace(const String oldstring, const String newstring); // Replace oldstring with newstring in this string;
		// return the number of replacements performed.
	uint replace(const char oldchar, const char newchar); // Replace oldchar with newchar in this string; return the
		// number of replacements performed.
	void reduceLengthTo(const uint newLength); // Set the length of this string.  Used when manipulating rawstring.
	char* increaseLengthTo(const uint newLength); // Increase length to specified length.  New characters at
		// the end of the string are uninitialized.  Returns pointer to beginning of the new space.
// Retrieval.
	String operator+(const String &reference) const;
	String operator+(const char* reference) const;
	uint atoi(const uint base); // Convert this string to a number in specified base.
		// Valid bases are 2 to 36, inclusive.
	char* nullterm(); // Return copy of string as null-terminated string.
	uint getlength() const; // Return the length of this string.
	const char* getrawstring(); // Return pointer to the raw string.  Used for copying.  Not guaranteed
		// to be valid after subsequent function calls on the object (e.g. reduceLengthTo and incraseLengthTo invalidate it under MS Windows; see the code).
	String getsubstring(uint beginoffset); // Return substring from beginoffset to the end of this string.
	String getsubstring(uint beginoffset, uint endoffset); // Return substring from beginoffset to endoffset.
	String copy() const; // Return copy of this string.
	BSTR getBSTR(); // Return copy of string as BSTR.
	uint hash();
	uint gettotalsize() const; // Get the total amount of state for the object, in bytes.  Used for serializing.
//	uint serialize(char* buffer); // Serialize this object into the specified buffer.  Number of bytes in
		// this object is specified by gettotalsize().  Return total number of bytes inserted into buffer.
	String hexencode();
// Output
	void print();
};

class DllExport StringLite { // Nothing more than a wrapper for a counted string with operator== overloaded.
	// Does NOT own the string itself, unlike class String.
public:
	char* string;
	uint length;
	bool operator==(const StringLite& ref);
	bool operator==(const char* ref);
	StringLite(char* newstring, uint newlength);
};

namespace Thash {
	int hash(const String& arg);
	int hash(const StringLite& arg);
};

#endif
