#include "dll_export.h"


#ifndef THASH_ALREADY_INCLUDED
#define THASH_ALREADY_INCLUDED

#include <utility> // For pair
using std::pair;
using std::make_pair;
#define TABLESIZE 91 // Arbitrary.

namespace Thash {
	template <typename T> class List { // Doubly linked list.
		public:
			T thevalue;
			List* next;
			List* prev;
			List(const T& newvalue, List<T>* newnext=NULL, List<T>* newprev=NULL, bool attach_to_tail=false);
			~List();
			void append(List<T>& newtail);
			void prepend(List<T>& newhead);
	};
	
	template <typename Keytype, typename Valuetype> class HashTable {
		private:
			List<pair<Keytype, Valuetype> > * thelist[TABLESIZE];
		public:
			HashTable();
			~HashTable();
			void add(const Keytype& key, const Valuetype& value);
			bool remove(const Keytype& key);
			bool changeValue(const Keytype& key, const Valuetype& value);
			Valuetype* const getValue(const Keytype& key);
	};

	template <typename T>
	inline void link(List<T>*& l1, List<T>*& l2) {
		if(l1) l1->next=l2;
		if(l2) l2->prev=l1;
	}

	template <typename T>
	inline List<T>::List(const T& newvalue, List<T>* newnext, List<T>* newprev, bool attach_to_tail): thevalue(newvalue),
		next(newnext), prev(newprev) {
		// Automatically set up appropriate links.
		if(prev) prev->next=this;
		if(next) next->prev=this;
	}
	
	template <typename T>
	inline List<T>::~List() {} // Intentionally currently null.

	template <typename T>
	inline void List<T>::append(List<T>& newtail) {
		List<T>* oldtail=this;
		while(oldtail->next!=NULL) oldtail=oldtail->next; // Search for the tail of the list.
		oldtail->next=&newtail; // Append the new element.
		newtail.prev=oldtail;
		newtail.next=NULL; // Ground the end.
	}

	template <typename T>
	inline void List<T>::prepend(List<T>& newhead) {
		List<T>* oldhead=this;
		while(oldhead->prev!=NULL) oldtail=oldhead->prev; // Search for the head of the list.
		oldhead->prev=&newhead; // Append the new element.
		newhead.next=oldhead;
		newhead.prev=NULL; // Ground the end.
	}

	template <typename Keytype, typename Valuetype>
	inline HashTable<Keytype, Valuetype>::HashTable() {
		for(int i=0; i<TABLESIZE; i++) thelist[i]=NULL;
	}

	template <typename Keytype, typename Valuetype>
	inline HashTable<Keytype, Valuetype>::~HashTable() {
		for(int i=0; i<TABLESIZE; i++) if(thelist[i]!=NULL) delete thelist[i];
	}

	template <typename Keytype, typename Valuetype>
	inline void HashTable<Keytype, Valuetype>::add(const Keytype& key, const Valuetype& value) {
		int hashvalue=hash(key)%TABLESIZE;
		remove(key); // Just in case it is already there, don't allow a key collision.
		if(thelist[hashvalue]==NULL) { // No element in this bucket yet.
			thelist[hashvalue]=new List<pair<Keytype, Valuetype> >(make_pair(key, value));
		} else { // Bucket collision.
			thelist[hashvalue]->append(*(new List<pair<Keytype, Valuetype> >(make_pair(key, value))));
		}
	}

	template <typename Keytype, typename Valuetype>
	inline bool HashTable<Keytype, Valuetype>::remove(const Keytype& key) {
		int hashvalue=hash(key)%TABLESIZE;
		if(thelist[hashvalue]==NULL) return false; // Key didn't exist.
			List<pair<Keytype, Valuetype> >* p=thelist[hashvalue];
			while(p!=NULL) { // Search all list elements.
				if(p->thevalue.first==key) break; // Specified key found.
				p=p->next;
			}
			if(p==NULL) return false; // Specified key not found.
			link(p->prev, p->next); // Link the list around p in preparation for removal of p.
			if(p==thelist[hashvalue]) {
				thelist[hashvalue]=p->next; // link() function above doesn't take care of this special case.
			}
			delete p;
			return true;
	}

	template <typename Keytype, typename Valuetype>
	inline bool HashTable<Keytype, Valuetype>::changeValue(const Keytype& key, const Valuetype& value) {
		int hashvalue=hash(key)%TABLESIZE;
		if(thelist[hashvalue]==NULL) return false; // Key didn't exist.
			List<pair<Keytype, Valuetype> >* p=thelist[hashvalue];
			while(p!=NULL) { // Search all list elements.
				if(p->thevalue.first==key) break; // Specified key found.
				p=p->next;
			}
			if(p==NULL) return false; // Specified key not found.
			p->thevalue.second=value;
			return true;
	}

	template <typename Keytype, typename Valuetype>
	inline Valuetype* const HashTable<Keytype, Valuetype>::getValue(const Keytype& key) {
		int hashvalue=hash(key)%TABLESIZE;
		if(thelist[hashvalue]==NULL) return NULL; // Key didn't exist.
			List<pair<Keytype, Valuetype> >* p=thelist[hashvalue];
			while(p!=NULL) { // Search all list elements.
				if(p->thevalue.first==key) break; // Specified key found.
				p=p->next;
			}
			if(p==NULL) return NULL; // Specified key not found.
			return &(p->thevalue.second);
	}

	inline int hash(int arg) {return arg;};
	inline int hash(const char* arg) {
		if(arg==NULL) return 0;
		int retval=0;
		char* p=const_cast<char*>(arg);
		while(*p!='\0') { // Simple checksum.
			retval+=*p;
			p++;
		}
		return retval;
	}
};

#endif
