#include "FontReader.hpp"

#include <stdlib.h>
#include <stdio.h>

const int FontReader::OUTPUT_LEVEL=1;

void FontReader::readFont(PdfFont *font) {

	if(font == NULL)
		printerr("Error - null font\n");

	if(font->dataLength < 1) 
		printerr("Error - no data in font object for reading\n");

	char *data = font->data;
	short *ptr = (short*) data;
	int i;
	char *temp, *fontName;

	char encoding[30];
	unsigned char* cmapIndex;
	unsigned long* widths;

	unsigned long scalerType;
	unsigned long numAdvanceWidths;
	unsigned short numTables, searchRange, entrySelector, rangeShift;
	unsigned short unitsPerEM;

	IndexEntry *tableIndex;
	IndexEntry *table;

	// printf("Font length: %d\n", font->dataLength);
	if(font->dataLength < 12)
		printerr("Error - Font file too short to contain main header\n");

	// Header
	scalerType = pdfc_ntohl( *((long*)ptr));
	ptr += 2;
	numTables = pdfc_ntohs( *ptr++ );
	searchRange = pdfc_ntohs( *ptr++ );
	entrySelector = pdfc_ntohs( *ptr++ );
	rangeShift = pdfc_ntohs( *ptr++ );

	// Index Table
	tableIndex = new IndexEntry[numTables];
	long *tableData = (long*) (data+12);

	if(font->dataLength < (12+( ((unsigned long)numTables)*4)))
		printerr("Error - Font file too short to contain main table\n");

	for(i=0; i<numTables; i++) {
		memcpy(tableIndex[i].name, (char*) (tableData+4*i), 4);
		tableIndex[i].name[4] = 0;
		tableIndex[i].offset = pdfc_ntohl( tableData[4*i+2] );
		tableIndex[i].length = pdfc_ntohl( tableData[4*i+3] );
	}

	if(OUTPUT_LEVEL > 3) {
		printf("Scaler Type: %lx\n", scalerType);
		printf("NumTables: %d\n", numTables);
		printf("SearchRange: %d\n", searchRange);
		printf("EntrySelector: %d\n", entrySelector);
		printf("RangeShift: %d\n", rangeShift);
		for(i=0; i<numTables; i++) {
			printf("Tag: %s", tableIndex[i].name);
			printf(" Offset: %08lx", tableIndex[i].offset);
			printf(" Table Length: %ld\n", tableIndex[i].length);
		}
		printf("\n");
	}

	// Read Head table
	table = getIndex("head", tableIndex, numTables, font->dataLength);
	if(table == NULL) 
		printerr("Error - could not find head table\n");
	temp = data + table->offset;
	unitsPerEM = readHead(temp, font->fontBBox);

	// Read Name table
	table = getIndex("name", tableIndex, numTables, font->dataLength);
	if(table == NULL) 
		printerr("Error - could not find name table\n");
	temp = data + table->offset;
	fontName = readName(temp);

	// Read CMap table
	table = getIndex("cmap", tableIndex, numTables, font->dataLength);
	if(table == NULL) 
		printerr("Error - could not find cmap table\n");
	temp = data + table->offset;
	cmapIndex = readCmap(temp, encoding);
	
	// Read HHea table
	table = getIndex("hhea", tableIndex, numTables, font->dataLength);
	if(table == NULL) 
		printerr("Error - could not find hhea table\n");
	temp = data + table->offset;
	numAdvanceWidths = readHHea(temp);
	
	// Read Hmtx table
	table = getIndex("hmtx", tableIndex, numTables, font->dataLength);
	if(table == NULL) 
		printerr("Error - could not find hmtx table\n");
	temp = data + table->offset;
	widths = readHmtx(temp, numAdvanceWidths, unitsPerEM, cmapIndex);

	font->subType = ("TrueType");
	font->fontName = (fontName);
	font->firstChar = 0;
	font->lastChar = 255;
	font->numWidths = 256;

	if(font->widths != NULL) {
		delete [] font->widths; 
	}

	font->widths = (long*) widths;
	font->encoding = encoding;

	if(OUTPUT_LEVEL > 1) {
		printf("Font info set\n");
		printf("SubType: %s\n", font->subType.nullterm() );
		printf("BaseFont: %s\n", font->fontName.nullterm() );
		printf("FirstChar: %d\n", font->firstChar);
		printf("LastChar: %d\n", font->lastChar);
		printf("NumWidths: %d\n", font->numWidths);
		printf("Encoding: %s\n", font->encoding.nullterm() );
	}

	if(OUTPUT_LEVEL > 2) {
		for(i=0; i<256; i++) {
			printf("Width %d: %ld\n",i, widths[i]);
		}
	}

	delete [] fontName;
	delete [] cmapIndex;
	delete [] tableIndex;
}

IndexEntry* FontReader::getIndex(char *name, IndexEntry *list, int numEntries, unsigned long fSize) {
	int i;

	for(i=0; i<numEntries; i++) {
		if(strcmp(list[i].name, name) == 0) {
			if(list[i].offset < 0)
				printerr("Error - illeagal %s table address\n", name);
			if(list[i].length < 0)
				printerr("Error - invalid %s table size\n", name);
			if(list[i].offset > fSize)
				printerr("Error - %s table address out of range\n", name);
			if(list[i].offset+list[i].length > fSize)
				printerr("Error - %s table extends out of range\n", name);

			return list+i;
		}
	}

	return NULL;
}

unsigned short FontReader::readHead(char *data, double *bbox) {
	unsigned long version, fontRevision;
	short *ptr = (short*) data;

	if(OUTPUT_LEVEL > 2) printf("Reading Head\n");

	version = pdfc_ntohl( *((long*) ptr) );
	ptr += 2;
	fontRevision = pdfc_ntohl( *((long*) ptr) );
	ptr += 2;

	unsigned long checkSumAdjustment, magicNumber;
	unsigned short flags, unitsPerEM;

	checkSumAdjustment = pdfc_ntohl( *((long*) ptr) );
	ptr += 2;
	magicNumber = pdfc_ntohl( *((long*) ptr) );
	ptr += 2;
	flags = pdfc_ntohs( *ptr++ );
	unitsPerEM = pdfc_ntohs( *ptr++ );

	ptr += 4; // "Creation date".
	ptr += 4; // "Last modified".

	bbox[0] = ((double) ((short)pdfc_ntohs( *ptr++ ))) * 1000 /unitsPerEM;
	bbox[1] = ((double) ((short)pdfc_ntohs( *ptr++ ))) * 1000 /unitsPerEM;
	bbox[2] = ((double) ((short)pdfc_ntohs( *ptr++ ))) * 1000 /unitsPerEM;
	bbox[3] = ((double) ((short)pdfc_ntohs( *ptr++ ))) * 1000 /unitsPerEM;

	unsigned short macStyle, lowestRecPPEM;
	short fontDirectionHint, indexToLocFormat, glyphDataFormat;

	macStyle = pdfc_ntohs( *ptr++ );
	lowestRecPPEM = pdfc_ntohs( *ptr++ );
	fontDirectionHint = pdfc_ntohs( *ptr++ );
	indexToLocFormat = pdfc_ntohs( *ptr++ );
	glyphDataFormat = pdfc_ntohs( *ptr++ );

	if(OUTPUT_LEVEL > 3) {
		printf("Version: %ld.%ld\n", version >> 16, version & 0xffff);
		printf("FontRevision: %ld.%ld\n", fontRevision >> 16, fontRevision & 0xffff);
		printf("CheckSumAdjustment: 0x%lX\n", checkSumAdjustment);
		printf("MagicNumber: 0x%lX\n", magicNumber);
		printf("Flags: 0x%04X\n", flags);
		printf("UnitsPerEM: %d\n", unitsPerEM);
		printf("CreationDate: \n" );
		printf("LastModified: \n" );
		printf("MinX: %.0f\n", bbox[0]);
		printf("MinY: %.0f\n", bbox[1]);
		printf("MaxX: %.0f\n", bbox[2]);
		printf("MaxY: %.0f\n", bbox[3]);
		printf("MacStyle: 0x%04x\n", macStyle);
		printf("lowestRecPPEM: %d\n", lowestRecPPEM);
		printf("FontDirectionHint: %d\n", fontDirectionHint);
		printf("IndexToLocFormat: %d\n", indexToLocFormat);
		printf("GlyphDataFormat: %d\n", glyphDataFormat);
		printf("\n");
	}

	return unitsPerEM;
}

char* FontReader::readName(char *data) {
	unsigned short *ptr = (unsigned short*) data;
	char *offsetPtr;
	char *str;

	unsigned short format, count, stringOffset;
	int i;

	if(OUTPUT_LEVEL > 2) printf("Reading Name\n");

	format = pdfc_ntohs( *ptr++ );
	count = pdfc_ntohs( *ptr++ );
	stringOffset = pdfc_ntohs( *ptr++ );
	offsetPtr = ((char*) data) + stringOffset;

	NameEntry *names = new NameEntry[count];

	for(i=0; i<count; i++) {
		names[i].platformId = pdfc_ntohs( *ptr++ );
		names[i].platformSpecificId = pdfc_ntohs( *ptr++ );
		names[i].languageId = pdfc_ntohs( *ptr++ );
		names[i].nameId = pdfc_ntohs( *ptr++ );
		names[i].length = pdfc_ntohs( *ptr++ );
		names[i].offset = pdfc_ntohs( *ptr++ );
	}

	if(OUTPUT_LEVEL > 3) {
		printf("Format: %d\n", format);
		printf("Count: %d\n", count);
		printf("StringOffset: %d\n", stringOffset);
		for(i=0; i<count; i++) {
			printf(" PlatformId: %d\n", names[i].platformId);
			printf(" PlatformSpecificId: %d\n", names[i].platformSpecificId);
			printf(" LanguageId: %d\n", names[i].languageId);
			printf(" NameId: %d\n", names[i].nameId);
			printf(" Length: %d\n", names[i].length);
			printf(" Offset: %d\n", names[i].offset);
			printf("\n");
		}
		printf("\n");
	}

	for(i=0; i<count; i++) {
		if(names[i].platformId == 1 && 
		   names[i].platformSpecificId == 0 && 
		   names[i].languageId == 0 && 
		   names[i].nameId == 6) {

			str = new char[names[i].length+1];
			memcpy(str, offsetPtr + names[i].offset, names[i].length);
			str[names[i].length] = 0;

			if(OUTPUT_LEVEL > 2) printf("Found MacRoman name: %s\n", str);
			delete [] names;
			return str;
		}
		else if(names[i].platformId == 3 && 
		        (names[i].platformSpecificId == 0 || names[i].platformSpecificId == 1) && 
		        names[i].languageId == 0x409 && 
		        names[i].nameId == 6) {

			int j;
			str = new char[(names[i].length/2)+1];
			short *namePtr = (short*) (offsetPtr + names[i].offset);
			for(j=0; j<names[i].length/2; j++) {
				str[j] = (char) pdfc_ntohs(namePtr[j]);
			}
			str[j] = 0;

			if(OUTPUT_LEVEL > 2) printf("Found Windows name: %s\n", str);
			delete [] names;
			return str;
		}
	}

	delete [] names;
	return NULL;
}

unsigned long FontReader::readHHea(char *data) {
	unsigned long version;
	short ascent, descent, linegap;
	unsigned short advanceWidthMax;
	short minLeftSideBearing, minRightSideBearing;
	short xMaxExtent, caretSlopeRise, caretSlopeRun, caretOffset;
	short metricDataFormat;
	unsigned short numOfLongHorMetrics;

	short *ptr = (short*) data;

	if(OUTPUT_LEVEL > 2) printf("Reading HHea\n");

	version = pdfc_ntohl( *((long*) ptr) );
	ptr += 2;

	ascent = pdfc_ntohs( *ptr++ );
	descent = pdfc_ntohs( *ptr++ );
	linegap = pdfc_ntohs( *ptr++ );
	advanceWidthMax = pdfc_ntohs( *ptr++ );
	minLeftSideBearing = pdfc_ntohs( *ptr++ );
	minRightSideBearing = pdfc_ntohs( *ptr++ );
	xMaxExtent = pdfc_ntohs( *ptr++ );
	caretSlopeRise = pdfc_ntohs( *ptr++ );
	caretSlopeRun = pdfc_ntohs( *ptr++ );
	caretOffset = pdfc_ntohs( *ptr++ );
	ptr += 4; // reserved bytes

	metricDataFormat = pdfc_ntohs( *ptr++ );
	numOfLongHorMetrics = pdfc_ntohs( *ptr++ );

	if(OUTPUT_LEVEL > 3) {
		printf("Version: %ld.%ld\n", version >> 16, version & 0xffff);
		printf("Ascent: %d\n", ascent);
		printf("Descent: %d\n", descent);
		printf("Linegap: %d\n", linegap);
		printf("AdvanceWidthMax: %d\n", advanceWidthMax);
		printf("MinLeftSideBearing: %d\n", minLeftSideBearing);
		printf("MinRightSideBearing: %d\n", minRightSideBearing);
		printf("XMaxExtent: %d\n", xMaxExtent);
		printf("CaretSlopeRise: %d\n", caretSlopeRise);
		printf("CaretSlopeRun: %d\n", caretSlopeRun);
		printf("CaretOffset: %d\n", caretOffset);
		printf("MetricDataFormat: %d\n", metricDataFormat);
		printf("NumOfLongHorMetrics: %d\n", numOfLongHorMetrics);
		printf("\n");
	}

	return numOfLongHorMetrics;
}

unsigned long* FontReader::readHmtx(char *data, unsigned long numEntries, unsigned short unitsPerEM, unsigned char* cmap) {
	unsigned short *list = (unsigned short*) data;
	unsigned short *advance = new unsigned short[numEntries];
	short *bearingX = new short[numEntries];
	uint i;

	unsigned long *widths = new unsigned long[256];

	if(OUTPUT_LEVEL > 2) printf("Reading Hmtx\n");

	for(i=0; i<numEntries; i++) {
		advance[i] = pdfc_ntohs(list[2*i]) * 1000 / unitsPerEM;
		bearingX[i] = pdfc_ntohs(list[2*i+1]);
	}

	if(OUTPUT_LEVEL > 3) {
		for(i=0; i<numEntries; i++) {
			printf("%3d Advance: %4d BearingX: %4d\n", i, advance[i], bearingX[i]);
		}
		printf("\n");
	}

	for(i=0; i<256; i++) {
		widths[i] = advance[cmap[i]];
	}

	delete [] advance;
	delete [] bearingX;

	return widths;
}

unsigned char* FontReader::readCmap(char *data, char* encoding) {
	unsigned short version, numSubTables;
	CMapEntry *cmapTables;
	int i;
	short *ptr = (short*) data;

	unsigned char *index = NULL;

	if(OUTPUT_LEVEL > 2) printf("Reading CMap\n");

	version = pdfc_ntohs( *ptr++ );

	numSubTables = pdfc_ntohs( *ptr++ );

	cmapTables = new CMapEntry[numSubTables];

	for(i=0; i<numSubTables; i++) {
		cmapTables[i].platformId = pdfc_ntohs( *ptr++ );
		cmapTables[i].platformSpecificId = pdfc_ntohs( *ptr++ );
		cmapTables[i].offset = pdfc_ntohl( *((long*) ptr) );
		ptr += 2;
	}
	
	*encoding = 0;

	for(i=0; i<numSubTables; i++) {
		if(cmapTables[i].platformId == 1 && cmapTables[i].platformSpecificId == 0) { 
			index = readMacRomanCmapTable( (char*)data + cmapTables[i].offset);
			if(index != NULL) {
				strcpy(encoding, "MacRomanEncoding");
				break;
			}
		}
		else if(cmapTables[i].platformId == 3 && cmapTables[i].platformSpecificId == 1) {
			index = readWinAnsiCmapTable( (char*)data + cmapTables[i].offset);
			if(index != NULL) {
				strcpy(encoding, "WinAnsiEncoding");
				break;
			}
		}
	}

	if(pdfc_strlen(encoding) < 1)
		printerr("Error - could not find suitable cmap encoding table\n");

	if(OUTPUT_LEVEL > 3) {
		printf("Version: %d\n", version);
		printf("NumSubTables: %d\n", numSubTables);
		for(i=0; i<numSubTables; i++) {
			printf(" PlatformId: %d", cmapTables[i].platformId);
			printf(" PlatformSpecificId: %d", cmapTables[i].platformSpecificId);
			printf(" Offset: %08lx\n", cmapTables[i].offset);
		}
		printf("\n");
	}

	delete [] cmapTables;

	return index;
}

unsigned char* FontReader::readMacRomanCmapTable(char *cmapData ) {
	unsigned short format, length, language;
	int i;
	unsigned char *index;
	unsigned short *ptr = (unsigned short*) cmapData;
	unsigned char *cptr;

	if(OUTPUT_LEVEL > 2) printf("Reading MacRoman CMap Subtable\n");

	format = pdfc_ntohs( *ptr++ );
	if(format != 0) // Mac CMap formats other that 0 are currently unsupported
		return NULL;

	length = pdfc_ntohs( *ptr++ );
	language = pdfc_ntohs( *ptr++ );
	// if(language != 0)
		// return NULL;

	index = new unsigned char[256];

	cptr = (unsigned char*) ptr;
	for(i=0; i<256; i++) {
		index[i] = cptr[i];
	}

	if(OUTPUT_LEVEL > 3) {
		printf("Format: %d\n", format);
		printf("Length: %d\n", length);
		printf("Language: %d\n", language);
		for(i=0; i<256; i++) {
			printf(" Character Code: %x Glyph Index: %d\n", i, index[i]);
		}
		printf("\n");
	}

	return index;
}

unsigned char* FontReader::readWinAnsiCmapTable(char *cmapData ) {
	unsigned short format, length, language;
	unsigned short segCount, searchRange, entrySelector, rangeShift;
	int i, j;

	unsigned char *index;

	unsigned short *ptr = (unsigned short*) cmapData;

	if(OUTPUT_LEVEL > 2) printf("Reading WinAnsi CMap Subtable\n");

	format = pdfc_ntohs( *ptr++ );
	if(format != 4) // Windows CMap formats other that 4 are currently unsupported
		return NULL;

	length = pdfc_ntohs( *ptr++ );
	language = pdfc_ntohs( *ptr++ );
	// if(language != 0)
		// return NULL;

	index = new unsigned char[256];

	segCount = pdfc_ntohs( *ptr++ ) / 2;
	searchRange = pdfc_ntohs( *ptr++ );
	entrySelector = pdfc_ntohs( *ptr++ );
	rangeShift = pdfc_ntohs( *ptr++ );

	unsigned short *endCode, *startCode;
	short *idDelta;
	unsigned short *idRangeOffset;

	long numEntries;

	numEntries = (length / 2); // All values are 16bit values;
	numEntries = numEntries - 8;
	numEntries = numEntries - (4*segCount);

	endCode = ptr;
	ptr += segCount;

	if(*ptr++ != 0)
		printerr("Error - reserved pad missing\n");

	startCode = ptr;
	ptr += segCount;

	idDelta = (short*) ptr;
	ptr += segCount;

	idRangeOffset = ptr;
	ptr += segCount;

	for(i=0; i<segCount; i++) {
		startCode[i] = pdfc_ntohs(startCode[i]);
		endCode[i] = pdfc_ntohs(endCode[i]);
		idDelta[i] = pdfc_ntohs(idDelta[i]);
		idRangeOffset[i] = pdfc_ntohs(idRangeOffset[i]);
	}

	unsigned short *glyphIndex = idRangeOffset + segCount;
	unsigned short val;

	for(i=0; i<numEntries; i++) {
		glyphIndex[i] = pdfc_ntohs(glyphIndex[i]);
	}

	// Create cmap index
	j = 0;
	for(i=0; i<256; i++) {
		if(i < startCode[j] || j >= segCount) {
			index[i] = 0;
		}
		else {
			while(j < segCount && i > endCode[j]) {
				j++;
			}

			if(i < startCode[j] || j >= segCount) {
				index[i] = 0;
			}
			else {
				if(idRangeOffset[j] == 0) {
					index[i] = i + idDelta[j];
				}
				else {
					index[i] = idRangeOffset[idRangeOffset[j]/2 + (i-startCode[j]) + j];
				}
			}
		}
	}

	if(OUTPUT_LEVEL > 3) {
		printf("Format: %d\n", format);
		printf("Length: %d\n", length);
		printf("Language: %d\n", language);
		printf("SegCount: %d\n", segCount);
		printf("SearchRange: %d\n", searchRange);
		printf("EntrySelector: %d\n", entrySelector);
		printf("RangeShift: %d\n", rangeShift);

		printf("%ld glyph entries\n", numEntries);
		for(i=0; i<segCount; i++) {
			printf(" %3d s: %04x e: %04x d: %5d ro: %5d\n", i, startCode[i], endCode[i], idDelta[i], idRangeOffset[i]);
		}

		for(i=0; i<numEntries; i++) {
			printf(" %3d Glyph Code: %d\n", i, glyphIndex[i]);
		}

		for(i=0; i<segCount-1; i++) {
			for(j=startCode[i]; j<=endCode[i]; j++) {
				if(idRangeOffset[i] == 0) {
					printf("Character Code: %4x Glyph Index: %3d\n", j, j + idDelta[i]);
				}
				else {
					val = idRangeOffset[idRangeOffset[i]/2 + (j-startCode[i]) + i];
					printf("Character Code: %4x Glyph Index: %3d\n", j, val);
				}
			}
		}
		printf("\n");
	}

	return index;
}

