#include "ObjParser.hpp"

#include <stdlib.h>
#include <stdio.h>

#include "FontReader.hpp"
#include "pdfc_basicobjects.h"
#include "PDFTokenizer.hpp"

const int ObjParser::OUTPUT_LEVEL=1;
const int ObjParser::DECODE_IMAGES=0;

//-----------------------------------------------

PdfRoot* ObjParser::parse(PObj *root) {
	PDF_dictionary *catalog;
	PObj *obj;
	PdfRoot *pdfRoot = new PdfRoot();

	if(OUTPUT_LEVEL > 2) printf("Starting ObjParser::parse\n");
	if(OUTPUT_LEVEL > 1) printf("Parsing Document structure\n");

	// Verify root node
	if(root == NULL) 
		printerr("Error - root object not found\n");
	if(root->get_type() != t_root) 
		printerr("Error - root object not of proper type\n");

	PDF_root* rootcast=(PDF_root*) root;
	if(rootcast==NULL) printerr("Error - root object not of proper type\n");
	
	// Read Document Info if any exists
	if( ((PDF_root *)root)->trailer != NULL && 
	    ((PDF_root *)root)->trailer->dictionary != NULL) {
			PDF_dictionary *info;
			PDF_string *item;

			obj = getValidItem(root, ((PDF_root *)root)->trailer->dictionary, "Encrypt", t_dictionary);
			if(obj != NULL)
				printerr("Encrypted documents are not supported at this time.\n");

			info = (PDF_dictionary*) getValidItem(root, ((PDF_root *)root)->trailer->dictionary, "Info", t_dictionary);
			if(info != NULL) {
				item = (PDF_string*) getValidItem(root, info, "Author", t_string);
				if(item != NULL) {
					pdfRoot->author = item->value;
					// printf("Author: %s\n", pdfRoot->author.nullterm());
				}

				item = (PDF_string*) getValidItem(root, info, "CreationDate", t_string);
				if(item != NULL) {
					pdfRoot->creationDate = item->value;
					// printf("CreationDate: %s\n", pdfRoot->creationDate.nullterm());
				}

				item = (PDF_string*) getValidItem(root, info, "ModDate", t_string);
				if(item != NULL) {
					pdfRoot->modDate = item->value;
					// printf("Previous ModDate: %s\n", pdfRoot->modDate.nullterm());
				}

				item = (PDF_string*) getValidItem(root, info, "Creator", t_string);
				if(item != NULL) {
					pdfRoot->creator = item->value;
					// printf("Creator: %s\n", pdfRoot->creator.nullterm());
				}

				item = (PDF_string*) getValidItem(root, info, "Producer", t_string);
				if(item != NULL) {
					pdfRoot->producer = item->value;
					// printf("Previous Producer: %s\n", pdfRoot->producer.nullterm());
				}

				item = (PDF_string*) getValidItem(root, info, "Title", t_string);
				if(item != NULL) {
					pdfRoot->title = item->value;
					// printf("Title: %s\n", pdfRoot->title.nullterm());
				}

				item = (PDF_string*) getValidItem(root, info, "Subject", t_string);
				if(item != NULL) {
					pdfRoot->subject = item->value;
					// printf("Subject: %s\n", pdfRoot->subject.nullterm());
				}

				item = (PDF_string*) getValidItem(root, info, "Keywords", t_string);
				if(item != NULL) {
					pdfRoot->keywords = item->value;
					// printf("Keywords: %s\n", pdfRoot->keywords.nullterm());
				}
			}
	} 

	// Find Main Catalog
	catalog = (PDF_dictionary*) validateObj(root, (rootcast)->rootNode, t_dictionary);
	if(catalog == NULL)
		printerr("Error - main catalog not found\n");
	
	obj = getValidItem(root, catalog, "Type", t_name);
	if(obj == NULL)
		printerr("Error - Could not find catalog type attribute\n");
	if( ((PDF_name*)obj)->value != "Catalog" ) 
		printerr("Error - Illegal object found while looking for catalogPages objectn");

	if(OUTPUT_LEVEL > 1) printf("Catalog found\n");

	// Check for Document display cues
	obj = getValidItem(root, catalog, "PageMode", t_name);
	if(obj != NULL) {
		if( ((PDF_name*)obj)->value == "UseOutlines")
			pdfRoot->showBookmarks = true;
	}

	// Find Main Pages object
	obj = catalog->getValueForKey("Pages");
	if(OUTPUT_LEVEL > 1) printf("Descending into pages structure\n");
	parsePages(obj, root, pdfRoot, NULL, NULL);

	// Deal with link references
	PdfLinkList *ptr = pdfRoot->linkList;
	PdfLink *link;
	PdfPage *page;

	while(ptr != NULL) {
		link = ptr->link;
		// printf("Link: %s\n", link->destKey.nullterm());

		page = (PdfPage*) pdfRoot->pageHash.get(link->destKey);
		if(page == NULL) {
			printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "invalid link id\n");
			//FIX - remove link
		}
		link->destination = page;

		ptr = ptr->next;
	}

	if(pdfRoot->linkList != NULL) {
		delete pdfRoot->linkList;
		pdfRoot->linkList = NULL;
	}

	// Get Bookmarks
	PObj *bookmarks = getValidItem(root, catalog, "Outlines", t_dictionary);
	if(bookmarks != NULL) {
		obj = getValidItem(root, bookmarks, "First", t_dictionary);
		while(obj != NULL) {
			pdfRoot->addBookmark(parseBookmark(obj, root, pdfRoot));
			obj = getValidItem(root, obj, "Next", t_dictionary);
		}
	}

	if(OUTPUT_LEVEL > 2) printf("Ending ObjParser::parse\n");
	return pdfRoot;
}

//-----------------------------------------------

PdfBookmark* ObjParser::parseBookmark(PObj *bookmark, PObj *root, PdfRoot *pdfRoot) {
	PObj *obj;
	PdfPage *page;
	PdfBookmark *pdfBookmark = new PdfBookmark();
	PdfBookmark *child;
	int childCount = 0;
	int numChildren = 0;

	// Verify bookmark exists
	if(bookmark == NULL) {
		printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - bookmark missing\n");
		return NULL;
	}

	// printf("Found bookmark\n");

	// Get bookmark parameters
	obj = getValidItem(root, bookmark, "Title", t_string);
	if(obj == NULL) {
		printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - bookmark has a missing title\n");
		return NULL;
	}
	pdfBookmark->title = ((PDF_string*)obj)->value;
	// printf(" Title: %s\n", pdfBookmark->title.nullterm());	

	obj = getValidItem(root, bookmark, "Count", t_integer);
	if(obj != NULL) {
		pdfBookmark->count = ((PDF_int*)obj)->value;
		// printf(" Count: %d\n", pdfBookmark->count);
	}

	// Get bookmark destination
	obj = getValidItem(root, bookmark, "Dest", t_array);
	if(obj == NULL)  {
		printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Missing destination for bookmark\n");
		return NULL;
	}
	else {
		PObjList *element;
		char buffer[32];
		// printf(" Found destination\n");

		if(obj->get_type() == t_array) {
			// printf("Found bookmark Dest array\n");
			element=obj->subobjects;
			if(element == NULL) {
				printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - empty bookmark destination\n");
				return NULL;
			}

			obj = element->theobject;
			obj = validateObj(root, obj, t_dictionary);
			if(obj == NULL) {
				printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - invalid bookmark destination\n");
				return NULL;
			}

			sprintf(buffer, "%d",  obj->id);
			// printf(" destKey: %s\n", buffer);
			pdfBookmark->destKey = buffer;

			page = (PdfPage*) pdfRoot->pageHash.get(pdfBookmark->destKey);
			if(page == NULL) {
				printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - invalid bookmark destination id\n");
				return NULL;
			}

			pdfBookmark->destination = page;
		}
		else if(obj->get_type() == t_name) {
			printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Bookmark named destinations not supported\n");
			return NULL;
		}
		else {
			printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - unknown type of Dest object found in bookmark %s\n", obj->get_printable_type());
			return NULL;
		}
	}

	// Get child bookmarks
	obj = getValidItem(root, bookmark, "First", t_dictionary);
	while(obj != NULL) {
		child = parseBookmark(obj, root, pdfRoot);
		if(child != NULL) {
			pdfBookmark->addBookmark(child);
			if(child->count < 0)
				childCount += -(child->count);
			else
				childCount += child->count;

			numChildren++;
		}
		obj = getValidItem(root, obj, "Next", t_dictionary);
	}

	// Determine if children are visable by default
	if(pdfBookmark->count == childCount)
		pdfBookmark->isOpen = false;
	else if( -(pdfBookmark->count) == (childCount + numChildren))
		pdfBookmark->isOpen = false;
	else if(pdfBookmark->count == (childCount + numChildren))
		pdfBookmark->isOpen = true;
	else {
		printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - faulty bookmark count\n");
	}

	return pdfBookmark;
}

//-----------------------------------------------

void ObjParser::parseAnnots(PObj *obj, PObj *root, PdfPage *pdfPage, PdfRoot *pdfRoot) {
	PdfAnnot *annot = NULL;
	PObj *type, *rectList, *rectVal, *val;
	PObjList *list, *element;
	int i;
	char buffer[256];

	if(OUTPUT_LEVEL > 2) printf("Starting ObjParser::parseAnnots\n");

	obj = dereference(root, obj);
	if(obj== NULL)
		printerr("Error empty annotation object\n");

	if(obj->get_type() == t_array) {
		if(OUTPUT_LEVEL > 1) printf("Annotation Array found\n");
		list = obj->subobjects;
		while(list != NULL) {
			parseAnnots(list->theobject, root, pdfPage, pdfRoot);
			list = list->next;
		}
	}
	else if(obj->get_type() == t_dictionary) {
		if(OUTPUT_LEVEL > 1) printf("Annotation Dictionary found\n");

		val = getValidItem(root, obj, "Type", t_name);
		if(val == NULL) {
			printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - Type not found in annotation\n");
			return;
		}
		if( ((PDF_name*)val)->value != "Annot") {
			printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - annotation Type must be \"Annot\"\n");
			return;
		}

		type = getValidItem(root, obj, "Subtype", t_name);
		if(type == NULL) {
			printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - SubType not found in annotation\n");
			return;
		}

		if(((PDF_name*)type)->value == "Text") {
			annot = new PdfTextAnnot();

			val = getValidItem(root, obj, "Contents", t_string);
			if(val == NULL) {
				printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - Contents not found in text_annotation\n");
				return;
			}
			((PdfTextAnnot*)annot)->contents = ((PDF_name*)val)->value;

			val = getValidItem(root, obj, "Open", t_boolean);
			if(val != NULL)
				((PdfTextAnnot*)annot)->isOpen = ((PDF_boolean*)val)->value;
		}
		else if(((PDF_name*)type)->value == "Link") {
			annot = new PdfLink(pdfRoot);

			val = ((PDF_dictionary*) obj)->getValueForKey("Dest");
			if(val == NULL) {
				printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - Dest not found in link\n");
				return;
			}

			if(val->get_type() == t_array) {
				// printf("Found link Dest array\n");
				element=val->subobjects;
				if(element == NULL) {
					printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - empty annotation link destination\n");
					return;
				}

				val = element->theobject;
				val = validateObj(root, val, t_dictionary);
				if(val == NULL) {
					printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - invalid annotation link destination\n");
					return;
				}

				sprintf(buffer, "%d",  val->id);
				((PdfLink*)annot)->destKey = buffer;
				// element = element->next;
				// Deal with rest of parameters - FIX

				pdfRoot->addLink( ((PdfLink*)annot) );
			}
			else if(val->get_type() == t_name) {
				printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Named Link destinations not supported: %s\n", ((PDF_name*)val)->value.nullterm()); // FIX
				return;
			}
			else if(val->get_type() == t_string) {
				printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Named link destinations not supported: %s\n", ((PDF_string*)val)->value.nullterm()); // FIX
				return;
			}
			else {
				printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - unknown type of Dest object found in link: %s\n", val->get_printable_type());
				return;
			}
		}

		rectList = getValidItem(root, obj, "Rect", t_array);
		if(rectList == NULL) {
			printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Could not find Rect in annotation\n");
			return;
		}

		list = rectList->subobjects;
		for(i=0; i<4; i++) {
			if(list == NULL) {
				printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - annotation rect missing elements\n");
				return;
			}

			rectVal = validateObj(root, list->theobject, t_numeric);
			if(rectVal == NULL) {
				printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - rect annotation elements must be numbers\n");
				return;
			}

			annot->rect[i] = ((PDF_numeric*)rectVal)->getNumericValue();
			list = list->next;
		}
		if(list != NULL) {
			printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - annotation rect has extra elements\n");
			return;
		}

		pdfPage->addAnnotation(annot);
	}

	if(OUTPUT_LEVEL > 2) printf("Ending ObjParser::parseAnnots\n");
}

//-----------------------------------------------

void ObjParser::parsePages(PObj *pages, PObj *root, PdfRoot *pdfRoot, double *inheritedMediaBox, ResourceRefList *inheritedResources) {
	PObj *obj, *contentsRef, *resourcesDict, *mediaBoxObj;
	PObj *annots;
	PObjList *list;
	int i;
	double *mediaBox = NULL;
	ResourceRefList *resources = NULL;

	if(OUTPUT_LEVEL > 2) printf("Starting ObjParser::parsePages\n");

	pages = dereference(root, pages);
	if(pages == NULL) {
		printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error empty pages object\n");
		return;
	}

	if(pages->get_type() == t_array) {
		if(OUTPUT_LEVEL > 1) printf("Pages Array found\n");
		list = pages->subobjects;
		while(list != NULL) {
			parsePages(list->theobject, root, pdfRoot, inheritedMediaBox, inheritedResources);
			list = list->next;
		}
	}
	else if(pages->get_type() == t_dictionary) {
		if(OUTPUT_LEVEL > 1) printf("Pages Object found\n");
	
		obj = getValidItem(root, pages, "Type", t_name);
		if(obj == NULL) {
			printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - Could not find pages type attribute\n");
			return;
		}

		if( ((PDF_name*)obj)->value == String("Pages") ) {
			if(OUTPUT_LEVEL > 1) printf("Found Pages dictionary\n");
			obj = ((PDF_dictionary*)pages)->getValueForKey("Kids");

			mediaBoxObj = getValidItem(root, pages, "MediaBox", t_array);
			mediaBox = parseMediaBox(mediaBoxObj, root);
			if(mediaBox != NULL)
				inheritedMediaBox = mediaBox;

			resourcesDict = getValidItem(root, pages, "Resources", t_dictionary);
			resources = parseResources(resourcesDict, root, pdfRoot);
			if(resources != NULL) {
				// printf("Found Pages resources\n");
				if(inheritedResources != NULL) {
					ResourceRefList *list, *ptr, *last;
					while(inheritedResources != NULL) {
						if(list == NULL) {
							list = new ResourceRefList( (PdfResourceRef*)inheritedResources->resourceRef->copy(true));
							ptr = list;
						}
						else {
							ptr->next = new ResourceRefList( (PdfResourceRef*)inheritedResources->resourceRef->copy(true));
						}
						last = ptr;
						ptr = ptr->next;
					}
					last->next = resources;
					resources = list;
				}
				inheritedResources = resources;
			}

			parsePages(obj, root, pdfRoot, inheritedMediaBox, inheritedResources);

			if(mediaBox != NULL)
				delete [] mediaBox;

			ResourceRefList *temp;
			while(resources != NULL) {
				temp = resources;
				resources = resources->next;
				delete temp->resourceRef;
				delete temp;
			}
		}
		else if( ((PDF_name*)obj)->value == String("Page") ) {

			// printf("Page PObj id: %d\n", pages->id);
			PdfPage *pdfPage = new PdfPage();
			pdfPage->id = pages->id;
			pdfRoot->addPage(pdfPage);

			if(OUTPUT_LEVEL > 1) printf("Found Page dictionary\n");
			resourcesDict = getValidItem(root, pages, "Resources", t_dictionary);
			resources = parseResources(resourcesDict, root, pdfRoot);

			// Add Resource Refs to pdfPage
			ResourceRefList *temp;
			while(inheritedResources != NULL) {
				// printf("Found inherited page resource\n");
				pdfPage->addResourceRef( (PdfResourceRef*) inheritedResources->resourceRef->copy(true));
				inheritedResources = inheritedResources->next;
			}
			while(resources != NULL) {
				// printf("Found page resource\n");
				pdfPage->addResourceRef(resources->resourceRef);

				temp = resources;
				resources = resources->next;
				delete temp;
			}

			mediaBoxObj = getValidItem(root, pages, "MediaBox", t_array);
			if(mediaBoxObj != NULL) 
				mediaBox = parseMediaBox(mediaBoxObj, root);

			if(mediaBox != NULL) {
				for(i=0; i<4; i++)
					pdfPage->mediaBox[i] = mediaBox[i];
				delete [] mediaBox;
			}
			else if(mediaBox == NULL && inheritedMediaBox != NULL) {
				for(i=0; i<4; i++)
					pdfPage->mediaBox[i] = inheritedMediaBox[i];
			}
			else {
				printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Could not find Page MediaBox\n");
			}

			annots = getValidItem(root, pages, "Annots", t_array);
			if(annots != NULL) {
				parseAnnots(annots, root, pdfPage, pdfRoot);
			}

			if(OUTPUT_LEVEL > 1) printf("\n");
			contentsRef = ((PDF_dictionary*)pages)->getValueForKey("Contents");
			parseContents(contentsRef, root, pdfPage);
		}
		else {
			printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - Illegal object found while looking for Pages object\n");
		}
	}
	else {
		printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - Illegal PObj found when looking for Pages reference - %s\n", pages->get_printable_type());
	}

	if(OUTPUT_LEVEL > 2) printf("Ending ObjParser::parsePages\n");
}

double* ObjParser::parseMediaBox(PObj *mediaBoxObj, PObj *root) {
	double *mediaBox = NULL;
	int i;
	PObjList *list;
	PObj *mediaBoxVal;

	if(mediaBoxObj == NULL)
		return NULL;

	mediaBox = new double[4];

	list = mediaBoxObj->subobjects;
	for(i=0; i<4; i++) {
		if(list == NULL) {
			printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - page MediaBox missing elements\n");
			delete [] mediaBox;
			return NULL;
		}
		else {
			mediaBoxVal = validateObj(root, list->theobject, t_numeric);
			if(mediaBoxVal == NULL) {
				printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - page mediabox elements must be numbers\n");
				delete [] mediaBox;
				return NULL;
			}
			else {
				mediaBox[i] = ((PDF_numeric*)mediaBoxVal)->getNumericValue();
			}
			list = list->next;
		}
	}
	if(i == 4 && list != NULL) {
		printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - page MediaBox has extra elements\n");
		delete [] mediaBox;
		return NULL;
	}

	return mediaBox;
}

//-----------------------------------------------

void ObjParser::parseContents(PObj *contentsRef, PObj *root, PdfPage *pdfPage) {
	PObj *contentsStream;
	PObj *streamDict;
	PObj *blobObj;
	PObjList *list;

	if(OUTPUT_LEVEL > 2) printf("Starting ObjParser::parseContents\n");

	contentsRef = dereference(root, contentsRef);

	if(contentsRef == NULL) {
		if(OUTPUT_LEVEL > 0) printf("Empty contents object\n");
		if(OUTPUT_LEVEL > 2) printf("Ending ObjParser::parseContents\n");
		return;
	}

	if(contentsRef->get_type() == t_array) {
		if(OUTPUT_LEVEL > 1) printf("Contents Array found\n");
		list = contentsRef->subobjects;
		while(list != NULL) {
			parseContents(list->theobject, root, pdfPage);
			list = list->next;
		}
		return;
	}
	
	if(OUTPUT_LEVEL > 1) printf("Found Contents Reference\n");

	contentsStream = validateObj(root, contentsRef, t_stream);
	if(contentsStream == NULL) {
		printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - invalid content stream type\n");
		return;
	}

	if(OUTPUT_LEVEL > 1) printf("Found Contents Stream\n");

	streamDict = ((PDF_stream*) contentsStream)->dictionary;
	streamDict = validateObj(root, streamDict, t_dictionary);
	if(streamDict == NULL) {
		printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - invalid content stream dictionary\n");
		return;
	}

	blobObj = ((PDF_stream*) contentsStream)->dataObject;
	blobObj = validateObj(root, blobObj, t_blob);
	if(blobObj==NULL) {
		printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - data missing from content stream\n");
		return;
	}

	if(OUTPUT_LEVEL > 1) printf("Content data found\n");

	parseFilter(pdfPage, streamDict, root);

	pdfPage->dataLength = ((PDF_blob*)blobObj)->length;
	if(pdfPage->dataLength > 0) { 
		pdfPage->data = new char[pdfPage->dataLength];
		memcpy(pdfPage->data, ((PDF_blob*)blobObj)->rawdata, pdfPage->dataLength);

		pdfPage->decode();
		if(pdfPage->filters != NULL) {
			printerr("Error - Unsupported Content encoding scheme - %s\n", pdfPage->filters->str.nullterm());
		}

		String temp( pdfPage->data, pdfPage->dataLength);
		delete [] pdfPage->data;
		pdfPage->data = NULL;
		pdfPage->dataLength = 0;

		pdfPage->addContent( parse_page_description_to_XML(temp));
	}

	if(OUTPUT_LEVEL > 2) printf("Ending ObjParser::parseContents\n");
}

//-----------------------------------------------

ResourceRefList* ObjParser::parseResources(PObj *resourcesDict, PObj *root, PdfRoot *pdfRoot) {
	String **resourceTypeList, **resourceType;
	PObj *resourceCatagory, *procSet;
	PObj *obj;
	String **keys, **keyPtr;
	PObj *item;
	PObjList *list;
	PdfResource *pdfResource = NULL;
	PdfResourceRef *pdfResourceRef;

	ResourceRefList *resources = NULL;
	ResourceRefList *refPtr = NULL;

	if(OUTPUT_LEVEL > 2) printf("Starting ObjParser::parseResources\n");

	resourcesDict = validateObj(root, resourcesDict, t_dictionary);
	if(resourcesDict == NULL) {
		return NULL;
	}

	if(OUTPUT_LEVEL > 1) printf("Found Page Resources Dictionary\n");

	resourceTypeList = ((PDF_dictionary*) resourcesDict)->getKeys();
	if(resourceTypeList == NULL)
		return NULL;

	resourceType = resourceTypeList;
	while(*resourceType != NULL) {
		if(**resourceType == NULL)
			continue;

		resourceCatagory = ((PDF_dictionary*) resourcesDict)->getValueForKey(**resourceType);
		resourceCatagory = dereference(root, resourceCatagory);

		if(resourceCatagory == NULL) {
			printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - Could not find Resource\n");
			continue;
		}
		
		if(OUTPUT_LEVEL > 1) printf("Resource Type: %s\n", (*resourceType)->nullterm());

		if(**resourceType == "ProcSet") {
			procSet = validateObj(root, resourceCatagory, t_array);
			if(procSet == NULL) {
				printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - invalid Resources ProcSet subobject\n");
				continue;
			}
			if(OUTPUT_LEVEL > 1) printf("Found Page Resource ProcSet Array\n");

			list = procSet->subobjects;
			while(list != NULL) {
				item = list->theobject;
				item = validateObj(root, item, t_name);
				if(item == NULL) {
					printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - name expected in ProcSet array\n");
					continue;
				}

				if(OUTPUT_LEVEL > 1) printf(" Found set: %s\n", ((PDF_name*) item)->value.nullterm());

				// FIX - use found item

				list = list->next;
			}
		}
		else {
			if(resourceCatagory->get_type() != t_dictionary) {
				printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - invalid Resources subobject - %s\n", resourceCatagory->get_printable_type());
				continue;
			}
	
			if(OUTPUT_LEVEL > 1) printf("Found Resource Dictionary\n");

			keys = ((PDF_dictionary*) resourceCatagory)->getKeys();
			keyPtr = keys;
			while(*keyPtr != NULL) {
				if(OUTPUT_LEVEL > 1) printf(" Resource: %s\n", (*keyPtr)->nullterm());

				pdfResource = NULL;

				obj = ((PDF_dictionary*) resourceCatagory)->getValueForKey(**keyPtr);
				pdfResource = getResource(obj, **resourceType, root, pdfRoot);

				if(obj == NULL) {
					printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - invalid named resource reference\n");
					continue;
				}

				// Get Pointer to resource based on the original object id
				// if this is the first reference to the resource, create
				// resource object
				pdfResourceRef = new PdfResourceRef();
				pdfResourceRef->resource = pdfResource;

				if(pdfResourceRef->resource != NULL) {
					pdfResourceRef->name = **keyPtr;
					pdfResourceRef->type = **resourceType;
					pdfResourceRef->resource->type = **resourceType;

					if(resources == NULL) {
						resources = new ResourceRefList(pdfResourceRef);
						refPtr = resources;
					}
					else {
						refPtr->next = new ResourceRefList(pdfResourceRef);
						refPtr = refPtr->next;
					}
				}

				delete *keyPtr;
				keyPtr++;
			}
			if(keys != NULL) {
				delete [] keys;
			}
		}
		delete *resourceType;
		resourceType++;
	}
	if(resourceTypeList != NULL) 
		delete [] resourceTypeList;

	if(OUTPUT_LEVEL > 2) printf("Ending ObjParser::parseResources\n");

	return resources;
}

//-----------------------------------------------

PdfResource* ObjParser::getResource(PObj *obj, String resourceType, PObj* root, PdfRoot *pdfRoot) {
	PObj *item;
	PdfResource *pdfResource = NULL;
	int id;
	char idStr[20];

	if(OUTPUT_LEVEL > 2) printf("Starting ObjParser::parseResource\n");

	obj = dereference(root, obj);
	if(obj == NULL)
		return NULL;

	sprintf(idStr, "%d", obj->id);
	if(obj->id != 0)
		pdfResource = (PdfResource*) pdfRoot->resourceHash.get(idStr);

	if(pdfResource == NULL) {
		if(resourceType == "Font") {
			if(OUTPUT_LEVEL > 1) printf("  Font found\n");
			pdfResource = parseFont(obj, root);
		}
		else if(resourceType == "XObject") {
			if(OUTPUT_LEVEL > 1) printf(" XObject found\n");
			pdfResource = parseXObject(obj, root);
		}
		else if(resourceType == "ColorSpace") {
			if(OUTPUT_LEVEL > 1) printf("  ColorSpace found\n");
			pdfResource = parseColorSpace(obj, root);
			((PdfColorSpace*)pdfResource)->isPageResource = true;
		}
		else if(resourceType == "ExtGState") {
			if(OUTPUT_LEVEL > 1) printf("  ExtGraphicsState found\n");
			pdfResource = parseExtGraphicsState(obj, root);
		}
		else {
			printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Unknown resource type: %s\n", resourceType.nullterm() );
		}

		if(pdfResource != NULL) {
			if(obj->id != 0)
				pdfRoot->addResource(pdfResource);
			pdfRoot->resourceHash.add(idStr, pdfResource);
		}
	}

	if(OUTPUT_LEVEL > 2) printf("Ending ObjParser::parseResource\n");

	return pdfResource;
}

//-----------------------------------------------

PdfFont* ObjParser::parseFont(PObj *font, PObj *root) {
	PObj *obj, *widths, *widthVal;
	PObjList *list;
	int i;

	if(OUTPUT_LEVEL > 2) printf("Starting ObjParser::parseFont\n");

	PdfFont *pdfFont = new PdfFont();

	font = validateObj(root, font, t_dictionary);
	if(font == NULL) {
		printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - Could not find Font\n");
		delete pdfFont;
		return NULL;
	}

	if(OUTPUT_LEVEL > 1) printf(" Parsing Font\n");

	// Required Fields
	obj = getValidItem(root, font, "Subtype", t_name);
	if(obj == NULL) {
		printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - font missing Subtype field\n");
		delete pdfFont;
		return NULL;
	}
	pdfFont->subType = ((PDF_name*) obj)->value;

	if(pdfFont->subType == "Type1" || pdfFont->subType == "TrueType") {

		obj = getValidItem(root, font, "BaseFont", t_name);
		if(obj == NULL) {
			printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - font missing BaseFont field\n");
			delete pdfFont;
			return NULL;
		}
		pdfFont->fontName = ((PDF_name*) obj)->value;
	
		obj = getValidItem(root, font, "data_subtype", t_name);
		if(obj != NULL) {
			pdfFont->dataSubType = ((PDF_name*) obj)->value;
		}

		if(pdfFont->isBase14Font())
			pdfFont->initBase14Font();

		// Fields Required if the font is not one of the base 14
		obj = getValidItem(root, font, "FirstChar", t_integer);
		if(obj == NULL) {
			if(! pdfFont->isBase14Font())
				printerr("Error - font missing FirstChar field\n");
		}
		else {
			pdfFont->firstChar = ((PDF_int*) obj)->value;
		}

		obj = getValidItem(root, font, "LastChar", t_integer);
		if(obj == NULL) {
			if(! pdfFont->isBase14Font())
				printerr("Error - font missing LastChar field\n");
		}
		else {
			pdfFont->lastChar = ((PDF_int*) obj)->value;
		}
	
		obj = getValidItem(root, font, "Encoding", t_name);
		if(obj != NULL)
			pdfFont->encoding = ((PDF_name*) obj)->value;
	
		widths = getValidItem(root, font, "Widths", t_array);
		if(widths == NULL) {
			if(! pdfFont->isBase14Font())
				printerr("Could not find font widths\n");
		}
		else {
			pdfFont->numWidths = pdfFont->lastChar - pdfFont->firstChar + 1;
			pdfFont->widths = new long[pdfFont->numWidths];
	
			list = widths->subobjects;
			for(i=0; i<pdfFont->numWidths; i++) {
				if(list == NULL) 
					printerr("Error - font missing elements of the Widths array\n");
	
				widthVal = validateObj(root, list->theobject, t_integer);
				if(widthVal == NULL) 
					printerr("Error - font width elements must be integers\n");
	
				pdfFont->widths[i] = ((PDF_int*)widthVal)->value;
				list = list->next;
			}
			if(list != NULL) 
				printerr("Error - font has extra elements in the Widths array\n");
		}
	
		obj = getValidItem(root, font, "FontDescriptor", t_dictionary);
	
		if(obj == NULL) {
			if(! pdfFont->isBase14Font())
				printerr("Error - font descriptor missing\n");
		}
		else {
			parseFontDescriptor(obj, root, pdfFont);
		}
	}
	else {
		printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Font type %s not supported\n", pdfFont->subType.nullterm());
		delete pdfFont;
		return NULL;
	}

	//if(OUTPUT_LEVEL > 1) printf("%s\n", font->render_xml(0).nullterm());

	//if(OUTPUT_LEVEL > 2) printf("Ending ObjParser::parseFont\n");

	return pdfFont;
}

//-----------------------------------------------

void ObjParser::parseFontDescriptor(PObj *fontDesc, PObj *root, PdfFont *pdfFont) {
	PObj *obj, *bBox, *bBoxVal;
	PObjList *list;
	int i;
	PObj *blob;
	char *data;
	int len;

	if(OUTPUT_LEVEL > 2) printf("Starting ObjParser::parseFontDescriptor\n");

	if(fontDesc == NULL)
		printerr("Error - could not find FontDescriptor\n");

	if(OUTPUT_LEVEL > 1) printf("  Found Font descriptor\n");
	if(OUTPUT_LEVEL > 1) printDictionary(fontDesc, "   ", root);

	obj = getValidItem(root, fontDesc, "Ascent", t_integer);
	if(obj == NULL)
		printerr("Error - fontDesc missing Ascent field\n");
	pdfFont->ascent = ((PDF_int*) obj)->value;

	obj = getValidItem(root, fontDesc, "Descent", t_integer);
	if(obj == NULL)
		printerr("Error - fontDesc missing Descent field\n");
	pdfFont->descent = ((PDF_int*) obj)->value;

	obj = getValidItem(root, fontDesc, "Flags", t_integer);
	if(obj == NULL)
		printerr("Error - fontDesc missing Flags field\n");
	pdfFont->flags = ((PDF_int*) obj)->value;

	bBox = getValidItem(root, fontDesc, "FontBBox", t_array);
	if(bBox == NULL)
		printerr("Could not find fontDescriptor fontBBox\n");

	list = bBox->subobjects;
	for(i=0; i<4; i++) {
		if(list == NULL) 
			printerr("Error - fontDescriptor fontBBox missing elements\n");

		bBoxVal = validateObj(root, list->theobject, t_numeric);
		if(bBoxVal == NULL) 
			printerr("Error - fontDescriptor fontBBox elements must be integers\n");

		pdfFont->fontBBox[i] = ((PDF_numeric*)bBoxVal)->getNumericValue();
		list = list->next;
	}
	if(list != NULL) 
		printerr("Error - fontDescriptor fontBBox has extra elements\n");

	if(pdfFont->subType == "Type1") {
		obj = getValidItem(root, fontDesc, "FontFile", t_stream);
		if(obj != NULL) {
			blob =((PDF_stream*)obj)->dataObject;
			if(blob == NULL)
				printerr("Error - stream is missing internal data object\n");

			len = ((PDF_blob*)blob)->length;
			data = ((PDF_blob*)blob)->rawdata;
			pdfFont->data = new char[len];
			pdfFont->dataLength = len;
			memcpy(pdfFont->data, data, len);

			// Find filters for stream
			parseFilter(pdfFont, ((PDF_stream*)obj)->dictionary, root);
		}
		else {
			obj = getValidItem(root, fontDesc, "FontFile3", t_stream);
			if(obj != NULL) {
				blob =((PDF_stream*)obj)->dataObject;
				if(blob == NULL)
					printerr("Error - stream is missing internal data object\n");

				len = ((PDF_blob*)blob)->length;
				data = ((PDF_blob*)blob)->rawdata;
				pdfFont->data = new char[len];
				pdfFont->dataLength = len;
				memcpy(pdfFont->data, data, len);

				if(((PDF_stream*)obj)->dictionary == NULL)
					printerr("Error - no dictionary for Font Stream\n");

				// Find filters for stream
				parseFilter(pdfFont, ((PDF_stream*)obj)->dictionary, root);

				obj = getValidItem(root, ((PDF_stream*)obj)->dictionary, "Subtype", t_name);
				if(obj == NULL) 
					printerr("Error - Type1 fonts using the FontFile3 tag must have a Subtype\n");
				
				if( ((PDF_name*)obj)->value != "Type1C")
					printerr("Unsupported Type1 Subtype encoding %s\n", ((PDF_name*)obj)->value.nullterm() );

				pdfFont->dataSubType = "Type1C";
			}
		}
	}
	else if(pdfFont->subType == "TrueType") {
		obj = getValidItem(root, fontDesc, "FontFile2", t_stream);
		if(obj != NULL) {
			blob =((PDF_stream*)obj)->dataObject;
			if(blob == NULL)
				printerr("Error - stream is missing internal data object\n");

			len = ((PDF_blob*)blob)->length;
			data = ((PDF_blob*)blob)->rawdata;
			pdfFont->data = new char[len];
			pdfFont->dataLength = len;
			memcpy(pdfFont->data, data, len);

			// Find filters for stream
			parseFilter(pdfFont, ((PDF_stream*)obj)->dictionary, root);
		}
	}
	else {
		printerr("Unknown font type found: %s\n", pdfFont->subType.nullterm());
	}

	if(pdfFont->subType == "TrueType" && pdfFont->dataLength > 0) {
		pdfFont->decode();
		if(pdfFont->filters != NULL) {
			printerr("Error - could not decode TrueTypeFont for parsing\n");
		}

		FontReader::readFont(pdfFont);
	}

	if(OUTPUT_LEVEL > 2) printf("Ending ObjParser::parseFontDescriptor\n");
}

//-----------------------------------------------

String ObjParser::getColorSpaceCatagory(PObj *colorSpace, PObj *root) {
	PObj *obj;

	colorSpace = dereference(root, colorSpace);
	if(colorSpace == NULL) {
		printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - missing ColorSpace field\n");
		return String();
	}
	if(colorSpace->get_type() == t_name) {
		return ((PDF_name*) colorSpace)->value;
	}
	else if(colorSpace->get_type() == t_array) {
		obj = validateObj(root, colorSpace->subobjects->theobject, t_name);
		if(obj == NULL) {
			printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - missing ColorSpace catagory\n");
			return String();
		}
		return ((PDF_name*) obj)->value;
	}
	else {
		return String();
	}
} 

//-----------------------------------------------

PdfColorSpace* ObjParser::parseColorSpace(PObj *colorSpace, PObj *root) {
	PdfColorSpace *result = new PdfColorSpace();
	PObjList *item;
	PObj *obj;
	int i;

	colorSpace = dereference(root, colorSpace);
	if(colorSpace == NULL) {
		printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - missing ColorSpace field\n");
		return NULL;
	}
	if(colorSpace->get_type() == t_name) {
		result->catagory = ((PDF_name*) colorSpace)->value;

		if(result->catagory != "DeviceGray" && result->catagory != "DeviceRGB" && result->catagory != "DeviceCMYK") {
			printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - incomplete ColorSpace\n");
			delete result;
			return NULL;
		}
	}
	else if(colorSpace->get_type() == t_array) {
		PObjList *list = colorSpace->subobjects;
		PDF_dictionary *dict;

		obj = validateObj(root, list->theobject, t_name);
		if(obj == NULL) {
			printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - missing ColorSpace catagory\n");
			delete result;
			return NULL;
		}
		result->catagory = ((PDF_name*) obj)->value;
		list = list->next;


		if(result->catagory == "CalGray" || result->catagory == "CalRGB" || result->catagory == "Lab") {
			// printf("Device Independant ColorSpace\n");
			// printf("Colorspace catagory: %s\n", result->catagory.nullterm());

			dict = (PDF_dictionary*) validateObj(root, list->theobject, t_dictionary);
			if(dict == NULL) {
				printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - missing ColorSpace dictionary\n");
				delete result;
				return NULL;
			}
			// printf("Found ColorSpace dictionary\n");

			obj = getValidItem(root, dict, "WhitePoint", t_array);
			if(obj == NULL) {
				printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - WhitePoint required for device independant ColorSpace\n");
				delete result;
				return NULL;
			}

			item = obj->subobjects;
			for(i=0; i<3; i++) {
				if(item == NULL || item->theobject == NULL || 
				   (item->theobject->get_type() != t_integer && item->theobject->get_type() != t_real) ) {
					printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - ColorSpace missing WhitePoint values\n");
					return NULL;
				}
				result->whitePoint[i] = ((PDF_numeric*) item->theobject)->getNumericValue();
			}

			obj = getValidItem(root, dict, "BlackPoint", t_array);
			if(obj != NULL) {
				printf("Found ColorSpace BlackPoint\n");

				item = obj->subobjects;
				for(i=0; i<3; i++) {
					if(item == NULL || item->theobject == NULL || 
					   (item->theobject->get_type() != t_integer && item->theobject->get_type() != t_real) ) {
						printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - ColorSpace missing BlackPoint values\n");
						return NULL;
					}
					result->blackPoint[i] = ((PDF_numeric*) item->theobject)->getNumericValue();
				}

			}
			if(result->catagory == "CalRGB") {
				obj = getValidItem(root, dict, "Gamma", t_array);
				if(obj != NULL) {
					// printf("Found ColorSpace Gamma\n");

					item = obj->subobjects;
					for(i=0; i<3; i++) {
						if(item == NULL || item->theobject == NULL || 
						   (item->theobject->get_type() != t_integer && item->theobject->get_type() != t_real) ) {
							printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - ColorSpace missing Gamma values\n");
							return NULL;
						}
						result->gamma[i] = ((PDF_numeric*) item->theobject)->getNumericValue();
					}
				}

				obj = getValidItem(root, dict, "Matrix", t_array);
				if(obj != NULL) {
					// printf("Found ColorSpace Matrix\n");

					item = obj->subobjects;
					for(i=0; i<3; i++) {
						if(item == NULL || item->theobject == NULL || 
						   (item->theobject->get_type() != t_integer && item->theobject->get_type() != t_real) ) {
							printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - ColorSpace missing Matrix values\n");
							return NULL;
						}
						result->matrix[i] = ((PDF_numeric*) item->theobject)->getNumericValue();
					}
				}
			}
			else if(result->catagory == "Lab") {
				obj = getValidItem(root, dict, "Range", t_array);
				if(obj != NULL) {
					// printf("Found ColorSpace Range\n");

					item = obj->subobjects;
					for(i=0; i<3; i++) {
						if(item == NULL || item->theobject == NULL || 
						   (item->theobject->get_type() != t_integer && item->theobject->get_type() != t_real) ) {
							printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - ColorSpace missing Range values\n");
							return NULL;
						}
						result->range[i] = ((PDF_numeric*) item->theobject)->getNumericValue();
					}
				}
			}

			list = list->next;
			if(list != NULL) {
				printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - too many arguments for a device independant ColorSpace\n");
				delete result;
				return NULL;
			}
		}
		else {
			// printf("Other ColorSpace\n");
			if(result->catagory == "Indexed") {
				String baseCatagory = getColorSpaceCatagory(list->theobject, root);
				if(baseCatagory == "" || baseCatagory == "Indexed" || baseCatagory == "Pattern") {
					printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - invalid Indexed ColorSpaces base\n");
					delete result;
					return NULL;
				}
				result->base = parseColorSpace(list->theobject, root);
				if(result->base == NULL) {
					printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - Indexed ColorSpaces must have a base\n");
					delete result;
					return NULL;
				}
				// printf("Found ColorSpace base\n");
				list = list->next;

				obj = validateObj(root, list->theobject, t_integer);
				if(obj == NULL) {
					printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - Indexed ColorSpaces must have a hival\n");
					delete result;
					return NULL;
				}
				result->hival = ((PDF_int*) obj)->value;
				// printf("Found ColorSpace hival\n");
				list = list->next;

				obj = validateObj(root, list->theobject, t_stream);
				if(obj == NULL || ((PDF_stream*) obj)->dataObject == NULL) {
					printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - Indexed ColorSpaces must have a look up table\n");
					delete result;
					return NULL;
				}

				// Find filters for stream
				parseFilter( result, ((PDF_stream*) obj)->dictionary, root);

				obj = ((PDF_stream*) obj)->dataObject;
				result->dataLength = ((PDF_blob*) obj)->length;
				result->data = new char[result->dataLength];
				memcpy(result->data, ((PDF_blob*) obj)->rawdata, result->dataLength);

				if(DECODE_IMAGES)
					result->decode();

				// printf("Found ColorSpace lookup\n");
				list = list->next;

				if(list != NULL) {
					printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - too many arguments for a Indexed ColorSpace\n");
					delete result;
					return NULL;
				}
			}
			else if(result->catagory == "Separation") {
				// printf("Found Separation ColorSpace\n");
				obj = validateObj(root, list->theobject, t_name);
				if(obj == NULL) {
					printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - Separation ColorSpaces must have a name\n");
					delete result;
					return NULL;
				}
				result->separationName = ((PDF_name*) obj)->value;
				// printf("Found ColorSpace separationName\n");
				list = list->next;

				String altSpaceCatagory = getColorSpaceCatagory(list->theobject, root);
				if(altSpaceCatagory == "" || altSpaceCatagory == "Indexed" || altSpaceCatagory == "Pattern" || altSpaceCatagory == "Separation") {
					printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - invalid Separation ColorSpaces altSpace\n");
					delete result;
					return NULL;
				}
				result->altSpace = parseColorSpace(list->theobject, root);
				if(result->altSpace == NULL) {
					printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - Separation ColorSpaces must have an alternate Space\n");
					delete result;
					return NULL;
				}
				// printf("Found ColorSpace altSpace\n");
				list = list->next;

				result->tintTransform = parseFunction(list->theobject, root);
				if(result->tintTransform == NULL) {
					printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - Separation ColorSpaces must have a tint transform\n");
					delete result;
					return NULL;
				}
				// printf("Found ColorSpace tint transform\n");
				list = list->next;

				if(list != NULL) {
					printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - too many arguments for a Separation ColorSpace\n");
					delete result;
					return NULL;
				}
			}
			else if(result->catagory == "Pattern") {
				printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - Pattern ColorSpaces are not yet supported\n");
				delete result;
				return NULL;
			}
			else {
				printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - The specified Image ColorSpace '%s' is not yet supported\n", result->catagory.nullterm() );
				delete result;
				return NULL;
			}
		}

	}
	else {
		printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - Invalid ColorSpace Format\n");
		delete result;
		return NULL;
	}

	return result;
} 

//-----------------------------------------------

PdfResource* ObjParser::parseXObject(PObj *xObject, PObj *root) {
	PObj *dict, *obj, *blob;
	PdfResource *pdfResource = NULL;

	if(OUTPUT_LEVEL > 2) printf("Starting ObjParser::parseXObject\n");

	xObject = validateObj(root, xObject, t_stream);
	if(xObject == NULL) {
		printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - Could not find XObject\n");
		return NULL;
	}

	if(OUTPUT_LEVEL > 1) printf(" Found XObject Stream\n");

	dict = ((PDF_stream*)xObject)->dictionary;
	dict = validateObj(root, dict, t_dictionary);
	if(dict == NULL) {
		printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - XObject missing dictionary\n");
		return NULL;
	}

	if(OUTPUT_LEVEL > 1) printf(" Found XObject Dictionary\n");

	if(OUTPUT_LEVEL > 1) printDictionary(dict, "   ", root);

	obj = getValidItem(root, dict, "Subtype", t_name);
	if(obj == NULL) {
		printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - could not find XObject subtype\n");
		return NULL;
	}

	if(((PDF_name*) obj)->value == String("Image")) {
		pdfResource = new PdfImage();
		if(OUTPUT_LEVEL > 1) printf(" Found Image\n");

		//if(OUTPUT_LEVEL > 1) printf("%s\n", dict->render_xml(0).nullterm());

		((PdfImage*)pdfResource)->format = "native";

		obj = getValidItem(root, dict, "Name", t_name);
		if(obj != NULL)
			((PdfImage*)pdfResource)->name = ((PDF_name*) obj)->value;

		obj = getValidItem(root, dict, "Height", t_integer);
		if(obj == NULL) {
			printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - Image missing Height field\n");
			delete pdfResource;
			return NULL;
		}
		((PdfImage*)pdfResource)->height = ((PDF_int*) obj)->value;

		obj = getValidItem(root, dict, "Width", t_integer);
		if(obj == NULL) {
			printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - Image missing Width field\n");
			delete pdfResource;
			return NULL;
		}
		((PdfImage*)pdfResource)->width = ((PDF_int*) obj)->value;

		obj = getValidItem(root, dict, "BitsPerComponent", t_integer);
		if(obj == NULL) {
			printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - Image missing BitsPerComponent field\n");
			delete pdfResource;
			return NULL;
		}
		((PdfImage*)pdfResource)->bitsPerPixel = ((PDF_int*) obj)->value;

		// Find filters for stream
		parseFilter(pdfResource, dict, root);

		obj = ((PDF_dictionary*) dict)->getValueForKey("ColorSpace");
		((PdfImage*) pdfResource)->colorSpace = parseColorSpace(obj, root);
		if(((PdfImage*) pdfResource)->colorSpace == NULL) {
			delete pdfResource;
			return NULL;
		}
	}
	else if(((PDF_name*) obj)->value == String("Form")) {
		if(OUTPUT_LEVEL > 0) printf(" Found Form\n");
	}
	else if(((PDF_name*) obj)->value == String("PS")) {
		if(OUTPUT_LEVEL > 0) printf(" Found PostScript Object\n");
	}
	else {
		printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - unknown XObject type found\n");
	}

	if(pdfResource != NULL) {
		blob = ((PDF_stream*) xObject)->dataObject;
		if(blob == NULL) {
			printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - XObject stream missing blob\n");
			delete pdfResource;
			return NULL;
		}

		pdfResource->dataLength = ((PDF_blob*)blob)->length;
		if(pdfResource->dataLength > 0) { 
			pdfResource->data = new char[pdfResource->dataLength];
			memcpy(pdfResource->data, ((PDF_blob*)blob)->rawdata, pdfResource->dataLength);

			if(DECODE_IMAGES)
				pdfResource->decode();
		}
	}

	if(OUTPUT_LEVEL > 2) printf("Ending ObjParser::parseXObject\n");

	return pdfResource;
}

//-----------------------------------------------

void ObjParser::parseFilter(PdfResource *pdfResource, PObj *dict, PObj *root) {
	PObj *obj, *filterObj;
	PObjList *filterItem;

	dict = validateObj(root, dict, t_dictionary);
	if(dict == NULL) {
		printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - Could not find dictionary containing filter\n");
		return;
	}

	obj = ((PDF_dictionary*) dict)->getValueForKey("Filter");
	obj = dereference(root, obj);

	if(obj == NULL) {
	}
	else if(obj->get_type() == t_name) {
		// printf("Filter: %s\n", ((PDF_name*) obj)->value.nullterm());
		((PdfResource*)pdfResource)->addFilter( ((PDF_name*) obj)->value );
	}
	else if(obj->get_type() == t_array) {
		filterItem = obj->subobjects;
		while(filterItem != NULL) {
			filterObj = validateObj(root, filterItem->theobject, t_name);
			if(filterObj == NULL) {
				printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - invalid object type found in image filter array\n");
				return;
			}

			// printf("Filter: %s\n", ((PDF_name*) filterObj)->value.nullterm());
			((PdfResource*)pdfResource)->addFilter( ((PDF_name*) filterObj)->value );
			filterItem = filterItem->next;
		}
	}
	else {
		printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - unkown object type for image Filter\n");
	}
}

//-----------------------------------------------

PdfFunction* ObjParser::parseFunction(PObj* functStream, PObj* root) {
	PObj *streamDict;
	PObj *blobObj;
	PObj *obj;
	PdfFunction *function = NULL;
	int numEntries, i;
	PObjList *ptr;

	// Check stream
	functStream = validateObj(root, functStream, t_stream);
	if(functStream == NULL) {
		printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - Could not find dictionary containing filter\n");
		return NULL;
	}
	// printf("Found Function Stream\n");

	// Get Dictionary
	streamDict = ((PDF_stream*) functStream)->dictionary;
	streamDict = validateObj(root, streamDict, t_dictionary);
	if(streamDict == NULL) {
		printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - invalid function stream dictionary\n");
		return NULL;
	}
	// printf("Found function stream dictionary\n");

	// Get Data blob
	blobObj = ((PDF_stream*) functStream)->dataObject;
	blobObj = validateObj(root, blobObj, t_blob);
	if(blobObj==NULL) {
		printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - data missing from function stream\n");
		return NULL;
	}
	// printf("Found function stream data\n");

	// Create Function
	function = new PdfFunction();

	// Get FunctinType
	obj = getValidItem(root, streamDict, "FunctionType", t_integer);
	if(obj == NULL) {
		printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - Function missing FunctionType field\n");
		delete function;
		return NULL;
	}
	function->functionType = ((PDF_int*) obj)->value;

	// Get Domain
	obj = getValidItem(root, streamDict, "Domain", t_array);
	if(obj == NULL) {
		printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - Function missing Domain Array\n");
		delete function;
		return NULL;
	}

	numEntries = obj->countsubs();
	if(numEntries % 2) {
		printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - Function has mismatched pairs for domain\n");
		delete function;
		return NULL;
	}
	function->numIn = numEntries / 2;

	function->domain = new double[numEntries];
	ptr = obj->subobjects;
	for(i=0; i<function->numIn; i++) {
		// Min 
		obj = validateObj(root, ptr->theobject, t_numeric);
		if(obj == NULL) {
			printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - could not read Function Domain Array\n");
			delete function;
			return NULL;
		}
		
		function->domain[2*i] = ((PDF_numeric*)obj)->getNumericValue();
		ptr = ptr->next;
	
		// Max
		obj = validateObj(root, ptr->theobject, t_numeric);
		if(obj == NULL) {
			printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - bad elements in Function Domain Array\n");
			delete function;
			return NULL;
		}
		
		function->domain[2*i+1] = ((PDF_numeric*)obj)->getNumericValue();
		ptr = ptr->next;

		if(function->domain[2*i] > function->domain[2*i+1]) {
			printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - the first element of each function domain pair must be less that the second\n");
			delete function;
			return NULL;
		}
	}

	// Get Range
	obj = getValidItem(root, streamDict, "Range", t_array);
	if(obj == NULL) {
		printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - Function missing Range Array\n");
		delete function;
		return NULL;
	}

	numEntries = obj->countsubs();
	if(numEntries % 2) {
		printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - Function has mismatched pairs for range\n");
		delete function;
		return NULL;
	}
	function->numOut = numEntries / 2;

	function->range = new double[numEntries];
	ptr = obj->subobjects;
	for(i=0; i<function->numOut; i++) {
		// Min 
		obj = validateObj(root, ptr->theobject, t_numeric);
		if(obj == NULL) {
			printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - could not read Function Range Array\n");
			delete function;
			return NULL;
		}
		
		function->range[2*i] = ((PDF_numeric*)obj)->getNumericValue();
		ptr = ptr->next;
	
		// Max
		obj = validateObj(root, ptr->theobject, t_numeric);
		if(obj == NULL) {
			printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - bad elements in Function Range Array\n");
			delete function;
			return NULL;
		}
		
		function->range[2*i+1] = ((PDF_numeric*)obj)->getNumericValue();
		ptr = ptr->next;

		if(function->range[2*i] > function->range[2*i+1]) {
			printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - the first element of each function range pair must be less that the second\n");
			delete function;
			return NULL;
		}
	}

	// Get Size
	obj = getValidItem(root, streamDict, "Size", t_array);
	if(obj == NULL) {
		printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - Function missing Size Array\n");
		delete function;
		return NULL;
	}

	numEntries = obj->countsubs();
	if(numEntries != function->numIn) {
		printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - Function size array has different number of dimensions that domain array\n");
		delete function;
		return NULL;
	}

	function->size = new int[function->numIn];
	ptr = obj->subobjects;
	for(i=0; i<function->numIn; i++) {
		obj = validateObj(root, ptr->theobject, t_integer);
		if(obj == NULL) {
			printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - could not read Function Size Array\n");
			delete function;
			return NULL;
		}
		
		function->size[i] = ((PDF_int*)obj)->value;
		ptr = ptr->next;
	}

	// Get BitsPerSample
	obj = getValidItem(root, streamDict, "BitsPerSample", t_integer);
	if(obj == NULL) {
		printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - Function missing BitsPerSample field\n");
		delete function;
		return NULL;
	}
	function->bitsPerSample = ((PDF_int*) obj)->value;

	// Get Order
	obj = getValidItem(root, streamDict, "Order", t_integer);
	if(obj != NULL) 
		function->order = ((PDF_int*) obj)->value;

	// Get Encode
	obj = getValidItem(root, streamDict, "Encode", t_array);
	if(obj != NULL) {
		numEntries = obj->countsubs();
		if(2*numEntries != function->numIn) {
			printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - Function Encode array has different number of dimensions that Domain array\n");
			delete function;
			return NULL;
		}

		function->encode = new double[numEntries];
		ptr = obj->subobjects;
		for(i=0; i<function->numIn; i++) {
			// Min 
			obj = validateObj(root, ptr->theobject, t_numeric);
			if(obj == NULL) {
				printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - could not read Function Encode Array\n");
				delete function;
				return NULL;
			}
			
			function->encode[2*i] = ((PDF_numeric*)obj)->getNumericValue();
			ptr = ptr->next;
		
			// Max
			obj = validateObj(root, ptr->theobject, t_numeric);
			if(obj == NULL) {
				printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - bad elements in Function Encde Array\n");
				delete function;
				return NULL;
			}
			
			function->encode[2*i+1] = ((PDF_numeric*)obj)->getNumericValue();
			ptr = ptr->next;

			if(function->encode[2*i] > function->encode[2*i+1]) {
				printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - the first element of each function encode pair must be less that the second\n");
				delete function;
				return NULL;
			}
		}
	}

	// Get Decode
	obj = getValidItem(root, streamDict, "Decode", t_array);
	if(obj != NULL) {
		numEntries = obj->countsubs();
		if(2*numEntries != function->numOut) {
			printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - Function Decode array has different number of dimensions that Range array\n");
			delete function;
			return NULL;
		}

		function->decode = new double[numEntries];
		ptr = obj->subobjects;
		for(i=0; i<function->numOut; i++) {
			// Min 
			obj = validateObj(root, ptr->theobject, t_numeric);
			if(obj == NULL) {
				printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - could not read Function Decode Array\n");
				delete function;
				return NULL;
			}
			
			function->decode[2*i] = ((PDF_numeric*)obj)->getNumericValue();
			ptr = ptr->next;
		
			// Max
			obj = validateObj(root, ptr->theobject, t_numeric);
			if(obj == NULL) {
				printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - bad elements in Function Decode Array\n");
				delete function;
				return NULL;
			}
			
			function->decode[2*i+1] = ((PDF_numeric*)obj)->getNumericValue();
			ptr = ptr->next;

			if(function->decode[2*i] > function->decode[2*i+1]) {
				printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - the first element of each function decode pair must be less that the second\n");
				delete function;
				return NULL;
			}
		}
	}

	function->dataLength = ((PDF_blob*)blobObj)->length;
	if(function->dataLength > 0) { 
		function->data = new char[function->dataLength];
		memcpy(function->data, ((PDF_blob*)blobObj)->rawdata, function->dataLength);
	}

	parseFilter(function, streamDict, root);

	return function;
}

// ----------------------------------------------

PdfExtGraphicsState* ObjParser::parseExtGraphicsState(PObj* dict, PObj* root) {
	PObj *obj;
	PdfExtGraphicsState *eGState = new PdfExtGraphicsState();
	int i;

	// Get Dictionary
	dict = validateObj(root, dict, t_dictionary);
	if(dict == NULL) {
		printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - invalid Extended Graphics State dictionary\n");
		return NULL;
	}
	// printf("Found Extended Graphics State dictionary\n");

	// Get SA
	obj = getValidItem(root, dict, "SA", t_boolean);
	if(obj != NULL) {
		if( ((PDF_boolean*)obj)->value )
			eGState->sa = 1;
		else
			eGState->sa = 0;
	}

	// Get OP
	obj = getValidItem(root, dict, "OP", t_boolean);
	if(obj != NULL) {
		if( ((PDF_boolean*)obj)->value )
			eGState->op = 1;
		else
			eGState->op = 0;
	}

	// Get BG
	obj = getValidItem(root, dict, "BG", t_stream);
	if(obj != NULL)
		eGState->bg = parseFunction(obj, root);

	// Get UCR
	obj = getValidItem(root, dict, "UCR", t_stream);
	if(obj != NULL)
		eGState->ucr = parseFunction(obj, root);

	// Get HTP
	obj = getValidItem(root, dict, "HTP", t_array);
	if(obj != NULL) {
		PObjList *list = obj->subobjects;

		eGState->htp = new int[2];
		if(list == NULL) {
			delete [] eGState->htp;
			eGState->htp = NULL;
			printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - empty ExtGState HTP array\n");
		}
		else {
			obj = validateObj(root, list->theobject, t_integer);
			if(obj == NULL) {
				delete [] eGState->htp;
				eGState->htp = NULL;
				printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - invalid ExtGState HTP element\n");
			}
			else {
				eGState->htp[0] = ((PDF_int*)obj)->value;
				list = list->next;
				if(list == NULL) {
					delete [] eGState->htp;
					eGState->htp = NULL;
					printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - invalid ExtGState HTP Element\n");
				}
				else {
					obj = validateObj(root, list->theobject, t_integer);
					if(obj == NULL) {
						delete [] eGState->htp;
						eGState->htp = NULL;
						printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - Invalid ExtGState HTP element\n");
					}
					else {
						eGState->htp[1] = ((PDF_int*)obj)->value;
						if(list->next != NULL) 
							printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - too many ExtGState HTP elements\n");
					}
				}
			}
		}
	}

	return eGState;
}

//===============================================
// Utility Functions

PObj* ObjParser::dereference(PObj *root, PObj *obj) {
	if(obj == NULL) {
		if(OUTPUT_LEVEL > 1) printf("INVALID - empty reference\n");
		return NULL;
	}

	while(obj->get_type() == t_reference) {
		obj = ((PDF_ref*) obj)->getObject(root);

		if(obj == NULL) {
			if(OUTPUT_LEVEL > 1) printf("INVALID - referenced object not found\n");
			return NULL;
		}
	}

	return obj;
}

PObj* ObjParser::validateObj(PObj *root, PObj *obj, ObjectType oType) {
	if(oType != t_reference)
		obj = dereference(root, obj);

	if(obj == NULL) {
		if(OUTPUT_LEVEL > 1) printf("INVALID - object not found\n");
		return NULL;
	}

	if(oType == t_numeric && obj->isNumeric())
		return obj;

	if(obj->get_type() != oType) {
		if(OUTPUT_LEVEL > 1) printf("INVALID - invalid object type - %s\n", obj->get_printable_type());
		return NULL;
	}

	return obj;
}

//-----------------------------------------------

PObj* ObjParser::getValidItem(PObj *root, PObj *dict, char* key, ObjectType oType) {
	PObj *result;

	if(dict == NULL) {
		if(OUTPUT_LEVEL > 1) printf("INVALID - dictionary object not found\n");
		return NULL;
	}
	if(dict->get_type() != t_dictionary) {
		if(OUTPUT_LEVEL > 1) printf("INVALID - dictionary type expected - %s\n", dict->get_printable_type());
		return NULL;
	}

	result = ((PDF_dictionary*) dict)->getValueForKey(key);

	if(result == NULL)
		return NULL;

	return validateObj(root, result, oType);
}

//-----------------------------------------------

void ObjParser::printDictionary(PObj *dict, char *prefix, PObj* root) {
	String **keys, **keyPtr;
	PObj *obj;

	if(dict == NULL) 
		return;

	keys = ((PDF_dictionary*) dict)->getKeys();
	if(keys != NULL) {
		keyPtr = keys;
		while(*keyPtr != NULL) {
			printf("%s%s: ", prefix, (*keyPtr)->nullterm());

			obj = ((PDF_dictionary*) dict)->getValueForKey(**keyPtr);
			obj = dereference(root, obj);
			if(obj == NULL) {
				printerr(ESTRICTSYNTAX|EOTHER_GENERICERROR, "Error - could not dictionary value\n");
				continue;
			}

			if(obj->get_type() == t_name) {
				printf("%s", ((PDF_name*) obj)->value.nullterm() );
			}
			else if(obj->get_type() == t_integer) {
				printf("%d", ((PDF_int*) obj)->value);
			}
			else {
				printf("%s", obj->get_printable_type());
			}

			printf("\n");

			delete *keyPtr;
			keyPtr++;
		}
		delete [] keys;
	}
}

//---------------------------------------------------------

ResourceRefList::ResourceRefList() : resourceRef(NULL), next(NULL) {
}

ResourceRefList::ResourceRefList(PdfResourceRef *newResourceRef) : resourceRef(newResourceRef), next(NULL) {
}

