#include "PDFReader.hpp"

#include <stdlib.h>
#include <stdio.h>

#include "RegExUtil.hpp"
#include "PDFTokenizer.hpp"

PDFReader::PDFReader() {
	DEBUG_OUTPUT = false;
}

//----------------------------------------

PObj* PDFReader::parse(DataReader *file) {
	RegExUtil regEx;
	char buffer[MAX_LINE_LENGTH+1];
	char **list;
	int state = START;
	char *comment;
	int byteOffset, xrefOffset, objId, objGen;
	PObj *rootNode = new PDF_root;
	PObj *node;
	PDF_root *rootObject = (PDF_root*) rootNode;
	PDF_trailer *trailer;
	Hashtable offsetHash;
	int cnt;

	PDFTokenizer tokens(file);

	xrefOffset = 0;
	buffer[MAX_LINE_LENGTH] = 0;
	byteOffset = file->getByteOffset();
	while( (cnt=file->getLine(buffer, MAX_LINE_LENGTH)) != -1) {
		buffer[cnt] = 0;
		// printf("MAIN %s\n", buffer);

		comment = strchr(buffer, '%');
		if(comment != NULL) {
			*comment = '\0';
			comment++;
			// printf("Comment: %s\n", comment);
		}

		if(regEx.match(buffer, "^[ \t\n]*$") && state != START && state != DONE_TRAILER) {
			// Empty line
		}
		else if(state == START) {
			if(comment != NULL && regEx.match(comment, "^PDF-[0-9]+\\.[0-9]+ *\n")) {
					state = BODY;
			}
		}
		else if(state == BODY) {
			if( (list = regEx.matchList(buffer, "^ *([0-9]+) +([0-9]+) +obj *\n")) != NULL) {
				objId = atoi(list[1]);
				objGen = atoi(list[2]);

				if(DEBUG_OUTPUT) printf("Found object %d revision %d offset: %d -------------------\n", objId, objGen, byteOffset );

				node = tokens.parse("endobj");
				if(node != NULL) {
					node->id = objId;
					node->generation = objGen;
					rootNode->addsub(node);

					if(rootObject->idHash.findobject(objId) == NULL || 
				   	   rootObject->idHash.findobject(objId)->generation < objGen) {
						rootObject->idHash.add(node, objId);
					}
					offsetHash.add(node, byteOffset);
				}

				state = BODY;
				regEx.freeMatches(list);
			}
			else if(regEx.match(buffer, "^ *xref *\n")) { // TEMPORARY - change to require xref
				PObj *deletions;

                if(DEBUG_OUTPUT) printf("Start of XRef\n");

				xrefOffset = byteOffset;
				deletions = readXRef(file, rootObject, &offsetHash);

                if(DEBUG_OUTPUT) printf("Start of Trailer\n");

				trailer = readTrailer(file, rootObject, deletions, xrefOffset);
				rootObject->updateTrailer(trailer);
				rootNode->addsub(trailer);

                state = DONE_TRAILER;
            }
			else {
				printerr("ERROR: Unknown line: '%s'\n", buffer);
				// printerr("ERROR: Unknown line in trailer\n");
			}
		}
		else if(state == DONE_TRAILER) {
			if(comment != NULL && (strcmp(comment, "%EOF") == 0 || strcmp(comment, "%EOF\n") == 0)) {
				if(DEBUG_OUTPUT) printf("PDF document finished\n");

				if(DEBUG_OUTPUT) printf("=========================================================\n");
				// state = FINISHED;
				state = BODY;
			}
			else {
				printf("Comment: %s\n", comment);
				printerr("ERROR: Unknown line after trailer: '%s'\n", buffer);
				// printerr("ERROR: Unknown line after trailer\n");
			}
		}
		else if(state == FINISHED) {
			break;
		}
		else {
			printerr("ERROR: Unknown state\n");
		}
		byteOffset = file->getByteOffset();
	}

	if(state == START) 
		printerr("Error - PDF header line not found\n");
	if(state != FINISHED && state != BODY) 
		printerr("Error - ended in unfinished state\n");

	return rootNode;
}

//----------------------------------------

PObj* PDFReader::readXRef(DataReader *file, PDF_root* rootObject, Hashtable *offsetHash) {
	char buffer[MAX_LINE_LENGTH+1];
	int i, cnt, numEntries;
	char **list;
	RegExUtil regEx;
	int val1, val2;
	PObj *collection = new PDF_collection();
	int numSections = 0;

	buffer[MAX_LINE_LENGTH] = 0;
	cnt = file->getLine(buffer, MAX_LINE_LENGTH);
	buffer[cnt] = 0;

	while( (list=regEx.matchList(buffer, "^([0-9]+) ([0-9]+) *\n")) != NULL) {
		numEntries = atoi(list[2]);
		if(DEBUG_OUTPUT) printf("Starting Xref first entry: %s num entries: %d\n", list[1], numEntries);
		regEx.freeMatches(list);
		numSections++;

		for(i=0; i<numEntries; i++) {
			cnt = file->getLine(buffer, MAX_LINE_LENGTH);
			buffer[cnt] = 0;
			buffer[10] = 0;
			buffer[16] = 0;

			val1 = atoi(buffer);
			val2 = atoi(buffer+11);

			if(buffer[17] == 'n') {
				PObj* obj = offsetHash->findobject(val1);
				if(DEBUG_OUTPUT) printf("In use entry: offset = %d generation = %d\n", val1, val2);
				if(obj == NULL) {
					// printerr("Could not find object at offset %d\n", val1); // TEMP DISABLED
				}
				else {
					if(DEBUG_OUTPUT) printf("Found object %d from offset %d\n", obj->id, val1);
				}
			}
			else {
				if(DEBUG_OUTPUT) printf("Free entry: next free = %d next generation = %d\n", val1, val2);
				PObj *obj = rootObject->idHash.findobject(val1);
				if(obj != NULL && obj != NULL) {
					PDF_deletion *delObj = new PDF_deletion();
					delObj->id = val1;
					collection->addsub(delObj);
				}
			}
		}
		cnt = file->getLine(buffer, MAX_LINE_LENGTH);
		buffer[cnt] = 0;
	}

	if(numSections < 1) {
		printerr("Could not find XRef table- '%s'\n", buffer);
	}

	if(! regEx.match(buffer, "^ *trailer *\n")) {
		printerr("Could not find Start of Trailer- '%s'\n", buffer);
	}

	return collection;
}

//----------------------------------------

PDF_trailer* PDFReader::readTrailer(DataReader *file, PDF_root *rootObject, PObj *deletions, int xrefOffset) {
	char buffer[MAX_LINE_LENGTH+1];
	int cnt, startXref;
	PDFTokenizer tokens(file);
	PDF_trailer *trailerObj = new PDF_trailer();
	PObj *trailerDictionary = NULL;
	PObj *rootRef;

	trailerObj->addsub(deletions);

	trailerDictionary = tokens.parse("startxref");
	if(trailerDictionary == NULL) {
		printerr("Error reading Trailer Dictionary\n");
	}
	else if(trailerDictionary->get_type() != t_dictionary) {
		printerr("Error - unknown object found while looking for Trailer Dictionary\n");
	}
	else {
		rootRef = ((PDF_dictionary*) trailerDictionary)->getValueForKey("Root");
 
		if(rootRef != NULL) {
			if(rootRef->get_type() != t_reference) {
				printerr("Error - invalid root reference object type - %s\n", rootRef->get_printable_type());
			}
			else {
				// printf("Found Root Reference\n");
				rootObject->rootNode = rootRef;
			}
		}

		trailerObj->dictionary = (PDF_dictionary*) trailerDictionary;
		trailerObj->addsub(trailerDictionary);
		cnt = file->getLine(buffer, MAX_LINE_LENGTH);
		buffer[cnt] = 0;

		if( ((PDF_dictionary*) trailerDictionary)->getValueForKey("Encrypt") != NULL)
			printerr("Encrypted documents are not supported at this time.\n");

		startXref = atoi(buffer);
		if(xrefOffset == 0) {
			printerr("Error - xref table not found\n");
		}
		else if(startXref == xrefOffset) {
			if(DEBUG_OUTPUT) printf("xref offset %d correct\n", xrefOffset);
		}
		else {
			// printerr("xref offset %d does not match startXref %d\n", xrefOffset, startXref); // TEMP DISABLED
		}
		if(DEBUG_OUTPUT) printf("StartXref: %d\n", startXref);
	}

	return trailerObj;
}

