#include "PDFTokenizer.hpp"

#include "RegExUtil.hpp"
#include "pdfc_basicobjects.h"
#include "DataFilter.hpp"
#include "PdfObjects.hpp"

#define DEBUG false

//------------------------------------

PDFTokenizer::PDFTokenizer(DataReader *newFile) {
	file = newFile;
	nextChar = buffer;
	bufferEnd = buffer;
	nextToken[0] = 0;
	lastType = t_base_object;
	DEBUG_OUTPUT = false;
	TOKEN_DEBUG = false;
	partialToken = false;
	inStream = false;
	inInlineImage = false;
	inString = false;
	inHexString = false;
}

//------------------------------------

PObj* PDFTokenizer::parse(char *endToken, bool parsingpostscript) {
	char token[MAX_LINE_LENGTH+1];
	PObj *result = NULL;
	int cnt;
	ObjectType objType;

	result = getNextObject(0, parsingpostscript);

	if(endToken == NULL) {
		return result;
	}	
	else {
		if( (cnt=getNextToken(token, &objType, true, true, parsingpostscript)) != -1 && strcmp(token, endToken) == 0) {
			return result;
		}	
		else {
			if(strcmp(token, endToken) == 0) {
				printerr("Error - could not find end of object - %s\n", token);
			}
			else {
				printerr("Error - PDFTokenizer::parse returned -1\n");
			}
			return NULL;
		}
	}
}

//------------------------------------

PObj* PDFTokenizer::getNextObject(int level, bool parsingpostscript) {
	char token[MAX_LINE_LENGTH+1];
	char **list;
	RegExUtil regEx;
	PObj *result = NULL;
	PObj *temp;
	int cnt;
	ObjectType objType = t_base_object;

	if(DEBUG_OUTPUT) indent(level);

	if( (cnt=getNextToken(token, &objType, true, true, parsingpostscript)) == -1) { 
		return NULL;
	}

	if(objType == t_dictionary && strcmp(token, "<<") == 0) {
		if(DEBUG_OUTPUT) printf("Start Dictionary %s\n", token);
		result = new PDF_dictionary();

		while( (cnt=getNextToken(token, &objType, true, false, parsingpostscript)) != -1 && strcmp(token, ">>") != 0) {
			temp = getNextObject(level+1, parsingpostscript);
			if(temp != NULL) 
				result->addsub(temp);
		}
		if( (cnt=getNextToken(token, &objType, true, false, parsingpostscript)) != -1 && strcmp(token, ">>") == 0) {
			cnt=getNextToken(token, &objType, true, true, parsingpostscript);

			if(DEBUG_OUTPUT) indent(level);
			if(DEBUG_OUTPUT) printf("End Dictionary %s\n", token);
			cnt=getNextToken(token, &objType, true, false, parsingpostscript);
			if(strcmp(token, "stream") == 0) {
				cnt=getNextToken(token, &objType, true, true, parsingpostscript);

				PDF_blob *blob = new PDF_blob();
				String data("");

				if(DEBUG_OUTPUT) indent(level);
				if(DEBUG_OUTPUT) printf("Found Stream\n");

				while( (cnt=getNextToken(token, &objType, true, true, parsingpostscript)) != -1 && strcmp(token, "endstream") != 0) {
					// if(DEBUG_OUTPUT) indent(level+1);
					// if(DEBUG_OUTPUT) printf("STREAM: %s\n", token);
					data.append(token, cnt);
				} // Illogical.  "endstream" can appear in a stream, and your code would choke on it.
					// You must examine the length attribute in the dictionary preceding the blob, rather
					// than trying to use endstream as a terminator.

				if(DEBUG_OUTPUT) indent(level);
				if(DEBUG_OUTPUT) printf("End Stream\n");

				if(! data.isBinary()) // See my comments in definition of isBinary(); this function is bad.
					data = DataFilter::convertEOLs(data);

				blob->setValue(data);

				if(blob->length > 1 && 
				   blob->rawdata[blob->length - 2] == '\r' &&
				   blob->rawdata[blob->length - 1] == '\n') {
					blob->length -= 2;
				}
				else if(blob->rawdata[blob->length - 1] == '\r') {
					blob->length--;
				}
				else if(blob->rawdata[blob->length - 1] == '\n') {
					blob->length--;
				}
				else {
					printerr("Error - no final EOL found for stream\n");
				}

				result = new PDF_stream( (PDF_dictionary*) result, blob);
			}
		}
		else {
			printerr("Error terminating dictionary - %s\n", token);
		}
	}
	else if(objType == t_dictionary && strcmp(token, ">>") == 0) {
		printerr("Error - Found unexpected End Dictionary\n");
	}
	else if(objType == t_array && strcmp(token, "[") == 0) {
		if(DEBUG_OUTPUT) printf("Start Array - %s\n", token);
		result = new PDF_array();

		while( (cnt=getNextToken(token, &objType, true, false, parsingpostscript)) != -1 && strcmp(token, "]") != 0) {
			temp = getNextObject(level+1, parsingpostscript);
			if(temp != NULL)
				result->addsub(temp);
		}
		if( (cnt=getNextToken(token, &objType, true, false, parsingpostscript)) != -1 && strcmp(token, "]") == 0) {
			cnt=getNextToken(token, &objType, true, true, parsingpostscript);
			
			if(DEBUG_OUTPUT) indent(level);
			if(DEBUG_OUTPUT) printf("End Array %s\n", token);
		}
		else {
			printerr("Error terminating array - %s\n", token);
		}
	}
	else if(objType == t_array && strcmp(token, "]") == 0) {
		printerr("Error - Found unexpected End Array\n");
	}
	else if(objType == t_name) {
		if(DEBUG_OUTPUT) printf("Name: %s\n", token);
		result = new PDF_name();
		result->setValue(token+1);
	}
	else if(objType == t_string && *token == '(') {
		if(DEBUG_OUTPUT) printf("String: %s\n", token);
		result = new PDF_string();

		String temp;
		while(token[pdfc_strlen(token)-1] == '\\') {
			token[pdfc_strlen(token)-1] = 0;
			temp.append(token);

			if( (cnt=getNextToken(token, &objType, true, true, parsingpostscript)) == -1) { 
				printerr("Error - input ended before end of string\n");
				break;
			}
			if(objType != t_string)
				printerr("Error - partial string is wrong type %d\n", objType);
		}

		if(token[pdfc_strlen(token)-1] == ')') {
			temp.append(token);
		}
		else {
			printerr("Error - nonterminated PDF string - '%s' on line %d\n", token, file->linesRead);
		}

		if(temp.getlength() > 1) {
			if(DEBUG_LOG_NULLTERM) printf("nullterm() #1.\n");
			char *str = temp.nullterm();

			str[pdfc_strlen(str)-1] = 0;
			result->setValue(DataFilter::decodePDF(String(str+1)));
			delete [] str;
		}
		else {
			printerr("Error - Illegal String\n");
		}
	}
	else if(objType == t_hexstring && *token == '<') {
		if(DEBUG_OUTPUT) printf("HexString: %s\n", token);
		result = new PDF_hexstring();

		token[pdfc_strlen(token)-1] = 0;

		result->setValue(token+1);
	}
	else if(objType == t_null) {
		if(DEBUG_OUTPUT) printf("Null: %s\n", token);
		result = new PDF_null();
	}
	else if(objType == t_boolean) {
		if(DEBUG_OUTPUT) printf("Boolean: %s\n", token);
		result = new PDF_boolean();
		result->setValue(token);
	}
	else if(objType == t_reference && (list = regEx.matchList(token, "^([0-9]+) +([0-9]+) R$")) != NULL) {
		PDF_ref *ref = new PDF_ref();
		result = ref;
		if(DEBUG_OUTPUT) printf("Reference: %s\n", token);

		ref->ref = atoi(list[1]);
		ref->generation = atoi(list[2]);
		regEx.freeMatches(list);
	}
	else if(objType == t_real) {
		if(DEBUG_OUTPUT) printf("Real: %s\n", token);
		result = new PDF_real();
		result->setValue(token);
	}
	else if(objType == t_integer) {
		if(DEBUG_OUTPUT) printf("Int: %s\n", token);
		result = new PDF_int();
		result->setValue(token);
	}
	else if(objType == t_pscommand && parsingpostscript) {
		if(DEBUG_OUTPUT) printf("Postscript command: '%s'\n", token);
		result = new PDF_pscommand();
		result->setValue(token);
	}
	else if(objType == t_inline_image && parsingpostscript) {
		if(strcmp(token, "ID") == 0) {
			PDF_inline_image *blob = new PDF_inline_image();
			String data;

			if(DEBUG_OUTPUT) indent(level);
			if(DEBUG_OUTPUT) printf("Found Inline Image\n");

			while( (cnt=getNextToken(token, &objType, true, true, parsingpostscript)) != -1 && strcmp(token, "EI") != 0) {
				// if(DEBUG_OUTPUT) indent(level+1);
				// if(DEBUG_OUTPUT) printf("INLINE IMAGE: %s\n", token);
				data.append(token, cnt);
			} 
			if(cnt == -1)
				printerr("Error - Could not find end of Inline Image\n");

			if(DEBUG_OUTPUT) indent(level);
			if(DEBUG_OUTPUT) printf("End InlineImage\n");

			if(! data.isBinary()) // See my comments in definition of isBinary(); this function is bad.
				data = DataFilter::convertEOLs(data);

			blob->setValue(data);

			result = blob;
		}
		else {
			printerr("Error - Bad start of inline image\n");
		}
	}
	else {
		if(objType == t_base_object) 
			printerr("Unknown base object Token: %s - line %d\n", token, file->linesRead); 
		else
			printerr("Unknown Token: %s - line %d\n", token, file->linesRead); 
	}

	return result;
}

//------------------------------------

int PDFTokenizer::getNextToken(char *token, ObjectType *objType, bool discardEOL, bool advanceToken, bool parsingpostscript) {
	int cnt;
	char *tokenPtr = token;
	char **list;
	RegExUtil regEx;
	
	// printf("Getting Token\n");
	*objType = t_base_object;
	*token = '\0';

	if(nextToken[0] != 0) {
		if(TOKEN_DEBUG) printf("Cached token: %s\n", nextToken);
		*objType = lastType;
		strcpy(token, nextToken);
		if(advanceToken)
			nextToken[0] = 0;
		return pdfc_strlen(token);
	}

	if(inStream) {
		// if(TOKEN_DEBUG) printf("inStream\n");
		*objType = t_stream;
		cnt = file->getLine(token, MAX_LINE_LENGTH, true);
		if(cnt == -1) {
			*token = 0;
			return -1;
		}

		// printf("Cnt: %d\n", cnt);
		if(cnt < 1) {
			printerr("ERROR: could not find end of stream - %d\n", cnt);
		}
 		else if(strncmp(token, "endstream", pdfc_strlen("endstream")) == 0 && 
               (token[pdfc_strlen("endstream")] == '\n' || token[pdfc_strlen("endstream")] == '\r')) {
			cnt = pdfc_strlen("endstream");
			token[cnt] = 0;
			inStream = false;
			partialToken = false;
		}

		nextChar = buffer;
		bufferEnd = buffer;

		return cnt;
	}

	if(inInlineImage) {
		// if(TOKEN_DEBUG) printf("inInlineImage\n");
		*objType = t_inline_image;
		cnt = file->getLine(token, MAX_LINE_LENGTH, true);
		if(cnt == -1) {
			*token = 0;
			return -1;
		}

		// printf("Cnt: %d\n", cnt);
		if(cnt < 1) {
			printerr("ERROR: could not find end of inline image - %d\n", cnt);
		}
 		else if(strncmp(token, "EI", pdfc_strlen("EI")) == 0 && 
               (token[pdfc_strlen("EI")] == '\n' || token[pdfc_strlen("EI")] == '\r')) {
			cnt = pdfc_strlen("EI");
			token[cnt] = 0;
			inInlineImage = false;
			partialToken = false;
		}

		nextChar = buffer;
		bufferEnd = buffer;

		return cnt;
	}

	if(jumpWhiteSpaceAndComments() == -1) {
		*token = 0;
		return -1;
	}

	*bufferEnd = 0;

	cnt = 0;
	if(! partialToken ) {
		if( (*nextChar >= 'a' && *nextChar <= 'z') || (*nextChar >= 'A' && *nextChar <= 'Z') ) {
			if(strncmp(nextChar, "stream", pdfc_strlen("stream")) == 0) {
				inStream = true;
				partialToken = true;
				cnt = pdfc_strlen("stream");
				if(TOKEN_DEBUG) printf("Stream\n");
				*objType = t_stream;
			}
			else if(strncmp(nextChar, "null", pdfc_strlen("null")) == 0) {
				cnt = pdfc_strlen("null");
				if(TOKEN_DEBUG) printf("Null\n");
				*objType = t_null;
			}
			else if(strncmp(nextChar, "true", pdfc_strlen("true")) == 0) {
				cnt = pdfc_strlen("true");
				if(TOKEN_DEBUG) printf("Boolean: true\n");
				*objType = t_boolean;
			}
			else if(strncmp(nextChar, "false", pdfc_strlen("false")) == 0) {
				cnt = pdfc_strlen("false");
				if(TOKEN_DEBUG) printf("Boolean: false\n");
				*objType = t_boolean;
			}
			else if(strncmp(nextChar, "endobj", pdfc_strlen("endobj")) == 0) {
				cnt = pdfc_strlen("endobj");
				if(TOKEN_DEBUG) printf("endobj\n");
			}
			else {
				while( (nextChar+cnt < bufferEnd) && 
				       (nextChar[cnt] >= 'a' && nextChar[cnt] <= 'z') || 
				       (nextChar[cnt] >= 'A' && nextChar[cnt] <= 'Z') || 
				       (nextChar[cnt] == '*')) {
					cnt++;
				}
				if(TOKEN_DEBUG) printf("Postscript command: %s\n", token);

				if(parsingpostscript && strncmp(nextChar, "ID", 2) == 0) {
					// printf("Found image data\n");
					inInlineImage = true;
					*objType = t_inline_image;
				}
				else {
					*objType = t_pscommand;
				}
			}

			if(cnt > 0) {
				strncpy(token, nextChar, cnt);
				token[cnt] = '\0';
				nextChar += cnt;
			}
		}
		else if( (*nextChar >= '0' && *nextChar <= '9') || *nextChar == '-') {
			if((list = regEx.matchList(nextChar, "^[0-9]+ +[0-9]+ obj")) != NULL) {
				strcpy(token, list[0]);
				cnt = pdfc_strlen(token);
				nextChar += cnt;
				if(TOKEN_DEBUG) printf("Indirect Object: %s\n", token);
				regEx.freeMatches(list);
			}
			else if((list = regEx.matchList(nextChar, "^[0-9]+ +[0-9]+ R")) != NULL) {
				strcpy(token, list[0]);
				cnt = pdfc_strlen(token);
				nextChar += cnt;
				if(TOKEN_DEBUG) printf("Reference: %s\n", token);
				regEx.freeMatches(list);
				*objType = t_reference;
			}
			else if((list = regEx.matchList(nextChar, "^-?[0-9]+\\.[0-9]+")) != NULL) {
				strcpy(token, list[0]);
				cnt = pdfc_strlen(token);
				nextChar += cnt;
				if(TOKEN_DEBUG) printf("Real: %s\n", token);
				regEx.freeMatches(list);
				*objType = t_real;
			}
			else if((list = regEx.matchList(nextChar, "^-?[0-9]+")) != NULL) {
				strcpy(token, list[0]);
				cnt = pdfc_strlen(token);
				nextChar += cnt;
				if(TOKEN_DEBUG) printf("Int: %s\n", token);
				regEx.freeMatches(list);
				*objType = t_integer;
			}
		}
		else if(*nextChar == '/') {
			if((list = regEx.matchList(nextChar, "^/[!\"$&'*+,-.0-9:;=?@A-Z^_`a-z|~#]+")) != NULL) {
				*objType = t_name;

				strcpy(token, list[0]);
				cnt = pdfc_strlen(token);
				nextChar += cnt;
				if(TOKEN_DEBUG) printf("Name: %s\n", token);
				regEx.freeMatches(list);
			}
		}
		else {
			if(strncmp(nextChar, "<<", 2) == 0) {
				if(TOKEN_DEBUG) printf("Start Dictionary\n");
				cnt = 2;
				*objType = t_dictionary;
			}
			else if(strncmp(nextChar, ">>", 2) == 0) {
				if(TOKEN_DEBUG) printf("End Dictionary\n");
				cnt = 2;
				*objType = t_dictionary;
			}
			else if(*nextChar == '[') {
				if(TOKEN_DEBUG) printf("Start Array\n");
				cnt = 1;
				*objType = t_array;
			}
			else if(*nextChar == ']') {
				if(TOKEN_DEBUG) printf("End Array\n");
				cnt = 1;
				*objType = t_array;
			}
			else if(*nextChar == '\"') {
				if(TOKEN_DEBUG) printf("Postscript command: \"\n");
				cnt = 1;
				*objType = t_pscommand;
			}
			else if(*nextChar == '\'') {
				if(TOKEN_DEBUG) printf("Postscript command: \'\n");
				cnt = 1;
				*objType = t_pscommand;
			}

			if(cnt > 0) {
				strncpy(token, nextChar, cnt);
				token[cnt] = '\0';
				nextChar += cnt;
			}
		}
	}

	if(cnt > 0) {
		// Token already found
	}
	else if(inString || *nextChar == '(') {
		char prevChar = '\0';
		int startLine = file->linesRead;

		*objType = t_string;

		while(prevChar == '\\' || *nextChar != ')') {
			*tokenPtr = *nextChar;

			if(*nextChar == '\\' && prevChar == '\\') {
				// printf("Found Double Slash\n");
				prevChar = '\0';
			}
			else {
				prevChar = *nextChar;
			}

			tokenPtr++;
			nextChar++;
			cnt++;

			if(*nextChar == '\n' || nextChar == bufferEnd) {
				if(prevChar == '\\') {
					inString = true;
					partialToken = true;
				}
				else {
					printerr("Error - end of string starting on line %d not found - prevChar = '%c'\n", startLine, prevChar);
				}
				*tokenPtr = 0;
				
				if(! advanceToken) {
					strcpy(nextToken, token);
					lastType = *objType;
				}
				return cnt; 
			}
			else if(cnt == MAX_LINE_LENGTH-1) {
				*tokenPtr++ = '\\';
				*tokenPtr = 0;
				inString = true;
				partialToken = true;

				if(! advanceToken) {
					strcpy(nextToken, token);
					lastType = *objType;
				}
				return cnt; 
			}
		}
		*tokenPtr++ = *nextChar++;
		cnt++;
		*tokenPtr = '\0';
		inString = false;
		partialToken = false;

		if(TOKEN_DEBUG) printf("String: %s\n", token);
	}
	else if(inHexString || *nextChar == '<') {
		*objType = t_hexstring;

		while(*nextChar != '>') {
			*tokenPtr++ = *nextChar++;
			cnt++;
		}
		*tokenPtr++ = *nextChar++;
		cnt++;
		*tokenPtr = '\0';

		if(TOKEN_DEBUG) printf("HexString: %s\n", token);
		inHexString = false;
		partialToken = false;
	}
	else {
		cnt = 0;
		while(nextChar < bufferEnd && *nextChar != ' ' && *nextChar != '\n') {
			*tokenPtr++ = *nextChar++;
			cnt++;
		}
		*tokenPtr = '\0';
		if(*token == '\0') {
			// Found Empty line
			printerr("Error - empty token on line %d\n", file->linesRead);
		}
		else {
			printerr("Unknown token '%s' - line %d\n", token, file->linesRead);
		}
	}

	if(! advanceToken) {
		strcpy(nextToken, token);
		lastType = *objType;
	}

	return cnt;
}
//------------------------------------

int PDFTokenizer::jumpWhiteSpaceAndComments() {
	int amt;
	char *ch;
	int numLeft;

	// if(TOKEN_DEBUG) printf("jumpWhiteSpaceAndComments()\n");
	while(nextChar == bufferEnd) {
		amt = file->getLine(buffer, MAX_LINE_LENGTH);
		if(amt == -1) {
			return -1;
		}
		nextChar = buffer;
		bufferEnd = buffer+amt;
	}

	while(*nextChar == ' ' || *nextChar == '\n' || *nextChar == '%') {
		if(*nextChar == '%') {
			// Comment runs til end of line

			// if(TOKEN_DEBUG) printf("Comment %s\n", nextChar); 

			amt = file->getLine(buffer, MAX_LINE_LENGTH);
			if(amt == -1) {
				return -1;
			}
			nextChar = buffer;
			bufferEnd = buffer+amt;
		}

		nextChar++;

		while(nextChar == bufferEnd) {
			amt = file->getLine(buffer, MAX_LINE_LENGTH);
			if(amt == -1) {
				return -1;
			}
			nextChar = buffer;
			bufferEnd = buffer+amt;
		}
	}

	// ALSO FIX REF AND OBJ LOOK AHEADS !!!
	ch = nextChar;
	while(ch < bufferEnd && *ch != ' ' && *ch != '\n') {
		ch++;
	}
	if(ch == bufferEnd) {
		numLeft = bufferEnd-nextChar;
		memmove(buffer, nextChar, numLeft);
		amt = file->getLine(buffer+numLeft, MAX_LINE_LENGTH-numLeft); 
		if(amt == -1) {
			nextChar = buffer;
			bufferEnd = buffer+numLeft;
			return 0;
		}

		nextChar = buffer;
		bufferEnd = buffer+amt+numLeft;
// printf("Get More Data: '");
// fwrite(nextChar, 1, amt+numLeft, stdout);
// printf("'\n");
	}
	else {
		char **list;
		RegExUtil regEx;

		if(*nextChar >= '0' && *nextChar <= '9' && 
		   (list = regEx.matchList(nextChar, "^[0-9]+ +([0-9]+( +(ob?)?)?)?$")) != NULL) {
			regEx.freeMatches(list);

			numLeft = bufferEnd-nextChar;
			memmove(buffer, nextChar, numLeft);
			amt = file->getLine(buffer+numLeft, MAX_LINE_LENGTH-numLeft); 
			if(amt == -1) {
				nextChar = buffer;
				bufferEnd = buffer+numLeft;
				return 0;
			}

			nextChar = buffer;
			bufferEnd = buffer+amt+numLeft;
		}
	}

	return 0;
}

//------------------------------------

void PDFTokenizer::indent(int level) {
	int i;

	for(i=0; i<level; i++)
		putchar(' ');
}

//------------------------------------

uint number_of_arguments(PObj* o) { // Assumes o contains a string.  If the string is a postscript command,
	// returns the number of arguments the command takes.  Otherwise, returns zero.
	if(o->get_type()!=t_pscommand) return 0; // Only pscommands have arguments.
	int r=get_number_of_args_primitive(((PDF_pscommand*)(o))->value);
	if(r==-1) {
		printerr("Unrecognized postscript command: '%s'\n", ((PDF_pscommand*)(o))->value.nullterm());
	}
	return r;
}

//------------------------------------

void convert_realname_to_alias(PObj* o) {
	if(o->get_type()!=t_pscommand) return;
	((PDF_pscommand*)(o))->value=(get_alias_primitive(((PDF_pscommand*)(o))->value));
}

//------------------------------------



/* PDF doesn't allow commands to push values onto the stack.  Therefore the only arguments to a command in a
page description are exactly those values which appear between the command in question and the previous command.
I'm taking advantage of this fact in the following function (in other words, the function would
fail miserably when processing genuine postscript).
*/
PObj* reorder_pscommands(PObj* o) { // If this function is NOP, commands get spit out in native (postfix)
	// sequence.  This function converts them to prefix and nests the arguments properly.
	// This function merely sets up the object hierarchy right.  The standard PObj::render_xml() function
	// actually renders the hierarchy into XML.
	int numsubs=0, c;
	PObjList* p=o->subobjects;
	while(p!=NULL) { // Count subobjects.
		p=p->next;
		numsubs++;
	}
	PObj** thesubs=new PObj*[numsubs];
	p=o->subobjects;
	for(c=0; c<numsubs; c++) { // Put pointers to all subobjects into an array, and convert names.
		convert_realname_to_alias(p->theobject);
		thesubs[c]=p->theobject;
		p->theobject=NULL; // Remove the PObjList object's reference so that when deleted, it doesn't delete the Pobj object.
		p=p->next;
	}
	delete o->subobjects;
	o->subobjects=NULL; // Remove the original subobjects listing from o, now that it's been copied to an array.
	for(c=0; c<numsubs; c++) { // Process all elements in the thesubs array.
		int d=c;
		if(DEBUG) setbuf(stdout, NULL);
		if(DEBUG) printf("First: c=%d d=%d\n", c, d);
		while(thesubs[d]->get_type()!=t_pscommand) { // Find the next command.
			if(DEBUG) printf("Search loop (before increment): c=%d d=%d\n", c, d);
			d++;
		}
		for(int e=c; e<d; e++) {
			thesubs[d]->addsub(thesubs[e]);
			thesubs[e]=NULL;
		}
		c=d;
	}
	for(c=0; c<numsubs; c++) if(thesubs[c]!=NULL) o->addsub(thesubs[c]); // Copy pointers from array back to o's subobjects listing, ignoring the ones marked as deleted.
	delete [] thesubs; // Temporary array is no longer needed.
	return o;
}



//------------------------------------

PObj* parse_page_description_to_XML(String contentData) {

	// printf("%s\n", contentData.nullterm());

	DataReader* f=new StringReader(contentData.getrawstring(), contentData.getlength());

	PDFTokenizer t(f);
	PObj* o=t.parseContents(true);

	o=reorder_pscommands(o);

	// printf("%s\n", o->render_xml(0).nullterm());

	return o;
}

//------------------------------------

PObj* PDFTokenizer::parseContents(bool parsingpostscript) {
	PObj *collection = new PDF_collection();
	PObj *temp = NULL;

	while( (temp = getNextObject(0, parsingpostscript)) != NULL) {
		collection->addsub(temp);
	}

	return collection;
}

