#include "XMLModule_lowLevel.hpp"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <iostream.h>

#include "PdfObjects.hpp"
#include "pdfc_basicobjects.h"
#include "stringclass.h"

XMLModule_lowLevel::XMLModule_lowLevel() : parsingPostscript(false), root(NULL), currentObj(NULL) {
    DEBUG_OUTPUT = false;
}

XMLModule_lowLevel::XMLModule_lowLevel(bool willParsePostscript) {
    DEBUG_OUTPUT = false;
	if(willParsePostscript) {
		if(DEBUG_OUTPUT) printf("Using Postscript Option\n");
		parsingPostscript = true;
		root = new PDF_collection();
		currentObj = root;
	}
	else {
		parsingPostscript = false;
		root = NULL;
		currentObj = NULL;
	}
}

PObj* XMLModule_lowLevel::getRoot() {
	return root;
}

void XMLModule_lowLevel::reset(bool deleteTree) {
	if(DEBUG_OUTPUT) printf("Resetting XMLModule\n");
	
	if(deleteTree && root != NULL)
		delete root;

	if(parsingPostscript) {
		root = new PDF_collection();
		currentObj = root;
	}
	else {
		root = NULL;
		currentObj = NULL;
	}
}

void XMLModule_lowLevel::enablePostscriptParsing() {
	if(DEBUG_OUTPUT) printf("Using Postscript Option\n");
	reset(true);
	
	parsingPostscript = true;

	root = new PDF_collection();
	currentObj = root;
} 

//================================================

void XMLModule_lowLevel::startElement(XMLElement** stack, int level) {
	PObj *pdfObject = NULL;
	char *type;

	// printf("Getting Basic XML Element\n");
	type = stack[level]->name;

	if(strcmp(type, "blob") == 0 || strcmp(type, "inline_image") == 0) {
    	PDF_blob *blob = NULL;

		if(strcmp(type, "blob") == 0) 
			blob = new PDF_blob();
		else if(strcmp(type, "inline_image") == 0) 
			blob = new PDF_inline_image();
		else
			printerr("Error creating blob\n");

		char *value;
		value = stack[level]->getAttribute("filename");
		if(value != NULL) {
			String data;
			char buffer[512];
			int cnt;
			FILE *blobFp = fopen(value, "rb");

			if(blobFp == NULL) {
				printerr("Error opening file: %s\n", value);
			}
			else {
				while( (cnt = fread(buffer, 1, 512, blobFp)) == 512) {
					data.append(buffer, cnt);
				}
				data.append(buffer, cnt);

				if(fclose(blobFp) != 0) {
					printerr("Error closing file: %s\n", value);
				}
				blob->setValue(data);
			}
		}
		else {
			value = stack[level]->getAttribute("encoding");
			if(value != NULL) {
				stack[level]->encoding = new char[pdfc_strlen(value)+1];
				strcpy(stack[level]->encoding, value);
			}
		}
    	pdfObject = blob;
	}
	else if(strcmp(type, "pdf") == 0) {
    	pdfObject = new PDF_root();
	}
	else if(strcmp(type, "null") == 0) {
    	pdfObject = new PDF_null();
	}
	else if(strcmp(type, "boolean") == 0) {
    	pdfObject = new PDF_boolean();
	}
	else if(strcmp(type, "int") == 0) {
    	pdfObject = new PDF_int();
	}
	else if(strcmp(type, "real") == 0) {
    	pdfObject = new PDF_real();
	}
	else if(strcmp(type, "string") == 0) {
		// This case basically copied verbatim from "blob" case above.
    	PDF_string *blob = new PDF_string();
		char *value;
		value = stack[level]->getAttribute("filename");
		if(value != NULL) {
			String data;
			char buffer[512];
			int cnt;
			FILE *blobFp = fopen(value, "rb");

			if(blobFp == NULL) {
				printerr("Error opening file: %s\n", value);
			}
			else {
				while( (cnt = fread(buffer, 1, 512, blobFp)) == 512) {
					data.append(buffer, cnt);
				}
				data.append(buffer, cnt);

				if(fclose(blobFp) != 0) {
					printerr("Error closing file: %s\n", value);
				}
				blob->setValue(data);
			}
		}
    	pdfObject = blob;
	}
	else if(strcmp(type, "hexstring") == 0) {
    	pdfObject = new PDF_hexstring();
	}
	else if(strcmp(type, "name") == 0) {
    	pdfObject = new PDF_name();
	}
	else if(strcmp(type, "array") == 0) {
    	pdfObject = new PDF_array();
	}
	else if(strcmp(type, "dictionary") == 0) {
    	pdfObject = new PDF_dictionary();
	}
	else if(strcmp(type, "stream") == 0) {
    	pdfObject = new PDF_stream(NULL, NULL);
	}
	else if(strcmp(type, "reference") == 0) {
		PDF_ref *ref = new PDF_ref();
		char *value;
		value = stack[level]->getAttribute("ref_id");
		ref->ref = atoi(value);
		value = stack[level]->getAttribute("generation");
		ref->generation = atoi(value);

    	pdfObject = ref;
	}
	else if(strcmp(type, "trailer") == 0) {
		pdfObject = new PDF_trailer();
	}
	else if(strcmp(type, "deletion") == 0) {
    	pdfObject = new PDF_deletion();
	}
	else if(strcmp(type, "box") == 0) {
		PDF_box* box=new PDF_box();
    	pdfObject=box;
		box->x=atoi(stack[level]->getAttribute("x"));
		box->y=atoi(stack[level]->getAttribute("y"));
		box->width=atoi(stack[level]->getAttribute("width"));
		box->height=atoi(stack[level]->getAttribute("height"));
	}
	else {
		if(parsingPostscript) {
			bool valid = name_isvalid_primitive(type)||name_isvalid_complex(type);
			if(valid) {
    			PDF_pscommand *command = new PDF_pscommand();
    			pdfObject = command;

				command->value = type; // Use the XML element name as the comand name.

				if(DEBUG_OUTPUT) indent(level);
				if(DEBUG_OUTPUT) printf("Found Command %s\n", type);
			}
			else {
				printerr("\nERROR - Unknown XML pscommand: %s\n", type);
			}
		}
		else {
			printerr("\nERROR - Unknown XML type: %s\n", type);
		}
	}
	
	if(pdfObject != NULL) {
		if(pdfObject->get_type() == t_root) {
			if(currentObj != NULL) {
				printerr("\nERROR - More that one PDF root object\n");
			}
			else {
				currentObj = pdfObject;
				root = currentObj;
	
				if(DEBUG_OUTPUT) indent(level);
				if(DEBUG_OUTPUT) printf("Initializing %s as root node\n", currentObj->get_printable_type() );
			}
		}
		else if(currentObj == NULL) {
			printerr("\nERROR - PObj found outside the PDF root object - %s\n", type);
		}
		else {
			PObj *temp = pdfObject;

			if(pdfObject->get_type() == t_dictionary &&
			        currentObj->get_type() == t_trailer) {
				((PDF_trailer*) currentObj)->dictionary = (PDF_dictionary*) temp;
			}
			else if(pdfObject->get_type() == t_dictionary &&
			        currentObj->get_type() == t_stream) {
				((PDF_stream*) currentObj)->dictionary = (PDF_dictionary*) temp;
			}
			else if(pdfObject->get_type() == t_blob &&
			        currentObj->get_type() == t_stream) {
				((PDF_stream*) currentObj)->dataObject = (PDF_blob*) temp;
			}

			currentObj->addsub(temp);
			currentObj= temp;

			if(DEBUG_OUTPUT) indent(level);
			if(DEBUG_OUTPUT) printf("Adding %s to %s\n", currentObj->get_printable_type() , currentObj->parentp->get_printable_type() );
		}

		char *value = stack[level]->getAttribute("id");
		if(value != NULL) {
			pdfObject->id = atoi(value);
    		((PDF_root*) root)->idHash.add( currentObj );
		}

		value = stack[level]->getAttribute("generation");
		if(value != NULL) {
			pdfObject->generation = atoi(value);
		}
	}
}

//-----------------------------------

void XMLModule_lowLevel::endElement(XMLElement** stack, int level) {
	if(stack[level]->cdata.getlength() > 0) {
		if(DEBUG_OUTPUT) indent(level);
		if(DEBUG_OUTPUT) printf("Data: %s\n", stack[level]->cdata.nullterm());
		
		if(currentObj->get_type() == t_blob) { // Overrides data input with filename attribute, if any.
			String output = stack[level]->cdata;

			if(stack[level]->encoding == NULL) {
			}
			else if(strcmp(stack[level]->encoding, "hex") == 0) {
				output = DataFilter::decodeHex(output);
			}
			else if(strcmp(stack[level]->encoding, "base85") == 0) {
				output = DataFilter::decode85(output);
			}
			else {
				printerr("Error - unknown encoding type\n");
			}

			currentObj->setValue(output);
		}
		else if(currentObj->get_type() == t_stream && currentObj->id == 0) {
			printerr("Error - each stream must have an associated object id\n");
		}
		else if(currentObj->get_type()==t_pscommand) {
			PDF_pscommand* command=(PDF_pscommand*)(currentObj);
			uint len=stack[level]->cdata.getlength();
			command->contentstring=new char[len+1];
			memcpy(command->contentstring, stack[level]->cdata.getrawstring(), len); // Grab the XML element content.
			command->contentstring[len] = 0;
		}
		else if(currentObj->get_type() == t_name) {
			
			currentObj->setValue(stack[level]->cdata);
			((PDF_name*)currentObj)->trimWhiteSpace();

			if( ! ((PDF_name*) currentObj)->isValidName() ) 
				printerr("'%s' is not a valid name\n", ((PDF_name*)currentObj)->value.nullterm());
		}
		else {
			currentObj->setValue(stack[level]->cdata);
		}
	}

	if(currentObj->get_type() == t_trailer) {
		((PDF_root*) root)->updateTrailer( (PDF_trailer*) currentObj );	

		PDF_dictionary *dict = ((PDF_trailer*) currentObj)->dictionary;
		if(dict != NULL && dict->get_type() == t_dictionary) {
			PObj *rootRef = dict->getValueForKey("Root");

			if(rootRef != NULL) {
				if(rootRef->get_type() != t_reference) {
					printerr("Error - invalid root reference object type - %s\n", rootRef->get_printable_type());
				}
				else {
					// printf("Found Root Reference\n");
					((PDF_root*) root)->rootNode = rootRef;
				}
			}
		}
	}

	if(currentObj->parentp != NULL)
		currentObj = currentObj->parentp;
}

