#include "pdfc_base.h"
#include "pdfc_basicobjects.h"

char* PDF_FILE_HEADER="%PDF-1.2\n%\n";
bool DllExport decompose=false; // Indicates whether blob xml output should decompose streams into separate files.
bool DllExport parseObjects=false; // Indicates if the document structure should pe parsed (2nd pass parsing)
bool DllExport parseHighLevelXML=true; // Indicates if the XML read in is a high level representation
bool DllExport renderHighLevelXML=true; // Indicates if the XML written out is a high level representation

int Xreftable::MAX_TABLE_SIZE = 65536;

uint gethashvalue(uint id) {
	return id%HASHTABLESIZE; // Keep hash value within 0 to HASHTABLESIZE.
}

Xrefentry::Xrefentry(): objnumber(0), gennumber(0), offset(0), totalsize(0) {}

Xreftable::Xreftable(): arraysize(0) {
	entry=new Xrefentry[MAX_TABLE_SIZE];
}

Xreftable::~Xreftable() {
	delete [] entry;
}

void Xreftable::accobjsize(int n) {
	entry[arraysize].totalsize+=n;
}

//-----------------------------------------------

PObjList::PObjList(): theobject(NULL), next(NULL), deleteObjectOnDestruct(true) {}

PObjList::PObjList(bool deleteObjectFlag): theobject(NULL), next(NULL), deleteObjectOnDestruct(deleteObjectFlag) {}

PObjList::~PObjList() { // Delete the list and the instances of PObj which it lists.
	if(theobject != NULL && deleteObjectOnDestruct) {
		if( (theobject->get_type() == t_collection || 
			 theobject->get_type() == t_shared_collection) &&
			((PDF_collection*)theobject)->sharedStream ) {
		}
		else {
			delete theobject;
		}
	}
	if(next) delete next;
}

//-----------------------------------------------

PDF_collection::PDF_collection() : PObj(), sharedStream(false), streamAlreadyRendered(false) {
}

PObj* PDF_collection::copy() {
    PDF_collection *beta = new PDF_collection();
 
    beta->copyFields(this);
    return beta;
}
 
ObjectType PDF_collection::get_type() {return t_collection;}
char* PDF_collection::get_printable_type() {return "collection";}

void PDF_collection::render_xml(int depth, String& dest) {
	render_xml_children(depth, dest);
}

void PDF_collection::render_xml_start(int depth, String& dest) {
	render_xml(depth, dest);
}

void PDF_collection::render_xml_end(int depth, String& dest) {
	// Intentionally NOP.
}

void PDF_collection::render_pdf(Xreftable* x, String& dest) {
	// x = render_pdf_top_level_start(x);
	render_pdf_children(x, dest);
	// render_pdf_top_level_end(x);
}

//-----------------------------------------------

Hashtable::Hashtable() {
}

Hashtable::~Hashtable() {
}

void Hashtable::add(PObj* newobject) {
	if(newobject != NULL)
		h.add(newobject->id, newobject);
}

void Hashtable::add(PObj* newobject, uint id) {
	if(newobject != NULL)
		h.add(id, newobject);
}

void Hashtable::remove(uint oldid) {
	h.remove(oldid);
}

PObj* Hashtable::findobject(uint id) {
    PObj** temp=h.getValue(id);
    if(temp==NULL) return NULL;
    else return *temp;
}

//-----------------------------------------------

PObj* PObj::copy() {
    PObj *beta = new PObj();
 
    beta->copyFields(this);
    return beta;
}
 
void PObj::copyFields(PObj *alpha) {
	PObj *temp;

    id = alpha->id;
    generation = alpha->generation;
 
    unique_id = alpha->unique_id;

    PObjList *listPtr = alpha->subobjects;
    while(listPtr != NULL) {
		if(listPtr->theobject == NULL)
			printerr("Error - bad link to PObj subobject\n");
		temp = listPtr->theobject->copy();
		if(temp == NULL)
			printerr("Error copying POBj\n");
        addsub(temp);
        listPtr = listPtr->next;
    }
}

uint PObj::gettotalsize() const {
	return getlength();
}

void PObj::addsub(PObj* newsub, bool deleteObjectOnDestruct) { // Insert new object at tail of list.
	if(newsub == NULL) 
		return;

	if(newsub->get_type() == t_collection || newsub->get_type() == t_shared_collection) {
		PObjList *element = newsub->subobjects;
		while(element != NULL) {
			addsub(element->theobject);

			element = element->next;
		}
	}
	else {
		// Note that insertion at the head is simpler to code and faster to execute, but we need insertions
		// at the end to preserve subobjects ordering for things like PDF arrays.
		PObjList* newlist=new PObjList(deleteObjectOnDestruct);
		newlist->theobject=newsub;
		newlist->next=NULL;
		PObjList** lpp=&subobjects;
		if(*lpp==NULL) { // No subobjects present.  Adding the first one.
			*lpp=newlist;
			lastsub=newlist;
		} else {
			lastsub->next=newlist;
			lastsub=newlist;
		}
		newsub->parentp=this; // Establish link from subobject to superobject (which is this).
	}
}

void PObj::setParentPage(PdfPage *newPage) {
	parentPage = newPage;
	PObjList *objs = subobjects;
	while(objs != NULL) {
		objs->theobject->setParentPage(newPage);
		objs = objs->next;
	}
}

uint PObj::countsubs() {
	uint c=0;
	PObjList* l=subobjects;
	while(l!=NULL) {
		c+=1;
		l=l->next;
	}
	return c;
}

PdfPage* PObj::getParentPage() {
	PObj* p=parentp;
	while(p!=NULL) {
		if(p->parentp!=NULL) p=p->parentp;
		else break;
	}
	return p->parentPage;
}

//-----------------------------------------------

unsigned long pdfc_ntohl(unsigned long arg) { // Network byte order is big endian.  This function deals with it.
	unsigned long x=1;
	if(*((char*)(&x))==(char)0) return arg; // Big endian machine.
	if(*((char*)(&x))!=(char)1) throw new Err(666); // This should never happen.
	// Little endian machine; deal with it.
	x=arg;
	char* a=(char*)(&x);
	char* b=((char*)(&x))+1;
	char* c=((char*)(&x))+2;
	char* d=((char*)(&x))+3;
	// Swap first and last bytes.
	*a=(*a)^(*d);
	*d=(*a)^(*d);
	*a=(*a)^(*d);
	// Swap middle two bytes.
	*b=(*b)^(*c);
	*c=(*b)^(*c);
	*b=(*b)^(*c);
	return x;
}

unsigned short pdfc_ntohs(unsigned short arg) {
	unsigned short x=1;
	if(*((char*)(&x))==(char)0) return arg; // Big endian machine.
	if(*((char*)(&x))!=(char)1) throw new Err(666); // This should never happen.
	// Little endian machine; deal with it.
	x=arg;
	char* a=(char*)(&x);
	char* b=((char*)(&x))+1;
	// Swap first and last bytes.
	*a=(*a)^(*b);
	*b=(*a)^(*b);
	*a=(*a)^(*b);
	return x;
}

//******************************************************************

PObj::PObj(): id(0), generation(-1), parentp(NULL), subobjects(NULL), lastsub(NULL), parentPage(NULL), xref(NULL) {
	// printf("Creating PObj\n");
}

PObj::~PObj() {
	// printf("Destroying PObj\n");
	if(subobjects) 
		delete subobjects; // Clean up linked lists of subobjects, if it exists.
}

void PObj::showObject() {
	printf("PObj: %s\n", get_printable_type());
}

void PObj::setValue(char *str) {
}

void PObj::setValue(String str) {
	if(DEBUG_LOG_NULLTERM) printf("nullterm() #9.  Culprit is %s\n", this->get_printable_type());

	char *buffer = str.nullterm();
	setValue(buffer);
	delete [] buffer;
}

uint PObj::getlength() const {
	return 0;
}

void PObj::render_xml_start(int depth, String& dest) {
    dest.append("<pdf_object>\n");
}

void PObj::render_xml_children(int depth, String& dest) {
    if(subobjects!=NULL) { // Render each subobject.
        dest.append("\n");
        PObjList* stepper=subobjects;
        while(stepper!=NULL) { // Process each subobject.
            stepper->theobject->render_xml(depth+1, dest);
            stepper=stepper->next;
        }
        for(int c=0; c<depth; c++) { // Prepend depth number of tabs.
            dest.append("\t");
        }
    }
}

void PObj::render_xml_end(int depth, String& dest) {
	dest.append("</pdf_object>\n");
}

void PObj::render_xml(int depth, String& dest) {
	render_xml_start(depth, dest);
	render_xml_children(depth, dest);
	render_xml_end(depth, dest);
}

void PObj::render_pdf(Xreftable* x, String& dest) {
	render_pdf_start(x, dest);

	if(xref != NULL && x == NULL)
		x = xref;
	render_pdf_children(x, dest);
	render_pdf_end(x, dest);
}

void PObj::render_pdf_children(Xreftable* x, String& dest) {
	if(subobjects!=NULL) { // Render each subobject.
		PObjList* stepper=subobjects;
		while(stepper!=NULL) { // Process each subobject.
			stepper->theobject->render_pdf(x, dest);
			stepper=stepper->next;
		}
	}
}

void PObj::render_pdf_start(Xreftable* x, String& dest) { // Wrapper.
	if(id!=0) rend_pdf_ind_begin(id, 0, x, dest); // Using all generation numbers of 0 for now.
}

void PObj::render_pdf_end(Xreftable* x, String& dest) { // Intentionally returns a blank string (all the work is done at the
	// beginning for type PObj).
	if(id!=0) rend_pdf_ind_end(x, dest);
}

void PObj::rend_pdf_ind_begin(uint id, uint gen, Xreftable* x, String& dest) {
	char buf[256];
	if(x->arraysize==0) { // First object is now being rendered, so the PDF file header must be dealt with.
		x->entry[0].offset=pdfc_strlen(PDF_FILE_HEADER);
	} else {
		x->entry[x->arraysize].offset=(x->entry[x->arraysize-1].offset)+(x->entry[x->arraysize-1].totalsize);
	}
	x->entry[x->arraysize].objnumber=id;
	x->entry[x->arraysize].gennumber=gen;
	sprintf(buf, "%d %d obj", id, gen);
	x->accobjsize(pdfc_strlen(buf));
	dest.append(buf);
}

void PObj::rend_pdf_ind_end(Xreftable* x, String& dest) {
	char* retval="\nendobj\n";
	x->accobjsize(pdfc_strlen(retval));
	x->arraysize+=1; // Finished rendering an indirect object, so arraysize is incremented.
	if(x->arraysize >= x->MAX_TABLE_SIZE)
		printerr("Error - Too many entries in Xref table\n");
	dest.append(retval);
}

Xreftable* PObj::render_pdf_top_level_start(Xreftable* x) { // Wrapper.
	// x is passed to render_pdf_ start and end, below.  It is the responsibility of those functions to set the data correctly.
	if(x==NULL) { // Only the first level call to render_pdf() will use the default x value of NULL.  All
		// recursive calls will pass the Xreftable which this if() condition creates.
		x=new Xreftable;
		xref=x;
	}
	return x;
}

void PObj::render_pdf_top_level_end(Xreftable* x) { // Wrapper.
	if(xref!=NULL) delete xref; // This will only occur for the first level call to render_pdf(), not the recurses.
}

ObjectType PObj::get_type() {return t_base_object;}
char* PObj::get_printable_type() {return "PObj";}

