#include "pdfc_basicobjects.h"
#include "stringformatter.h"
#include "thash.h"
#include "CmdList.hpp"

#define DEBUG_FUNCTIONBOUNDARIES false

char* pscommands_complex[NUMBER_OF_COMPLEX_COMMANDS];

#define TOTAL_PRIMITIVE_COMMANDS 74

HashTable<StringLite, Command>* cmd_nameindex;
HashTable<StringLite, Command>* cmd_aliasindex;

ObjectType PDF_blob::get_type() {return t_blob;}
ObjectType PDF_null::get_type() {return t_null;}
ObjectType PDF_boolean::get_type() {return t_boolean;}
ObjectType PDF_int::get_type() {return t_integer;}
ObjectType PDF_real::get_type() {return t_real;}
ObjectType PDF_string::get_type() {return t_string;}
ObjectType PDF_hexstring::get_type() {return t_hexstring;}
ObjectType PDF_name::get_type() {return t_name;}
ObjectType PDF_array::get_type() {return t_array;}
ObjectType PDF_dictionary::get_type() {return t_dictionary;}
ObjectType PDF_stream::get_type() {return t_stream;}
ObjectType PDF_ref::get_type() {return t_reference;}
ObjectType PDF_root::get_type() {return t_root;}
ObjectType PDF_trailer::get_type() {return t_trailer;}
ObjectType PDF_deletion::get_type() {return t_deletion;}
ObjectType PDF_pscommand::get_type() {return t_pscommand;}
ObjectType PDF_box::get_type() {return t_box;}

char* PDF_blob::get_printable_type() {return "blob";}
char* PDF_null::get_printable_type() {return "null";}
char* PDF_boolean::get_printable_type() {return "boolean";}
char* PDF_int::get_printable_type() {return "int";}
char* PDF_real::get_printable_type() {return "real";}
char* PDF_string::get_printable_type() {return "string";}
char* PDF_hexstring::get_printable_type() {return "hexstring";}
char* PDF_name::get_printable_type() {return "name";}
char* PDF_array::get_printable_type() {return "array";}
char* PDF_dictionary::get_printable_type() {return "dictionary";}
char* PDF_stream::get_printable_type() {return "stream";}
char* PDF_ref::get_printable_type() {return "ref";}
char* PDF_root::get_printable_type() {return "root";}
char* PDF_trailer::get_printable_type() {return "trailer";}
char* PDF_deletion::get_printable_type() {return "deletion";}
char* PDF_pscommand::get_printable_type() {return "pscommand";}
char* PDF_box::get_printable_type() {return "box";}

//----------------------------------------

int PDF_blob::filenum = 1;
 
PDF_blob::PDF_blob(): PObj(), rawdata(NULL), length(0) {
}
 
PDF_blob::~PDF_blob() {
	if(rawdata) delete [] rawdata;
}
 
PObj* PDF_blob::copy() {
    PDF_blob *beta = new PDF_blob();
 
    beta->copyFields(this);
    return beta;
}
 
void PDF_blob::copyFields(PDF_blob *alpha) {
    PObj::copyFields(alpha);

    length = alpha->length;
	if(length > 0) {
    	rawdata = new char[length];
		memcpy(rawdata, alpha->rawdata, length);
	}
}

bool PDF_blob::isBinary() {
	uint i; // Needs to be uint to match variable "length".

	for(i=0; i<length; i++) {
		if( (rawdata[i] < 0x20) && (rawdata[i] != '\n') && (rawdata[i] != '\r')) {
			return true;
		}
	}
	return false;
}

uint PDF_blob::getlength() {
	return length;
}
 
void PDF_blob::putlength(uint newlength) {
	length=newlength;
}
 
void PDF_blob::setValue(char *str) {
	// TEMPORARY - eventually this will need to be able to handle non-null terminated strings
	if(rawdata) delete [] rawdata;
	length = pdfc_strlen(str);
	rawdata = new char[length];
	strcpy(rawdata, str);
} 

void PDF_blob::setValue(char *str, int len) {
	if(rawdata) delete [] rawdata;
	length = len;
	rawdata = new char[len];
	memcpy(rawdata, str, len);
} 

void PDF_blob::setValue(String string) {
	const char *rawstring = string.getrawstring();
	if(rawdata) delete [] rawdata;
	rawdata = new char[string.getlength()+1];
	memcpy(rawdata, rawstring, string.getlength());
	rawdata[string.getlength()] = '\0';
	length = string.getlength();
} 

void PDF_blob::render_xml_start(int depth, String& dest) {
	char buf[256];
	char filename[256];
	String output;

	for(int c=0; c<depth; c++) { // Prepend depth number of tabs.
		dest.append("\t");
	}
	dest.append("<blob");
	if(id != 0) {
		sprintf(buf, "%d", id);
		dest.append(" id=\"");
		dest.append(buf);
		dest.append("\"");
	}
	if(generation != -1) {
		sprintf(buf, "%d", generation);
		dest.append(" generation=\"");
		dest.append(buf);
		dest.append("\"");
	}
	if(decompose) { // Save Blobs to Files
		if(rawdata != NULL) {
			sprintf(filename, "stream%ld-%d.pdfstream", time(NULL), filenum++);
			dest.append(" filename=\"");
			dest.append(filename);
			dest.append("\"");
 
			FILE *streamFp = fopen(filename, "wb");

			if(streamFp == NULL) {
				fprintf(stderr, "Error in PDF_blob::render_xml_start() while trying to open file %s\n", filename);
				exit(1);
			}
			if(fwrite(rawdata, 1, length, streamFp)!=length) { 
				// If length is wrong, Bad Things will happen here.
				fprintf(stderr, "Error in PDF_blob::render_xml_start() while trying to write file %s\n", filename);
				exit(1);
			}
			fclose(streamFp);
		}
		dest.append(">");
	}
	else {
		if(rawdata != NULL) {
			output.setval(rawdata, length);

			if(isBinary()) { 
				if(0) { // Flip Flop for encoding scheme - TEMP ENCODE
					dest.append(" encoding=\"hex\"");
					output = DataFilter::encodeHex(output);
				}
				else {
					dest.append(" encoding=\"base85\"");
					output = DataFilter::encode85(output);
				}
			}
			else {
				output = DataFilter::convertEOLs(output);
			}
		}
		dest.append(">");

		output = DataFilter::encodeXML(output);
		dest.append(output);
	}
}
 
void PDF_blob::render_xml_end(int depth, String& dest) {
	dest.append("</blob>\n");
}
 
void PDF_blob::render_pdf_start(Xreftable* x, String& dest) { // Wrapper.
	if(id!=0) throw new Err(666); // Blobs are only valid as part of streams, never as independent indirect objects.

	char* header="\nstream\n";
	dest.append(header);

	if(rawdata == NULL || length < 1)
		printerr("Error - missing blob data needed for rendering PDF\n");
	dest.append(rawdata, length); // Spit raw binary data into PDF file.  The dictionary for the stream
		// contains and entry for the length, and that length sure had better be right.
	if(x != NULL)
		x->accobjsize(pdfc_strlen(header)+length);
}
 
void PDF_blob::render_pdf_end(Xreftable* x, String& dest) {
	char* footer="\nendstream";
	if(x != NULL)
		x->accobjsize(pdfc_strlen(footer));
	dest.append(footer);
}
 
//----------------------------------------

PDF_inline_image::PDF_inline_image(): PDF_blob() {
}
 
PDF_inline_image::~PDF_inline_image() {
}

void PDF_inline_image::render_xml_start(int depth, String& dest) {
	char buf[256];
	char filename[256];
	String output;

	for(int c=0; c<depth; c++) { // Prepend depth number of tabs.
		dest.append("\t");
	}
	dest.append("<inline_image");
	if(id != 0) {
		sprintf(buf, "%d", id);
		dest.append(" id=\"");
		dest.append(buf);
		dest.append("\"");
	}
	if(generation != -1) {
		sprintf(buf, "%d", generation);
		dest.append(" generation=\"");
		dest.append(buf);
		dest.append("\"");
	}
	if(decompose) { // Save Blobs to Files
		if(rawdata != NULL) {
			sprintf(filename, "stream%ld-%d.pdfstream", time(NULL), filenum++);
			dest.append(" filename=\"");
			dest.append(filename);
			dest.append("\"");
 
			FILE *streamFp = fopen(filename, "wb");

			if(streamFp == NULL) {
				fprintf(stderr, "Error in PDF_blob::render_xml_start() while trying to open file %s\n", filename);
				exit(1);
			}
			if(fwrite(rawdata, 1, length, streamFp)!=length) { 
				// If length is wrong, Bad Things will happen here.
				fprintf(stderr, "Error in PDF_blob::render_xml_start() while trying to write file %s\n", filename);
				exit(1);
			}
			fclose(streamFp);
		}
		dest.append(">");
	}
	else {
		if(rawdata != NULL) {
			output.setval(rawdata, length);

			if(isBinary()) { 
				if(0) { // Flip Flop for encoding scheme - TEMP ENCODE
					dest.append(" encoding=\"hex\"");
					output = DataFilter::encodeHex(output);
				}
				else {
					dest.append(" encoding=\"base85\"");
					output = DataFilter::encode85(output);
				}
			}
			else {
				output = DataFilter::convertEOLs(output);
			}
		}
		dest.append(">");

		output = DataFilter::encodeXML(output);
		dest.append(output);
	}
}
 
void PDF_inline_image::render_xml_end(int depth, String& dest) {
	dest.append("</inline_image>\n");
}
 
void PDF_inline_image::render_pdf_start(Xreftable* x, String& dest) { // Wrapper.
	if(id!=0) throw new Err(666); // Blobs are only valid as part of streams, never as independent indirect objects.

	char* header="\nID\n";
	dest.append(header);

	if(rawdata != NULL && length > 0) {
		dest.append(rawdata, length);
	}
	if(x != NULL)
		x->accobjsize(pdfc_strlen(header)+length);
}
 
void PDF_inline_image::render_pdf_end(Xreftable* x, String& dest) {
	char* footer="EI";
	if(x != NULL)
		x->accobjsize(pdfc_strlen(footer));
	dest.append(footer);
}
 
//----------------------------------------

PObj* PDF_null::copy() {
    PDF_null *beta = new PDF_null();
 
    beta->copyFields(this);
    return beta;
}
 
uint PDF_null::getlength() {return 1;}

void PDF_null::putlength(uint newlength) {} // Intentionally defined as NOP since PDF_null doesn't use length.

void PDF_null::render_xml_start(int depth, String& dest) {
	char buf[256];
	for(int c=0; c<depth; c++) { // Prepend depth number of tabs.
		dest.append("\t");
	}
	dest.append("<null");
	sprintf(buf, "%d", id);
	if(id != 0) {
		dest.append(" id=\"");
		dest.append(buf);
		dest.append("\"");
	}
	sprintf(buf, "%d", generation);
	if(generation != -1) {
		dest.append(" generation=\"");
		dest.append(buf);
		dest.append("\"");
	}
	dest.append(">");
}

void PDF_null::render_xml_end(int depth, String& dest) {
	dest.append("</null>\n");
}

void PDF_null::render_pdf(Xreftable* x, String& dest) {
	char buf[256];
	sprintf(buf, "%d", id);
	if(id!=0) rend_pdf_ind_begin(id, 0, x, dest);
	char* val="\nnull";
	dest.append(val);
	if(x != NULL)
		x->accobjsize(pdfc_strlen(val));
	if(id!=0) rend_pdf_ind_end(x, dest);
}

//----------------------------------------

// Type PDF_boolean
PDF_boolean::PDF_boolean() : PObj() {
	value=false;
}

PDF_boolean::PDF_boolean(bool initVal) : PObj() {
	value=initVal;
}

PObj* PDF_boolean::copy() {
    PDF_boolean *beta = new PDF_boolean();
 
    beta->copyFields(this);
    return beta;
}
 
void PDF_boolean::copyFields(PDF_boolean *alpha) {
    PObj::copyFields(alpha);

	value= alpha->value;
}

uint PDF_boolean::getlength() {return 1;}

void PDF_boolean::putlength(uint newlength) {} // Intentionally defined as NOP since PDF_boolean doesn't use length.

void PDF_boolean::setValue(char *data) {
	if(strcmp(data, "true") == 0)
		value = true;
	else if(strcmp(data, "false") == 0)
		value = false;
	else
		printerr(EFATAL|EPDFIN|EWRONGARGTYPE, "Error parsing boolean value - %s\n", data);
}

void PDF_boolean::render_xml_start(int depth, String& dest) {
	char buf[256];
	for(int c=0; c<depth; c++) { // Prepend depth number of tabs.
		dest.append("\t");
	}
	dest.append("<boolean");
	sprintf(buf, "%d", id);
	if(id != 0) {
		dest.append(" id=\"");
		dest.append(buf);
		dest.append("\"");
	}
	sprintf(buf, "%d", generation);
	if(generation != -1) {
		dest.append(" generation=\"");
		dest.append(buf);
		dest.append("\"");
	}
	dest.append(">");

	if(value==true) {
		dest.append("true");
	} else {
		dest.append("false");
	}
}

void PDF_boolean::render_xml_end(int depth, String& dest) {
	dest.append("</boolean>\n");
}

void PDF_boolean::render_pdf(Xreftable* x, String& dest) {
	char buf[256];
	sprintf(buf, "%d", id);
	if(id!=0) rend_pdf_ind_begin(id, 0, x, dest);
	char* t="\ntrue";
	char* f="\nfalse";
	char* thebool=(value==true)?t:f;
	dest.append(thebool);
	if(x != NULL)
		x->accobjsize(pdfc_strlen(thebool));
	if(id!=0) rend_pdf_ind_end(x, dest);
}

//----------------------------------------

PDF_numeric::PDF_numeric() : PObj() {
}

//----------------------------------------

// Type PDF_int
PDF_int::PDF_int() : PDF_numeric() {
	value=0;
}

PDF_int::PDF_int(int i) : PDF_numeric() {
	value=i;
}

PObj* PDF_int::copy() {
    PDF_int *beta = new PDF_int();
 
    beta->copyFields(this);
    return beta;
}
 
void PDF_int::copyFields(PDF_int *alpha) {
    PObj::copyFields(alpha);

	value = alpha->value;
}

void PDF_int::setNumericValue(double newVal) {
	value = (int) newVal;
}

double PDF_int::getNumericValue() {
	return (double) value;
}

uint PDF_int::getlength() {return sizeof(int);}

void PDF_int::putlength(uint newlength) {}

void PDF_int::setValue(char *data) {
	value = atoi(data);
}

void PDF_int::render_xml_start(int depth, String& dest) {
	char buf[256];
	char buf2[256];
	sprintf(buf2, "%d", value);
	for(int c=0; c<depth; c++) { // Prepend depth number of tabs.
		dest.append("\t");
	}
	dest.append("<int");
	sprintf(buf, "%d", id);
	if(id != 0) {
		dest.append(" id=\"");
		dest.append(buf);
		dest.append("\"");
	}
	sprintf(buf, "%d", generation);
	if(generation != -1) {
		dest.append(" generation=\"");
		dest.append(buf);
		dest.append("\"");
	}
	dest.append(">");

	dest.append(buf2);
}

void PDF_int::render_xml_end(int depth, String& dest) {
	dest.append("</int>\n");
}

void PDF_int::render_pdf(Xreftable* x, String& dest) {
	char buf[256];
	sprintf(buf, "\n%d", value);
	if(id!=0) rend_pdf_ind_begin(id, 0, x, dest);
	if(x != NULL)
		x->accobjsize(pdfc_strlen(buf));
	dest.append(buf);
	if(id!=0) rend_pdf_ind_end(x, dest);
}

//----------------------------------------

// Type PDF_real
PDF_real::PDF_real() : PDF_numeric() {
	value=0;
}

// Type PDF_real
PDF_real::PDF_real(double d) : PDF_numeric() {
	value=d;
}

PObj* PDF_real::copy() {
    PDF_real *beta = new PDF_real();
 
    beta->copyFields(this);
    return beta;
}
 
void PDF_real::copyFields(PDF_real *alpha) {
    PObj::copyFields(alpha);

	value = alpha->value;
}

void PDF_real::setNumericValue(double newVal) {
	value = newVal;
}

double PDF_real::getNumericValue() {
	return value;
}

uint PDF_real::getlength() {return sizeof(double);}

void PDF_real::putlength(uint newlength) {}

void PDF_real::setValue(char *data) {
	value = atof(data);
}

void PDF_real::render_xml_start(int depth, String& dest) {
	char buf[256];
	char buf2[256];
	sprintf(buf2, "%.6E", value);
	for(int c=0; c<depth; c++) { // Prepend depth number of tabs.
		dest.append("\t");
	}
	dest.append("<real");
	sprintf(buf, "%d", id);
	if(id != 0) {
		dest.append(" id=\"");
		dest.append(buf);
		dest.append("\"");
	}
	sprintf(buf, "%d", generation);
	if(generation != -1) {
		dest.append(" generation=\"");
		dest.append(buf);
		dest.append("\"");
	}
	dest.append(">");
	dest.append(buf2);
}

void PDF_real::render_xml_end(int depth, String& dest) {
	dest.append("</real>\n");
}

void PDF_real::render_pdf(Xreftable* x, String& dest) {
	char buf[256];
	sprintf(buf, "\n%f", value);
	if(id!=0) rend_pdf_ind_begin(id, 0, x, dest);
	if(x != NULL)
		x->accobjsize(pdfc_strlen(buf));
	dest.append(buf);
	if(id!=0) rend_pdf_ind_end(x, dest);
}

//----------------------------------------

// Type PDF_string
PDF_string::PDF_string() : PObj() {
}

PDF_string::PDF_string(String newValue) : PObj() {
	value = newValue;
}

PDF_string::PDF_string(char *newValue) : PObj() {
	value = String(newValue);
}

PDF_string::~PDF_string() {} // Intentionally defined as NOP.  The encapsulated String can care for itself.

PObj* PDF_string::copy() {
    PDF_string *beta = new PDF_string();
 
    beta->copyFields(this);
    return beta;
}
 
void PDF_string::copyFields(PDF_string *alpha) {
    PObj::copyFields(alpha);

	value = alpha->value;
}

uint PDF_string::getlength() {return value.getlength();}

void PDF_string::putlength(uint newlength) {}

void PDF_string::setValue(char *data) {
	value = (data);
}

void PDF_string::setValue(String data) {
	value = data;
}

void PDF_string::render_xml_start(int depth, String& dest) {
	char buf[256];
	for(int c=0; c<depth; c++) { // Prepend depth number of tabs.
		dest.append("\t");
	}
	dest.append("<string");
	sprintf(buf, "%d", id);
	if(id != 0) {
		dest.append(" id=\"");
		dest.append(buf);
		dest.append("\"");
	}
	sprintf(buf, "%d", generation);
	if(generation != -1) {
		dest.append(" generation=\"");
		dest.append(buf);
		dest.append("\"");
	}
	dest.append(">");
	dest.append(DataFilter::encodeXML(value));
}

void PDF_string::render_xml_end(int depth, String& dest) {
	dest.append("</string>\n");
}

void PDF_string::render_pdf(Xreftable* x, String& dest) {
	char buf[256];
	sprintf(buf, "%d", id);
	if(id!=0) rend_pdf_ind_begin(id, 0, x, dest);
	String appendage=("\n(");
	appendage.append(DataFilter::encodePDF(String(value)));
	appendage.append(")");
	if(x != NULL)
		x->accobjsize(appendage.getlength());
	dest.append(appendage);
	if(id!=0) rend_pdf_ind_end(x, dest);
}

//----------------------------------------

PObj* PDF_hexstring::copy() {
    PDF_hexstring *beta = new PDF_hexstring();
 
    beta->copyFields(this);
    return beta;
}
 
// Type PDF_hexstring
void PDF_hexstring::setValue(char *data) {
	value = DataFilter::decodeHex(String(data));
}

// Type PDF_hexstring
void PDF_hexstring::setValue(String data) {
	value = DataFilter::decodeHex(data);
}

void PDF_hexstring::render_xml_start(int depth, String& dest) {
	char buf[256];
	for(int c=0; c<depth; c++) { // Prepend depth number of tabs.
		dest.append("\t");
	}
	dest.append("<hexstring");
	sprintf(buf, "%d", id);
	if(id != 0) {
		dest.append(" id=\"");
		dest.append(buf);
		dest.append("\"");
	}
	sprintf(buf, "%d", generation);
	if(generation != -1) {
		dest.append(" generation=\"");
		dest.append(buf);
		dest.append("\"");
	}
	dest.append(">");

	dest.append(DataFilter::encodeXML(DataFilter::encodeHex(value)));
}

void PDF_hexstring::render_xml_end(int depth, String& dest) {
	dest.append("</hexstring>\n");
}

void PDF_hexstring::render_pdf(Xreftable* x, String& dest) {
	char buf[256];
	sprintf(buf, "%d", id);
	if(id!=0) rend_pdf_ind_begin(id, 0, x, dest);
	String appendage=("\n<");
	appendage.append(String(value).hexencode());
	appendage.append(">");
	if(x != NULL)
		x->accobjsize(appendage.getlength());
	dest.append(appendage);
	if(id!=0) rend_pdf_ind_end(x, dest);
}

//----------------------------------------

PDF_name::PDF_name(String name) : PDF_string() {
	setValue(name);
}

PDF_name::PDF_name() : PDF_string() {
}

PObj* PDF_name::copy() {
    PDF_name *beta = new PDF_name();
 
    beta->copyFields(this);
    return beta;
}
 
void PDF_name::render_xml_start(int depth, String& dest) {
	char buf[256];
	for(int c=0; c<depth; c++) { // Prepend depth number of tabs.
		dest.append("\t");
	}
	dest.append("<name");
	sprintf(buf, "%d", id);
	if(id != 0) {
		dest.append(" id=\"");
		dest.append(buf);
		dest.append("\"");
	}
	sprintf(buf, "%d", generation);
	if(generation != -1) {
		dest.append(" generation=\"");
		dest.append(buf);
		dest.append("\"");
	}
	dest.append(">");

	dest.append(DataFilter::encodeXML(value));
}

void PDF_name::render_xml_end(int depth, String& dest) {
	dest.append("</name>\n");
}

void PDF_name::render_pdf(Xreftable* x, String& dest) {
	char buf[256];
	sprintf(buf, "%d", id);
	if(id!=0) rend_pdf_ind_begin(id, 0, x, dest);
	String appendage=("\n/");
	appendage.append(value);
	if(x != NULL)
		x->accobjsize(appendage.getlength());
	dest.append(appendage);
	if(id!=0) rend_pdf_ind_end(x, dest);
}

void PDF_name::trimWhiteSpace() {
	char *input = value.nullterm();
	char *start, *end;
 
	if(input == NULL)
		return;
 
	start = input;
	while(*start != '\0' && (*start == ' ' || *start == '\n' || *start == '\t'))
		start++;
 
	end = start + strlen(start) - 1;
	while(end >= start && (*end == ' ' || *end == '\n' || *end == '\t'))
		end--;
	*(end+1) = '\0';
 
	value = start;

	delete input;
} 

bool PDF_name::isValidName() {
	int i, len;
	const char *str = value.getrawstring();

	len = value.getlength();
	if(len < 1)
		return false;
	// if( (str[0] < 'a' || str[0] > 'z') && (str[0] < 'A' || str[0] > 'Z') ) 
		// return false;
	for(i=0; i<len; i++) {
		if(str[i] == ' ') 
			return false;
	}

	return true;
}

//----------------------------------------

PObj* PDF_array::copy() {
    PDF_array *beta = new PDF_array();
 
    beta->copyFields(this);
    return beta;
}
 
void PDF_array::setValue(char *data) {
}

void PDF_array::render_xml_start(int depth, String& dest) {
	char buf[256];
	for(int c=0; c<depth; c++) { // Prepend depth number of tabs.
		dest.append("\t");
	}
	dest.append("<array");
	sprintf(buf, "%d", id);
	if(id != 0) {
		dest.append(" id=\"");
		dest.append(buf);
		dest.append("\"");
	}
	sprintf(buf, "%d", generation);
	if(generation != -1) {
		dest.append(" generation=\"");
		dest.append(buf);
		dest.append("\"");
	}
	dest.append(">");
}

void PDF_array::render_xml_end(int depth, String& dest) {
	dest.append("</array>\n");
}

void PDF_array::render_pdf_start(Xreftable* x, String& dest) {
	if(id!=0) rend_pdf_ind_begin(id, 0, x, dest);
	char* header="\n[";
	if(x != NULL)
		x->accobjsize(pdfc_strlen(header));
	dest.append(header);
}

void PDF_array::render_pdf_end(Xreftable* x, String& dest) {
	char* footer="]";
	if(x != NULL)
		x->accobjsize(pdfc_strlen(footer));
	dest.append(footer);
	if(id!=0) rend_pdf_ind_end(x, dest);
}

//----------------------------------------

PObj* PDF_dictionary::copy() {
    PDF_dictionary *beta = new PDF_dictionary();
 
    beta->copyFields(this);
    return beta;
}
 
void PDF_dictionary::setValue(char *data) {
}

void PDF_dictionary::render_xml_start(int depth, String& dest) {
	char buf[256];
	for(int c=0; c<depth; c++) { // Prepend depth number of tabs.
		dest.append("\t");
	}
	dest.append("<dictionary");
	sprintf(buf, "%d", id);
	if(id != 0) {
		dest.append(" id=\"");
		dest.append(buf);
		dest.append("\"");
	}
	sprintf(buf, "%d", generation);
	if(generation != -1) {
		dest.append(" generation=\"");
		dest.append(buf);
		dest.append("\"");
	}
	dest.append(">");
}

void PDF_dictionary::render_xml_end(int depth, String& dest) {
	dest.append("</dictionary>\n");
}

void PDF_dictionary::render_pdf_start(Xreftable* x, String& dest) {
	if(id!=0) rend_pdf_ind_begin(id, 0, x, dest);
	char* header="\n<<\n";
	if(x != NULL)
		x->accobjsize(pdfc_strlen(header));
	dest.append(header);
}

void PDF_dictionary::render_pdf_end(Xreftable* x, String& dest) {
	char* footer="\n>>\n";
	if(x != NULL)
		x->accobjsize(pdfc_strlen(footer));
	dest.append(footer);
	if(id!=0) rend_pdf_ind_end(x, dest);
}

void PDF_dictionary::showObject() {
	String **keys, **keyPtr;

	printf("PObj: %s\n", get_printable_type());

	keys = getKeys();
	if(keys != NULL) {
		keyPtr = keys;
		while(*keyPtr != NULL) {
			printf(" Key: %s", (*keyPtr)->nullterm() );
			printf(" Value: %s\n", getValueForKey(**keyPtr)->get_printable_type());
		
			delete *keyPtr;
			keyPtr++;
		}
		delete [] keys;
	} 
} 

String** PDF_dictionary::getKeys() {
	int cnt;
	PObjList* list;
	String** keys = NULL;

	cnt = 0;
	list = subobjects;
	while(list != NULL) {
		cnt++;
		list = list->next->next;
	}
	keys = new String*[cnt+1];

	cnt = 0;
	list = subobjects;
	while(list != NULL) {
		if(list->theobject->get_type() == t_name) {
			keys[cnt] = new String( ((PDF_name*) list->theobject)->value );	
			if(list->next == NULL) {
				printerr("Error - invalid number of objects in dictionary\n");
			}
			list = list->next->next;
			cnt++;	
		}
		else {
			printerr("Error - dictionary key not of type name\n");
		}
	}
	keys[cnt] = NULL;

	return keys;	
}

PObj* PDF_dictionary::getValueForKey(String key) {
	String currentKey;

	PObjList *list = subobjects;
 
	while(list != NULL) {
		if(list->theobject->get_type() == t_name) {
			PDF_name *nameObject = (PDF_name*) list->theobject;
			currentKey = nameObject->value;
		}
		else {
			printerr("Error - name expected as key for Dictionary\n");
		}
		if(list->next == NULL) {
			printerr("Error - invalid number of entries in dictionary\n");
		}
		list = list->next;

		if(currentKey == key) {
			return list->theobject;
		}
		list = list->next;
	}
	return NULL;
}

void PDF_dictionary::setValueForKey(String key, PObj *value, bool deleteObjectOnDestruct) {
	String currentKey;

	PObjList *list = subobjects;
 
	while(list != NULL) {
		if(list->theobject->get_type() == t_name) {
			PDF_name *nameObject = (PDF_name*) list->theobject;
			currentKey = nameObject->value;
		}
		else {
			printerr("Error - name expected as key for Dictionary\n");
		}
		if(list->next == NULL) {
			printerr("Error - invalid number of entries in dictionary\n");
		}
		list = list->next;

		if(currentKey == key) {
			if(list->deleteObjectOnDestruct)
				delete list->theobject;

			list->theobject = value;
			list->deleteObjectOnDestruct = deleteObjectOnDestruct;
			value->parentp = this;
			return;
		}
		list = list->next;
	}

	addsub(new PDF_name(key));
	addsub(value, deleteObjectOnDestruct);
}

//----------------------------------------

PDF_stream::PDF_stream() : PObj() {
	dictionary = new PDF_dictionary();
	addsub(dictionary);
	dataObject = new PDF_blob();
	addsub(dataObject);
}

PDF_stream::PDF_stream(PDF_dictionary *newDict, PDF_blob *newBlob) : PObj() {
	if(newDict == NULL && newBlob == NULL) {
		dictionary = NULL;
		dataObject = NULL;
		return;
	}
	else if( (newDict == NULL && newBlob != NULL) || (newDict != NULL && newBlob == NULL) ) {
		printerr("Error - a new PDF_stream constuctor may not contain only one NULL argument\n");
	}

	dictionary = newDict;
	addsub(dictionary);
	dataObject = newBlob;
	addsub(dataObject);
}

PDF_stream::PDF_stream(String newData) : PObj() {
	dictionary = new PDF_dictionary();
	addsub(dictionary);
	dataObject = new PDF_blob();
	dataObject->setValue(newData);
	addsub(dataObject);
}

PDF_stream::~PDF_stream() {
}

PObj* PDF_stream::copy() {
    PDF_stream *beta = new PDF_stream();
 
    beta->copyFields(this);
    return beta;
}
 
void PDF_stream::copyFields(PDF_stream *alpha) {
    PObj::copyFields(alpha);

	if(subobjects == NULL || subobjects->theobject == NULL || subobjects->theobject->get_type() != t_dictionary)
		printerr("Error copying stream dictionary\n");
	if(subobjects->next == NULL || subobjects->next->theobject == NULL || subobjects->next->theobject->get_type() != t_blob)
		printerr("Error copying stream dataObject\n");

	dictionary = (PDF_dictionary*) subobjects->theobject;
	dataObject = (PDF_blob*) subobjects->next->theobject;
}                                                                                                                                        

void PDF_stream::setValue(char *data) {
}

void PDF_stream::render_xml_start(int depth, String& dest) {
	char buf[256];
	for(int c=0; c<depth; c++) { // Prepend depth number of tabs.
		dest.append("\t");
	}
	dest.append("<stream");
	sprintf(buf, "%d", id);
	if(id != 0) {
		dest.append(" id=\"");
		dest.append(buf);
		dest.append("\"");
	}
	sprintf(buf, "%d", generation);
	if(generation != -1) {
		dest.append(" generation=\"");
		dest.append(buf);
		dest.append("\"");
	}
	dest.append(">");
}

void PDF_stream::render_xml_end(int depth, String& dest) {
	dest.append("</stream>\n");
}

void PDF_stream::render_pdf_start(Xreftable* x, String& dest) {
	if(id==0) printerr("Error - all streams must have an associated object id\n");

	dictionary->setValueForKey("Length", new PDF_int(dataObject->length));
	rend_pdf_ind_begin(id, 0, x, dest);
}

void PDF_stream::render_pdf_end(Xreftable* x, String& dest) {
	if(id==0) printerr("Error - all streams must have an associated object id\n");
	rend_pdf_ind_end(x, dest);
}

//----------------------------------------

PObj* PDF_ref::copy() {
    PDF_ref *beta = new PDF_ref();
 
    beta->copyFields(this);
    return beta;
}
 
void PDF_ref::copyFields(PDF_ref *alpha) {
    PObj::copyFields(alpha);

	ref = alpha->ref;
}                                                                                                                                        

void PDF_ref::PDF_ref_init() {
	ref=0;
	generation=0;
}

// Type PDF_ref
PDF_ref::PDF_ref() : PObj() {
	PDF_ref_init();
}

PDF_ref::PDF_ref(int newId) : PObj() {
	PDF_ref_init();

	ref = newId;
}

PDF_ref::PDF_ref(PObj* obj) {
	PDF_ref_init();

	if(obj == NULL || obj->id == 0) {
		printf("Invalid ref object\n");
		return;
	}

	ref = obj->id;
}

void PDF_ref::setValue(char *data) {
}

void PDF_ref::render_xml_start(int depth, String& dest) {
	char buf[256];
	for(int c=0; c<depth; c++) { // Prepend depth number of tabs.
		dest.append("\t");
	}
	dest.append("<reference");
	sprintf(buf, "%d", ref);
	dest.append(" ref_id=\"");
	dest.append(buf);
	dest.append("\"");
	sprintf(buf, "%d", generation);
	dest.append(" generation=\"");
	dest.append(buf);
	dest.append("\"");
	dest.append(">");
}

void PDF_ref::render_xml_end(int depth, String& dest) {
	dest.append("</reference>\n");
}

void PDF_ref::render_pdf(Xreftable* x, String& dest) {
	char buf[256];
	sprintf(buf, "\n%d %d R", ref, generation);
	if(x != NULL)
		x->accobjsize(pdfc_strlen(buf));
	dest.append(buf);
}

PObj* PDF_ref::getObject(PObj* root) {
	// printf("PObj ID: %d\n", ref);
 
	if(root == NULL)
		return NULL;

	return ((PDF_root*) root)->idHash.findobject(ref);
}

//----------------------------------------

PDF_root::PDF_root(): PObj(), rootNode(NULL) {
	trailer = new PDF_trailer;
	trailer->dictionary = new PDF_dictionary();
	trailer->addsub(trailer->dictionary);
}
 
PDF_root::~PDF_root() {
	delete trailer;
}
 
PObj* PDF_root::copy() {
	printerr("Error - copying PDF_root is currently unsupported\n");

    PDF_root *beta = new PDF_root();
 
    beta->copyFields(this);
    return beta;
}
 
void PDF_root::copyFields(PDF_root *alpha) {
    PObj::copyFields(alpha);

	// The fields need to be handled before the copy constructor can be used

	// rootNode
	// idHash
	// trailer

}                                                                                                                                        

void PDF_root::setValue(char *data) {
}

void PDF_root::render_xml_start(int depth, String& dest) {
	int c;

	for(c=0; c<depth; c++) { // Prepend depth number of tabs.
		dest.append("\t");
	}
	dest.append("<!DOCTYPE pdf SYSTEM \"pdf.dtd\">\n");
	for(c=0; c<depth; c++) { // Prepend depth number of tabs.
		dest.append("\t");
	}
	dest.append("<pdf>");
}

void PDF_root::render_xml_end(int depth, String& dest) {
	dest.append("</pdf>\n");
}

void PDF_root::render_pdf_start(Xreftable* x, String& dest) {
	x = render_pdf_top_level_start(x);

	// Dealing the the offsets in the xref table is done by rend_pdf_ind_begin(), so nothing is needed here.
	dest.append(PDF_FILE_HEADER); // Standard PDF header.
}

void PDF_root::render_pdf_end(Xreftable* x, String& dest) {
	render_pdf_top_level_end(x);
}

void PDF_root::updateTrailer(PDF_trailer *newTrailer) {
	String **keys, **keyPtr;
	PDF_dictionary *newDict;
	PDF_dictionary *mainDict = trailer->dictionary;

	if(newTrailer == NULL || newTrailer->dictionary == NULL)
		return;

	newDict = newTrailer->dictionary;

	keys = newDict->getKeys();
	if(keys != NULL) {
		keyPtr = keys;
		while(*keyPtr != NULL) {
			mainDict->setValueForKey(**keyPtr, newDict->getValueForKey(**keyPtr), false );
			delete *keyPtr;
			keyPtr++;
		}
		delete [] keys;
	} 
}

//----------------------------------------

PDF_trailer::PDF_trailer(): PObj(), dictionary(NULL) {
}
 
PDF_trailer::~PDF_trailer() {
}
 
PObj* PDF_trailer::copy() {
    PDF_trailer *beta = new PDF_trailer();
 
    beta->copyFields(this);
    return beta;
}
 
void PDF_trailer::copyFields(PDF_trailer *alpha) {
    PObj::copyFields(alpha);

	// dictionary
}                                                                                                                                        

void PDF_trailer::setValue(char *data) {
}

void PDF_trailer::render_xml_start(int depth, String& dest) {
	int c;

	for(c=0; c<depth; c++) { // Prepend depth number of tabs.
		dest.append("\t");
	}
	dest.append("<trailer");
	dest.append(">");
}

void PDF_trailer::render_xml_end(int depth, String& dest) {
	dest.append("</trailer>\n");
}

void PDF_trailer::render_pdf_start(Xreftable* x, String& dest) {
	char buf[256];
	uint i;
	uint *orderedIndex = sortXrefEntries(x);
	uint startIndex, index;
	uint prevEntry;
	bool firstPass = true;

	startIndex = 0;
	while(startIndex < x->arraysize) {
		if(startIndex == 0 && firstPass) 
			prevEntry = 0;
		else
			prevEntry = x->entry[orderedIndex[startIndex]].objnumber - 1;

		for(index=startIndex; 
		    index < x->arraysize && x->entry[orderedIndex[index]].objnumber == prevEntry+1; 
		    index++) {

			prevEntry = x->entry[orderedIndex[index]].objnumber;
		}

		if(startIndex == 0 && firstPass) {
			sprintf(buf, "xref\n0 %d\n0000000000 65535 f \n", index - startIndex + 1 );
			firstPass = false;
		}
		else {
			sprintf(buf, "%d %d\n", x->entry[orderedIndex[startIndex]].objnumber, index - startIndex );
		}
		dest.append(buf);
		

		for(i=startIndex; i < index; i++) {
			sprintf(buf, "%.10d %.5d n \n", x->entry[orderedIndex[i]].offset, x->entry[orderedIndex[i]].gennumber);
			dest.append(buf);
		}

		startIndex = index;
	}

	dest.append("trailer"); 

	delete [] orderedIndex;
}

uint* PDF_trailer::sortXrefEntries(Xreftable *xrefTable) {
	uint i, j, numLeft, temp;
	uint *orderedIndex = new uint[xrefTable->arraysize];

	for(i=0; i<xrefTable->arraysize; i++) {
		orderedIndex[i] = i;
	}

	for(i=0; i<xrefTable->arraysize; i++) {
		numLeft = xrefTable->arraysize - i - 1;
		for(j=0; j<numLeft; j++) {
			if(xrefTable->entry[orderedIndex[j]].objnumber > xrefTable->entry[orderedIndex[j+1]].objnumber) {
				temp = orderedIndex[j];
				orderedIndex[j] = orderedIndex[j+1];
				orderedIndex[j+1] = temp;
			}
		}
	}

	return orderedIndex;
}

void PDF_trailer::render_pdf_end(Xreftable* x, String& dest) {
	dest.append("\nstartxref");
	uint xrefoffset=(x->entry[x->arraysize-1].offset)+(x->entry[x->arraysize-1].totalsize);
		// Xref table immediately follows all indirect objects, so the xref table offset is calculated by
		// summing the offset and size of the last indirect object.
	char buf[256];
	sprintf(buf, "\n%d\n%%%%EOF\n", xrefoffset);
	dest.append(buf);
}

//----------------------------------------

PObj* PDF_deletion::copy() {
    PDF_deletion *beta = new PDF_deletion();
 
    beta->copyFields(this);
    return beta;
}
 
void PDF_deletion::setValue(char *data) {
}

void PDF_deletion::render_xml_start(int depth, String& dest) {
	char buf[256];
	int c;

	for(c=0; c<depth; c++) { // Prepend depth number of tabs.
		dest.append("\t");
	}
	dest.append("<deletion");
	sprintf(buf, "%d", id);
	dest.append(" id=\"");
	dest.append(buf);
	dest.append("\"");
	dest.append(">");
}

void PDF_deletion::render_xml_end(int depth, String& dest) {
	dest.append("</deletion>\n");
}

void PDF_deletion::render_pdf_start(Xreftable* x, String& dest) {
	dest.append(""); 
}

void PDF_deletion::render_pdf_end(Xreftable* x, String& dest) {
	dest.append("");
}

//----------------------------------------
 
PDF_shared_collection::PDF_shared_collection() : PDF_collection(), currentStream(NULL) {
}
 
PObj* PDF_shared_collection::copy() {
    PDF_shared_collection *beta = new PDF_shared_collection();
 
    beta->copyFields(this);
    return beta;
}
 
ObjectType PDF_shared_collection::get_type() {return t_shared_collection;}
char* PDF_shared_collection::get_printable_type() {return "shared_collection";}
 
void PDF_shared_collection::addNewStream(bool deleteObjectOnDestruct) {
	addNewStream(new PDF_collection(), deleteObjectOnDestruct);
}

void PDF_shared_collection::addNewStream(PDF_collection *collection, bool deleteObjectOnDestruct) {

	PObjList* newlist=new PObjList(deleteObjectOnDestruct);
	newlist->theobject=collection;
	newlist->next=NULL;

	PObjList** lpp=&subobjects;
	if(*lpp==NULL) { // No subobjects present.  Adding the first one.
		*lpp=newlist;
		lastsub=newlist;
	} else {
		lastsub->next=newlist;
		lastsub=newlist;
	}
	collection->parentp=this; // Establish link from subobject to superobject (which is this).

	currentStream = collection;
}

void PDF_shared_collection::addsub(PObj* newsub, bool deleteObjectOnDestruct) {
	if(newsub == NULL)
		return;

	if(currentStream == NULL) {
		addNewStream(deleteObjectOnDestruct);
	}

	currentStream->addsub(newsub, deleteObjectOnDestruct);
} 

//----------------------------------------

PDF_pscommand::PDF_pscommand() {
	PDF_string();
	numAttr=0;
	attrName=NULL;
	attrValue=NULL;
	contentstring=NULL;
}

PDF_pscommand::PDF_pscommand(String name) {
	PDF_string();
	numAttr=0;
	attrName=NULL;
	attrValue=NULL;
	contentstring=NULL;
	value = name;
}

PObj* PDF_pscommand::copy() {
    PDF_pscommand *beta = new PDF_pscommand();
 
    beta->copyFields(this);
    return beta;
}
 
void PDF_pscommand::copyFields(PDF_pscommand *alpha) {
    PObj::copyFields(alpha);

	int i;

    value = alpha->value;

    numAttr = alpha->numAttr;
	if(numAttr > 0) {
		attrName = new String[numAttr];
		attrValue = new String[numAttr];
		for(i=0; i<numAttr; i++) {
			attrName[i] = alpha->attrName[i];
			attrValue[i] = alpha->attrValue[i];
		}
	}

	if(alpha->contentstring != NULL) {
		contentstring = new char[strlen(alpha->contentstring)+1];
    	strcpy(contentstring, alpha->contentstring);
	}
}                                                                                                                                        

// Type PDF_pscommand
uint PDF_pscommand::getlength() {return value.getlength();}

void PDF_pscommand::putlength(uint newlength) {}

void PDF_pscommand::setValue(char *data) {
	value = (data);
}

void PDF_pscommand::setValue(String data) {
	value = data;
}

void PDF_pscommand::render_xml_start(int depth, String& dest) {
	for(int c=0; c<depth; c++) { // Prepend depth number of tabs.
		dest.append("\t");
	}
	dest.append("<");
	if(id != 0) printerr(EFATAL|EXMLOUT|EINTERNALERROR, "Postscript command \"%s\" has a nonzero ID, which is invalid.", value.nullterm());
	if(generation != -1) printerr(EFATAL|EXMLOUT|EINTERNALERROR, "Postscript command \"%s\" has non-negativeone generation, which is invalid.", value.nullterm());
	dest.append(DataFilter::encodeXML(value));
	dest.append(">");
}

void PDF_pscommand::render_xml_end(int depth, String& dest) {
	dest.append("</");
	dest.append(value);
	dest.append(">\n");
}

void PDF_pscommand::render_pdf(Xreftable* x, String& dest) {
	PObj* thetree=render_complex_command_to_primitives(this, getParentPage());
	if(thetree!=this) { // Command was complex.
		thetree->render_pdf(x, dest);
	} else { // Command was primitive.
		PObj::render_pdf(x, dest); // Use the usual semantics.
	}
}

void PDF_pscommand::render_pdf_start(Xreftable* x, String& dest) {
	// Actually rendering this command is done in render_pdf_end() so that this command is rendered _after_
	// its children (i.e. its arguments) are rendered, so that the final output is postfix, per the PDF spec.
	dest.append("");
}

void PDF_pscommand::render_pdf_end(Xreftable* x, String& dest) {
	char* thenewvalue=get_realname_primitive(value);
	if(thenewvalue==NULL) {
		dest.append(" ");
		dest.append(value);
	} else {
		dest.append(" ");
		dest.append(thenewvalue);
	}
	if(x != NULL)
		x->accobjsize(pdfc_strlen(thenewvalue)+1); // +1 for " " above.
	char* special="Tf"; // Special command for which something extra must be done.
	if((value==special)||(value==get_alias_primitive(special))) {
		// Tf command is being rendered, which changes the current font, so deal with it here.
		if(countsubs()!=2) printerr(EFATAL|EPDFOUT|EWRONGARGCOUNT, "postscript Tf command wasn't passed exactly two arguments.\n");
		if(subobjects->theobject->get_type()!=t_name) printerr(EFATAL|EPDFOUT|EWRONGARGTYPE, "first argument to postscript Tf command wasn't a name.\n");
		if(subobjects->next->theobject->get_type()!=t_integer && subobjects->next->theobject->get_type()!=t_real) printerr(EFATAL|EPDFOUT|EWRONGARGTYPE, "second argument to postscript Tf command wasn't an integer or a real.\n");
	}
	special="Tm";
	if((value==special)||(value==get_alias_primitive(special))) {
		if(countsubs()!=6) printerr(EFATAL|EPDFOUT|EWRONGARGCOUNT, "postscript Tm command wasn't passed exactly six arguments.\n");

		// Needs to take floats too.
		//if(subobjects->theobject->get_type()!=t_integer) printerr(EFATAL|EPDFOUT|EWRONGARGTYPE, "first argument to postscript Tm command wasn't an integer.\n");
		//if(subobjects->next->next->next->theobject->get_type()!=t_integer) printerr(EFATAL|EPDFOUT|EWRONGARGTYPE, "fourth argument to postscript Tm command wasn't an integer.\n");

		// PdfPage *parent = getParentPage();
		// if(parent != NULL) {
			// getParentPage()->textmatrix_x=((PDF_int*)(this->subobjects->theobject))->value;
			// getParentPage()->textmatrix_y=((PDF_int*)(this->subobjects->next->next->next->theobject))->value;
		// }
	}
}

//----------------------------------------

PDF_box::PDF_box(): PObj(), x(0), y(0), width(0), height(0) {}

PDF_box::PDF_box(int newX, int newY, int newWidth, int newHeight): PObj() {
	x = newX;
	y = newY;
	width = newWidth;
	height = newHeight;
}

PObj* PDF_box::copy() {
    PDF_box *beta = new PDF_box();
 
    beta->copyFields(this);
    return beta;
}
 
void PDF_box::copyFields(PDF_box *alpha) {
    PObj::copyFields(alpha);

	x = alpha->x;
	y = alpha->y;
	width = alpha->width;
	height = alpha->height;
}                                                                                                                                        

void PDF_box::render_xml_start(int depth, String& dest) {
	char buf[256];
	int c;

	for(c=0; c<depth; c++) { // Prepend depth number of tabs.
		dest.append("\t");
	}
	sprintf(buf, "<box x=\"%d\" y=\"%d\" width=\"%d\" height=\"%d\">", x, y, width, height);
	dest.append(buf);
}

void PDF_box::render_xml_end(int depth, String& dest) {
	dest.append("</box>");
}

void PDF_box::render_pdf_start(Xreftable* x, String& dest) {
	printerr(EFATAL|EPDFOUT|EINTERNALERROR, "PDF_box::render_pdf_start() got called!  That should never happen.\n");
	dest.append("The program that generated this PDF file is really buggy.\n");
}

void PDF_box::render_pdf_end(Xreftable* x, String& dest) {
	printerr(EFATAL|EPDFOUT|EINTERNALERROR, "PDF_box::render_pdf_end() got called!  That should never happen.\n");
	dest.append("The program that generated this PDF file is really buggy.\n");
}

void PDF_box::setValue(int newx, int newy, int newwidth, int newheight) {
	x=newx;
	y=newy;
	width=newwidth;
	height=newheight;
}

//----------------------------------------

void init_pscommands() {
	cmd_nameindex=new HashTable<StringLite, Command>;
	cmd_aliasindex=new HashTable<StringLite, Command>;
	for(int i=0; i<TOTAL_PRIMITIVE_COMMANDS; i++) {
		char* c=cmdlist[i].realname;
		cmd_nameindex->add(StringLite(c, pdfc_strlen(c)), cmdlist[i]);
		c=cmdlist[i].alias;
		cmd_aliasindex->add(StringLite(c, pdfc_strlen(c)), cmdlist[i]);
	}
	pscommands_complex[0]="init_text_page"; // Deprecated.
	pscommands_complex[1]="close_text_page"; // Deprecated.
	pscommands_complex[2]="drawstring";
	pscommands_complex[3]="drawimage";
}

int get_number_of_args_primitive(String& arg) {
	if(DEBUG_FUNCTIONBOUNDARIES) printf("get_number_of_args_primitive(String&) beginning.\n");
	char* argraw=const_cast<char*>(arg.getrawstring());
	uint arglen=arg.getlength();
	const struct Command* thing=cmd_aliasindex->getValue(StringLite(argraw, arglen));
	if(thing!=NULL) return thing->args;
	const struct Command* thing2=cmd_nameindex->getValue(StringLite(argraw, arglen));
	if(thing2!=NULL) return thing2->args;
	return -1; // No match found.
}

char* get_alias_primitive(char* realname) {
	if(DEBUG_FUNCTIONBOUNDARIES) printf("get_alias_primitive(char*) beginning.\n");
	const struct Command* thing=cmd_nameindex->getValue(StringLite(realname, pdfc_strlen(realname)));
	if(thing!=NULL) return thing->alias;
	return NULL;
}

char* get_realname_primitive(String& alias) {
	char* aliasraw=const_cast<char*>(alias.getrawstring());
	uint aliaslen=alias.getlength();
	const struct Command* thing=cmd_aliasindex->getValue(StringLite(aliasraw, aliaslen));
	if(thing!=NULL) return thing->realname;
	return NULL;
}

char* get_alias_primitive(String& realname) {
	if(DEBUG_FUNCTIONBOUNDARIES) printf("get_alias_primitive(String&) beginning.\n");
	char* realnameraw=const_cast<char*>(realname.getrawstring());
	uint realnamelen=realname.getlength();
	const struct Command* thing=cmd_nameindex->getValue(StringLite(realnameraw, realnamelen));
	if(thing!=NULL) {
		return thing->alias;
	}
	return NULL;
}

bool name_isvalid_primitive(char* name) {
	if(DEBUG_FUNCTIONBOUNDARIES) printf("name_isvalid_primitive(char*) beginning.\n");
	uint namelen=pdfc_strlen(name);
	const struct Command* thing=cmd_nameindex->getValue(StringLite(name, namelen));
	if(thing!=NULL) return true;
	const struct Command* thing2=cmd_aliasindex->getValue(StringLite(name, namelen));
	if(thing2!=NULL) return true;
	return false;
}

bool name_isvalid_complex(char* name) {
	if(DEBUG_FUNCTIONBOUNDARIES) printf("name_isvalid_complex(char*) beginning.\n");
	for(uint x=0; x<NUMBER_OF_COMPLEX_COMMANDS; x++) {
		if(strcmp(name, pscommands_complex[x])==0) return true;
	}
	return false;
}

bool name_isvalid_complex(String& name) {
	if(DEBUG_FUNCTIONBOUNDARIES) printf("name_isvalid_complex(String&) beginning.\n");
	for(uint x=0; x<NUMBER_OF_COMPLEX_COMMANDS; x++) {
		if(name==(pscommands_complex[x])) return true;
	}
	return false;
}

char* standard_text_page_initialization_sequence="\
/DocumentSetup<</PageOrigin [13 13 ]/RulerOrigin [0 0 ]>> BDC EMC\n\
/Layer<</Title (Layer 1)/Visible true /Preview true /Editable true /Printed true /Dimmed false /Color [20224 32768 65535 ]>> BDC\n\
/Attribute<</ShowCenterPoint false >> DP\n\
/PointText<</Matrix [1 0 0 1 0 0 ]/Vertical false >> BDC\n\
/Pgf BMC\n"; // Deprecated.

char* standard_text_page_closing_sequence="\nEMC EMC EMC\n"; // Deprecated.

PObj* render_complex_command_to_primitives(PObj* o, PdfPage* currentpage) {
	uint c;

	if(o->get_type()!=t_pscommand) return o; // Only deal with pscommand objects.
	if(!name_isvalid_complex(((PDF_pscommand*)o)->value)) {
		return o; // Only deal with complex commands here.
	}
	if(DEBUG_LOG_NULLTERM) printf("nullterm() #11.\n");
	char* name=((PDF_pscommand*)o)->value.nullterm();
	PObj* retval=NULL;
	PDF_pscommand* p=NULL;
	if(strcmp(name, "init_text_page")==0) { // Deprecated.
		if(o->countsubs()!=0) 
			printerr(EFATAL|EPDFOUT|EWRONGARGCOUNT, "incorrect number of arguments for complex command %s\n", name);
		p=new PDF_pscommand();
		p->value=(standard_text_page_initialization_sequence);
		retval=p;
	}
	else if(strcmp(name, "close_text_page")==0) { // Deprecated.
		if(o->countsubs()!=0) 
			printerr(EFATAL|EPDFOUT|EWRONGARGCOUNT, "incorrect number of arguments for complex command %s\n", name);
		p=new PDF_pscommand();
		p->value=(standard_text_page_closing_sequence);
		retval=p;
	} else if(strcmp(name, "drawstring")==0) {
		uint numsubs=o->countsubs();

		if(numsubs < 3) 
			printerr(EFATAL|EPDFOUT|EWRONGARGCOUNT, "at least tree arguments (a font id, a font size, and a box) must be specified for drawstring.\n");

		PObjList* stepper=o->subobjects;
		PObjList* boxes = NULL;

		if(stepper->theobject->get_type()!=t_name) 
			printerr(EFATAL|EPDFOUT|EWRONGARGTYPE, "incorrect argument type for complex command \"drawstring\" a name is required for the font id)\n");
		String fontId = ((PDF_name*) stepper->theobject)->value;
		stepper=stepper->next;

		PdfResource *resource = currentpage->getResource(fontId);
		if(resource == NULL)
			printerr("Error - could not find font '%s' specified for drawstring command\n", fontId.nullterm());
		if(resource->type != "Font")
			printerr("Error - invalid resource type passed as font to drawstring command\n");
	 
		PdfFont *font = (PdfFont*) resource;

		if(stepper->theobject->get_type()!=t_integer) 
			printerr(EFATAL|EPDFOUT|EWRONGARGTYPE, "incorrect argument type for complex command \"drawstring\" an int is required for the font size)\n");
		int fontSize = ((PDF_int*) stepper->theobject)->value;
		stepper=stepper->next;

		boxes = stepper;

		while(stepper != NULL) {
			if(stepper->theobject->get_type()!=t_box) 
			printerr(EFATAL|EPDFOUT|EWRONGARGTYPE, "incorrect argument type for complex command \"drawstring\" (only allowed type is \"box\")\n");
			stepper=stepper->next;
		}

		PDF_collection* collection=new PDF_collection;
		collection->parentPage = currentpage;
		retval=collection;

		char* temp=((PDF_pscommand*)o)->contentstring; // Get the actual content string for the drawstring element.
		String s=StringFormatter::stripCR(temp);
		// Why did I write the next line?!  Am I really supposed to delete temp here?
		// delete [] temp;
		temp=s.nullterm();
		char* leftovers;
		char* currentposition=temp;
		int x, y, width, height;
		stepper=boxes;

		while(stepper != NULL) {
			x=((PDF_box*)(stepper->theobject))->x;
			y=((PDF_box*)(stepper->theobject))->y;
			width=((PDF_box*)(stepper->theobject))->width;
			height=((PDF_box*)(stepper->theobject))->height;
			leftovers=NULL;
			s=StringFormatter::wrapstring(currentposition, font, width, height, true, leftovers, fontSize); // The "true" is to hardcode-enable word wrapping.
			currentposition=leftovers;
			if(DEBUG_LOG_NULLTERM) printf("nullterm() #12.\n");
			char* temp2=s.nullterm();

			retval->addsub(StringFormatter::renderLF(temp2, x, y, currentpage, fontId, fontSize)); // Generate the primitive postscript commands that render the text.

			delete [] temp2;
			temp2=NULL;
			stepper=stepper->next;
			if(leftovers==NULL) break; // All of the text has been processed; leftover boxes are ignored.
		}

		if(leftovers!=NULL) printerr(EMINORWARN|EPDFOUT|ETOOMUCHDATA, "too much text to fit in specified box(es).  Remaining text is:\n%s\n", leftovers);
		delete [] temp;
		temp=NULL;
	} 

	else if(strcmp(name, "drawimage")==0) {
		uint numsubs=o->countsubs();
		if(numsubs!=1) printerr(EFATAL|EPDFOUT|EWRONGARGCOUNT, "exactly one argument (a box) must be specified for drawimage.\n");
		if(o->subobjects->theobject->get_type()!=t_box) printerr(EFATAL|EPDFOUT|EWRONGARGTYPE, "incorrect argument type for complex command \"drawimage\" (only allowed type is \"box\")\n");
		int x, y, width, height;
		x=((PDF_box*)(o->subobjects->theobject))->x;
		y=((PDF_box*)(o->subobjects->theobject))->y;
		width=((PDF_box*)(o->subobjects->theobject))->width;
		height=((PDF_box*)(o->subobjects->theobject))->height;
		PDF_collection* collection=new PDF_collection;
		retval=collection;
		p=new PDF_pscommand();
		p->setValue("save_graphics_state");
		retval->addsub(p);

		p=new PDF_pscommand();
		p->setValue("concatenate_CTM_matrix");
		PObj* concat=p;
		PDF_int* something;
		// Set up the matrix.
		something=new PDF_int(width);
		concat->addsub(something);
		something=new PDF_int(0);
		concat->addsub(something);
		something=new PDF_int(0);
		concat->addsub(something);
		something=new PDF_int(height);
		concat->addsub(something);
		something=new PDF_int(x);
		concat->addsub(something);
		something=new PDF_int(y);
		concat->addsub(something);
		retval->addsub(concat);
		// Place the image on the page.
		p=new PDF_pscommand();
		p->setValue("execute_xobject");

		// Get the name of the image resource.
		PDF_name *imageName = new PDF_name( ((PDF_pscommand*)o)->contentstring ); 
		imageName->trimWhiteSpace();

		p->addsub(imageName);
		retval->addsub(p);

		p=new PDF_pscommand();
		p->setValue("restore_graphics_state");
		retval->addsub(p);
	} else {
		printerr(EFATAL|EPDFOUT|EUNKNOWNCOMMAND, "attempted to render unknown pscommand: %s\n", name);
	}
	delete [] name;
	return retval;
}

